// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include <set>
#include "nvgl/nvgl.h"
#include "nvutil/RCObject.h"

namespace nvsg
{
  class Texture;
}

namespace nvgl
{
  //! The GLObjects class is a container of OpenGL objects.
  class GLObjects
  {
    friend class GLContext;

    public:
      // Constructor / Destructor
      GLObjects();
      ~GLObjects();

    public:
      //! Add VBO buffers to the set of shareable objects.
      void addBuffers(GLsizei n, const GLuint * names);

      //! Remove VBO buffers to the set of shareable objects.
      void removeBuffers(GLsizei n, const GLuint * names);

      //! Add display lists to the set of shareable objects.
      void addLists(GLsizei n, GLuint first);

      //! Remove display lists from the set of shareable objects
      void removeLists(GLsizei n, GLuint first);

      //! Add programs to the set of shareable objects.
      void addPrograms(GLsizei n, const GLuint * names);

      //! Remove programs from the set of shareable objects
      void removePrograms(GLsizei n, const GLuint * names);

      //! Add textures to the set of shareable objects.
      void addTextures(GLsizei n, const GLuint * names);

      //! Remove textures from the set of shareable objects
      void removeTextures(GLsizei n, const GLuint * names);

      //! Add occlusion queries to the set of non-shareable objects.
      void addOcclusionQueries(GLsizei n, const GLuint * names);

      //! Remove occlusion queries from the set of non-shareable objects
      void removeOcclusionQueries(GLsizei n, const GLuint * names);

      //! Add queries to the set of non-shareable objects.
      void addQueries(GLsizei n, const GLuint * names);

      //! Remove queries from the set of non-shareable objects
      void removeQueries(GLsizei n, const GLuint * names);

    private:
      // Collection of shareable objects
      struct GLShareables : public nvutil::RCObject
      {
        ~GLShareables(); // for cleanup purposes
        
        std::set<GLuint> m_vboNames;
        std::set<GLuint> m_dlistNames;
        std::set<GLuint> m_programNames;
        std::set<GLuint> m_textureNames;
        std::map<std::string, const nvsg::Texture*> m_texLUT;
      };
      
      // shareable objects
      GLShareables * m_shareables;
      // non-shareable objects
      std::set<GLuint> m_occlusionQueryNames;
      std::set<GLuint> m_queryNames;
  };

  inline void GLObjects::addBuffers(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    m_shareables->m_vboNames.insert(&names[0], &names[n]);
  }

  inline void GLObjects::removeBuffers(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    for( GLsizei i=0; i<n; ++i )
    {
      m_shareables->m_vboNames.erase(names[i]);
    }
  }

  inline void GLObjects::addLists(GLsizei n, GLuint first)
  {
    __ASSERT(n>0 && first>0);
    for( GLuint i=first; i<first+n; ++i )
    {
      m_shareables->m_dlistNames.insert(i);
    }
  }

  inline void GLObjects::removeLists(GLsizei n, GLuint first)
  {
    __ASSERT(n>0 && first>0);
    for( GLuint i=first; i<first+n; ++i )
    {
      m_shareables->m_dlistNames.erase(i);
    }
  }

  inline void GLObjects::addPrograms(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    m_shareables->m_programNames.insert(&names[0], &names[n]);
  }

  inline void GLObjects::removePrograms(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    for( GLsizei i=0; i<n; ++i )
    {
      m_shareables->m_programNames.erase(names[i]);
    }
  }

  inline void GLObjects::addTextures(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    m_shareables->m_textureNames.insert(&names[0], &names[n]);
  }

  inline void GLObjects::removeTextures(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    for( GLsizei i=0; i<n; ++i )
    {
      m_shareables->m_textureNames.erase(names[i]);
    }
  }

  inline void GLObjects::addOcclusionQueries(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    m_occlusionQueryNames.insert(&names[0], &names[n]);
  }

  inline void GLObjects::removeOcclusionQueries(GLsizei n, const GLuint * names)

  {
    __ASSERT(n>0 && names);
    for( GLsizei i=0; i<n; ++i )
    {
      m_occlusionQueryNames.erase(names[i]);
    }
  }

  inline void GLObjects::addQueries(GLsizei n, const GLuint * names)
  {
    __ASSERT(n>0 && names);
    m_queryNames.insert(&names[0], &names[n]);
  }

  inline void GLObjects::removeQueries(GLsizei n, const GLuint * names)

  {
    __ASSERT(n>0 && names);
    for( GLsizei i=0; i<n; ++i )
    {
      m_queryNames.erase(names[i]);
    }
  }

} //  namespace nvgl
