// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvmath/Mat44f.h"
#include "nvmath/Quatf.h"

namespace nvmath
{
  //! transformation class.
  /** This class is used to ease transformation handling. It has an interface to rotate, scale, and translate and can
    * produce a Mat44f that combines them.  */
  class Trafo
  {
    public:
      //! Constructor: initialized to identity
      NVSG_API Trafo( void );

      //! Get the inverse of the current transformation.
      /** \returns  inverse transformation as a Mat44f  */
      NVSG_API Mat44f getInverse( void ) const;

      //! Get the center of rotation of this transformation.
      /** \returns  vector that describes the center or rotation. */
      NVSG_API const Vec3f & getCenter( void ) const;

      //! Get the rotational part of this transformation.
      /** \returns  quaternion that describes the rotational part */
      NVSG_API const Quatf & getOrientation( void ) const;

      //! Get the scaling part of this transformation.
      /** \returns  vector that describes the scaling part  */
      NVSG_API const Vec3f & getScaling( void ) const;

      //! Get the current transformation.
      /** \returns  transformation as a Mat44f  */
      NVSG_API Mat44f getMatrix( void ) const;

      //! Get the translational part of this transformation.
      /** \returns  vector that describes the translational part  */
      NVSG_API const Vec3f & getTranslation( void ) const;

      //! Set the center of ration of the transformation.
      NVSG_API void setCenter( const Vec3f &center              //!<  center of rotation
                             );

      //! Set the Trafo to identity.
      NVSG_API void setIdentity( void );

      //! Set the rotational part of the transformation, using a quaternion.
      NVSG_API void setOrientation( const Quatf &orientation    //!<  rotational part of transformation
                                  );

      //! Set the scaling part of the transformation.
      NVSG_API void setScaling( const Vec3f &scaling            //!<  scaling part of transformation
                              );

      //! Set the translational part of the transformation.
      NVSG_API void setTranslation( const Vec3f &translation    //!<  translational part of transformation
                                  );

    private:
      Vec3f   m_center;             //!<  center of rotation
      Quatf   m_orientation;        //!<  orientational part of the transformation
      Vec3f   m_scaling;            //!<  scaling part of the transformation
      Vec3f   m_translation;        //!<  translational part of the transformation
  };

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // non-member functions
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  /*! \relates nvmath::Trafo
   *  Linear interpolation between two trafos \a t0 and \a t1.
   *  t = t0 + alpha * ( t1 - t0 ) */
  inline Trafo lerp( float alpha      //!<  interpolation parameter
                   , const Trafo &t0  //!<  starting value
                   , const Trafo &t1  //!<  ending value
                   )
  {
    Trafo t;
    t.setCenter( lerp( alpha, t0.getCenter(), t1.getCenter() ) );
    t.setOrientation( lerp( alpha, t0.getOrientation(), t1.getOrientation() ) );
    t.setScaling( lerp( alpha, t0.getScaling(), t1.getScaling() ) );
    t.setTranslation( lerp( alpha, t0.getTranslation(), t1.getTranslation() ) );
    return( t );
  }

  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - -  - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  inline Trafo::Trafo( void )
    : m_center( Vec3f( 0.0f, 0.0f, 0.0f ) )
    , m_orientation( Quatf( 0.0f, 0.0f, 0.0f, 1.0f ) )
    , m_scaling( Vec3f( 1.0f, 1.0f, 1.0f ) )
    , m_translation( Vec3f( 0.0f, 0.0f, 0.0f ) )
  {
  };

  inline const Vec3f & Trafo::getCenter( void ) const
  {
    return( m_center );
  }

  inline const Quatf & Trafo::getOrientation( void ) const
  {
    return( m_orientation );
  }

  inline const Vec3f & Trafo::getScaling( void ) const
  {
    return( m_scaling );
  }

  inline const Vec3f & Trafo::getTranslation( void ) const
  {
    return( m_translation );
  }

  inline void Trafo::setCenter( const Vec3f &center )
  {
    m_center = center;
  }

  inline void Trafo::setIdentity( void )
  {
    m_center = Vec3f( 0.0f, 0.0f, 0.0f );
    m_orientation = Quatf( 0.0f, 0.0f, 0.0f, 1.0f );
    m_scaling = Vec3f( 1.0f, 1.0f, 1.0f );
    m_translation = Vec3f( 0.0f, 0.0f, 0.0f );
  }

  inline void Trafo::setOrientation( const Quatf &orientation )
  {
    __ASSERT( isNormalized( orientation ) );
    m_orientation = orientation;
  }

  inline void Trafo::setScaling( const Vec3f &scaling )
  {
    m_scaling = scaling;
  }

  inline void Trafo::setTranslation( const Vec3f &translation )
  {
    m_translation = translation;
  }
}
