// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Animation.h"
#include "nvsg/Quads.h"
#include "nvsg/VNVector.h"

namespace nvsg
{
  //! GeoSet for animated Quads.
  /** This class is a Quads with an Animation<VNVector>.
    * The Animation defines the vertices and normals for each animation step. */
  class AnimatedQuads : public Quads
  {
    public:
      //! Create an AnimatedQuads.
      /** There is an optional argument \a pAnimation to use a given Animation<VNVector>.
        * \returns a constant pointer to an AnimatedQuads. */
      NVSG_API static const AnimatedQuads * create( const Animation<VNVector> *pAnimation = NULL  //!< optional animation to use
                                                  );

      //! Create an AnimatedQuads by copying from a Quads.
      /** There is an optional argument \a pAnimation to use a given Animation<VNVector>.
        * \returns a constant pointer to an AnimatedQuads. */
      NVSG_API static const AnimatedQuads * createFromBase( const Quads &rhs                              //!< Quads to copy from
                                                          , const Animation<VNVector> *pAnimation = NULL  //!< optional animation to use
                                                          );

      //! Create a new AnimatedQuads by copying from this.
      /** \returns a constant pointer to an identical AnimatedQuads. */
      NVSG_API virtual const AnimatedQuads * clone( void ) const;

      //! Get a constant pointer to the Animation.
      /** \returns  constant pointer to the animation */
      NVSG_API const Animation<VNVector> * getAnimation( void ) const;

      //! Set the Animation.
      /** Any previous Animation is dereferenced, the new one is referenced, and the animation frame is reset to zero. */

      NVSG_API void setAnimation( const Animation<VNVector> *pAnimation   //!<  new animation to use
                                );

      //! Set the animation frame.
      /** The vertices/normals for animation step \a frame are used as the current. */
      NVSG_API void setAnimationFrame( size_t frame   //!<  current animation step
                                     );

      //--  Functions reimplemented from Drawable --
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //--  Functions reimplemented from Object  --
      NVSG_API virtual DataID getDataID( void ) const;
      NVSG_API virtual bool isDataShared( void ) const;

    protected:
      //! Constructor with an Animation.
      NVSG_API AnimatedQuads( const Animation<VNVector> *pAnimation );

      //! Constructor with a Quads and an Animation.
      NVSG_API AnimatedQuads( const Quads &rhs, const Animation<VNVector> *pAnimation );

      //! Copy Constructor
      NVSG_API AnimatedQuads( const AnimatedQuads &rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~AnimatedQuads(void);

      // for some extra work after vertex attribute data have been changed
      NVSG_API virtual void onSetVertexAttributeData(void);

    private:
      AnimatedQuads();    //  private default constructor to prevent use

    private:
      const Animation<VNVector> * m_animation;
      size_t                      m_currentFrame;
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline const Animation<VNVector> * AnimatedQuads::getAnimation( void ) const
  {
    __TRACE();
    return( m_animation );
  }

} //  namespace nvsg
