// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvutil/Allocator.h"

namespace nvsg
{  

  //! ID to describe the cull code for an object
  typedef enum
  {
    CC_OUT,   //!<  an object is completely outside
    CC_IN,    //!<  an object is completely inside
    CC_PART   //!<  an object is partially in and partially out
  } CullCode;

  //! Class holding culling information for an object and it's children.
  /** With CullData, a CullTraverser can build a (somewhat simplified) mirror of a scene tree. Each CullData corresponds
    * to one object in a scene tree. Such a CullData tree is used in the Traverser to determine if parts
    * of the scene tree are to be traversed.  */
  class CullData : public nvutil::IAllocator
  {
    public:
      //! Constructor
      NVSG_API CullData( CullCode cullCode );

      //! Destructor
      NVSG_API virtual ~CullData( void );

    public:
      //! Balance the CullData tree.
      /** After having constructed a CullData tree, it might be of value to improve it. When CullData has a CullCode  
        * of CC_PART and the children of CullData have a CullCode of CC_IN to CC_IN, then this function removes
        * these children. */
      NVSG_API void balance( void );

      //! Add a child to this CullData.
      /** CullData builds a somewhat simplified mirror of a scene tree. With \p addChild , the CullData of a child object
        * can be added to the parent object to which the current CullData belongs.  */
      NVSG_API void addChild( CullData *cd    //!<  CullData to add to the back of the parent object's CullData
                            );

      //! Get the \a i'th child of this CullData.
      /** It is assumed that this child does exist. */
      NVSG_API const CullData * getChild( size_t i    //!<  Index of children to get
                                        ) const;

      //! Get the CullCode of this CullData.
      /** \returns  CullCode of this CullData.  */
      NVSG_API CullCode getCullCode( void ) const;

      //! Get the number of children of this CullData.
      /** \returns  Number of children of this CullData. This should be zero or the same as the number of children of the
        * scene tree object that this CullData represents. */
      NVSG_API size_t getNumberOfChildren( void ) const;

    private:
      CullCode          m_cullCode;
      std::vector<CullData*> m_children;
  };

  // - - - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - - - -

  inline  CullData::CullData( CullCode cullCode )
    : m_cullCode( cullCode )
  {
    __TRACE();
  }

  inline  void  CullData::addChild( CullData *cd )
  {
    __TRACE();
    m_children.push_back( cd );
  }

  inline  const CullData  * CullData::getChild( size_t i ) const
  {
    __TRACE();
    __ASSERT( i < m_children.size() );
    return( m_children[i] );
  }

  inline  CullCode  CullData::getCullCode( void ) const
  {
    __TRACE();
    return( m_cullCode );
  }

  inline  size_t    CullData::getNumberOfChildren( void ) const
  {
    __TRACE();
    return( m_children.size() );
  }

} //  namespace nvsg
