// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include "nvsg/Node.h"
#include "nvsg/GeoSet.h"
#include "nvsg/StateSet.h"

// memory leak detection
#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif

namespace nvsg
{

  //! Node to hold a set of geometry.
  /** A GeoNode holds a number of geometries. A Geometry consists of a Drawable - defining the topology, and a StateSet - 
    * defining the visual appearance.
    */
  class GeoNode : public Node
  {
    public:
      //! Create a GeoNode.
      /** \returns a constant pointer to a GeoNode. */
      NVSG_API static const GeoNode * create( void );

      //! Create a GeoNode by copying from a Node.
      /** \returns a constant pointer to a GeoNode. */
      NVSG_API static const GeoNode * createFromBase(const Node &rhs   //!<  Node to copy from
      );

      //! Create a new GeoNode by copying from this.
      /** \returns a constant pointer to an identical GeoNode. */
      NVSG_API virtual GeoNode * clone( void ) const;

      //! Add a Geometry to the GeoNode.
      /** If the GeoNode already holds a geometry with the same Drawable, no geometry is added. Otherwise the pair
        * of \a pDrawable and \a pStateSet is added to the set of geometries.
        * \returns true, if the pair is added, otherwise false */
      NVSG_API bool addGeometry(const Drawable *pDrawable    //!<  Drawable to add
                              , const StateSet *pStateSet    //!<  StateSet to add
      );

      //! Get the number of Geometries of this GeoNode.
      /** \note The number of Drawables is equal to the number of Geometries.
      /** \returns number of Geometries  */
      NVSG_API size_t getNumberOfGeometries( void ) const;

      //! Returns the Drawable at position \a index.
      /** \returns A pointer to a constant Drawable at position \a index.
        * \note The behaviour is undefined if called with an invalid index. 
        * You should call getNumberOfDrawables prior to this call to determine 
        * a valid range of indices that can be passed to this function.  */
      NVSG_API const Drawable * getDrawable(size_t index    //!<  Zero-based index of Drawable to get
                                           ) const;

      //! Get the Drawables that use a given StateSet.
      /** Fills the vector \a v with the pointers to all const Drawables. If \a pStateSet is not NULL, only the
        * Drawables that use that StateSet are returned.
        * \returns number of Drawables that are collected into \a v. */
      NVSG_API size_t getDrawables(std::vector<const Drawable *> &v //!< Receive requested Drawables here.
                                  , const StateSet *pStateSet = NULL //!< If not NULL, only drawables with the specified StateSet will be copied to \a v.
                                  ) const;

      //! Get the number of Drawables in this GeoNode.
      /** \note The number of Drawables is equal to the number of Geometries.
        * \returns number of Drawables. */
      NVSG_API size_t getNumberOfDrawables( void ) const;


      //! Replace all occurrences of a specified Drawable.
      /** The function replaces all occurrences of \a oldDrawable by \a newDrawable. 
        * \note The behavior is undefined when passing an invalid pointer for \a newDrawable. */
      NVSG_API void replaceDrawable( const Drawable* newDrawable //!< Specifies the new Drawable.
                                   , const Drawable* oldDrawable //!< Specifies the Drawable to be replaced by \a newDrawable.
                                   );

      //! Returns the StateSet at position \a index.
      /** \returns pointer to const StateSet at position \a index. */
      NVSG_API const StateSet * getStateSet( size_t index    //!<  Zero-based index of StateSet to get
                                           ) const;

      //! Returns the StateSet of a specified Drawable.
      /** \returns A pointer to the const StateSet corresponding to the Drawable \a pDrawable, or NULL, if \a pDrawable is
        * not part of this GeoNode. */
      NVSG_API const StateSet * getStateSet( const Drawable *pDrawable   //!<  Drawable to get associated StateSet from.
                                           ) const;

      //! Get all StateSets used.
      /** Fills the vector \a v with the constant pointers to all (different) StateSets.
        * \returns number of (different) StateSets that are collected into \a v. */
      NVSG_API size_t getStateSets( std::vector<const StateSet *> &v //!< Receive requested StateSets here 
                                  ) const;

      //! Get the number of (different) StateSets in this GeoNode.
      /** \returns number of (different) StateSets. */
      NVSG_API size_t getNumberOfStateSets( void ) const;

      //! Replace all occurrences of a specified StateSet.
      /** The function replaces all occurrences of \a oldStateSet by \a newStateSet. */
      NVSG_API void replaceStateSet( const StateSet * newStateSet //!< Specifies the new StateSet.
                                   , const StateSet * oldStateSet //!< Specifies the StateSet to be replaced by \a newStateSet.
                                   );

      //! Replace the StateSet for a specified Drawable.
      /** The function replaces the StateSet currently assigned to \s drawable by \a newStateSet. */
      NVSG_API void replaceStateSet( const StateSet * newStateSet //!< Specifies the new StateSet.
                                   , const Drawable * drawable    //!< Specifies the Drawable whose current StateSet is to be replaced by \a newStateSet.
                                   );

      //! Remove one Geometry from this GeoNode.
      /** Removes the Geometry with Drawable \a pDrawable from the GeoNode. 
        * If \a pStateSet is not NULL, the Geometry is only removed if it corresponds to \a pDrawable. 
        * \returns true, if a Geometry is removed, otherwise false. */
      NVSG_API bool removeGeometry( const Drawable *pDrawable        //!< Specifies the Drawable of the geometry to be removed.
                                  , const StateSet *pStateSet = NULL //!< Conditional StateSet. Remove the geometry only if \a pDrawable and \a pStateSet are associated.
                                  );

      //! Remove the Geometry at position \a index.
      /** \note The behaviour is undefined if called with an invalid index. */
      NVSG_API void removeGeometry( size_t index //!< Zero-based index of the geometry to be removed.
                                  );

      //! Remove the Geometries with a given StateSet.
      /** If \a pStateSet is given, all Geometries that use that StateSet are removed. Otherwise all Geometries are
        * removed (i.e. the GeoNode is cleared).
        * \returns true if at least one Geometry was removed, otherwise false. */
      NVSG_API bool removeGeometries( const StateSet *pStateSet = NULL //!< Conditional StateSet. If not NULL, only geometries associated with \a pStateSet will be removed from the GeoNode. 
                                    );

      //! Generate texture coordinates on all geometries.
      /** There are three different methods for texture generation: TCT_CYLINDRICAL, TCT_PLANAR, and TCT_SPHERICAL. */
      NVSG_API void generateTexCoords( TextureCoordType tct   //!< texture coordinate type to create
                                     , size_t tu = 0          //!< texture unit to create coords for
                                     );

      //! Generate tangents and binormals for all geometries.
      /** To generate tangents and binormals there have to be texture coordinates.  */
      NVSG_API void calculateTangentSpace( size_t tu = 0    //!<  texture unit that holds the texture coordinates to use for generation
                                         , size_t tg = 1    //!<  texture unit to create tangents in
                                         , size_t bn = 2    //!<  texture unit to create binormals in
                                         );

      //! Ask if this GeoNode contains any CgFx attributes.
      /** \returns  \c true if the GeoNode contains Geometry with a CgFx attribute, otherwise \c false.  */
      NVSG_API virtual bool containsCgFx( void ) const;

      //! Invalidate the CgFx containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateCgFxContainment( void );

      //! Ask if this GeoNode contains any transparent Material attributes.
      /** \returns  \c true if the GeoNode contains Geometry with a transparent Material attribute, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

      //! Invalidate the transparent Material containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateTransparentMaterialContainment( void );

      //! Ask if this GeoNode contains any transparent Texture.
      /** \returns  \c true if the GeoNode contains Geometry with a transparent Texture, otherwise \c false.  */
      NVSG_API virtual bool containsTransparentTexture( void ) const;

      //! Invalidate the transparent Texture containment cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateTransparentTextureContainment( void );

      //! Invalidate the number of animation frames cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateNumberOfFrames( void ) const;

      //--  Functions reimplemented from Node --
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //--  Functions reimplemented from Object  --
      NVSG_API virtual DataID getDataID( void ) const;
      NVSG_API virtual bool isDataShared( void ) const;

    protected:
      //! Default Constructor
      NVSG_API GeoNode( void );

      //! Partial Constructor
      NVSG_API GeoNode( const Node &rhs );

      //! Copy Constructor
      NVSG_API GeoNode( const GeoNode& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~GeoNode(void);

      //! Calculate the bounding sphere of this GeoNode.
      /** The bounding sphere of a GeoNode is calculated by combining the bounding spheres of all its geometries.
        * \returns \c true if the calculated bounding sphere is valid, otherwise \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const;

      //! Invalidate all caches of this GeoNode, maybe depending on the added or removed StateSet.
      /** Invalidates the bounding sphere of this GeoNode. If the CgFx containment changes by adding/removing \a
        * pStateSet, the CgFx cache is invalidated. */
      NVSG_API void invalidateCaches( const StateSet *pStateSet  //!<  StateSet that has been added or removed
                                    , bool insert                //!< Assigns whether \a pSateSet has been added or removed
                                    );

    private:
      void        _removeGeometry( std::vector<std::pair<const Drawable*,const StateSet*> >::iterator it );

    private:
      // collection of Drawables and their corresponding StateSet
      std::vector<std::pair<const Drawable*, const StateSet*> > m_geometry;

      //  State caches
      mutable bool    m_containsCgFx;
      mutable bool    m_containsCgFxValid;
      mutable bool    m_containsTransparentMaterial;
      mutable bool    m_containsTransparentMaterialValid;
      mutable bool    m_containsTransparentTexture;
      mutable bool    m_containsTransparentTextureValid;
      mutable size_t  m_numberOfFrames;
      mutable bool    m_numberOfFramesValid;
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // inline function implementations can go here
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  inline size_t GeoNode::getNumberOfDrawables( void ) const
  {
    __TRACE();
    return( m_geometry.size() );
  }

  inline size_t GeoNode::getNumberOfGeometries( void ) const
  {
    __TRACE(); 
    return( m_geometry.size() );
  }

  inline size_t GeoNode::getNumberOfStateSets( void ) const
  {
    __TRACE();
    std::vector<const StateSet*> v;
    return( getStateSets( v ) );
  }

  inline const Drawable * GeoNode::getDrawable( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_geometry.size()); // undefined behavior for invalid index
    return m_geometry[index].first;
  }

  inline const StateSet * GeoNode::getStateSet( size_t index ) const
  {
    __TRACE();
    __ASSERT(index<m_geometry.size()); // undefined behaviour for invalid index
    return m_geometry[index].second;
  }

} //  namespace nvsg

#ifdef _DEFINED_DBGNEW
# undef new
#endif

