// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include <map>
#include "nvmath/BoundingSphere.h"
#include "nvsg/Drawable.h"
#include "nvsg/HighlightObject.h"

#ifdef _DEBUG
# ifndef new
#  define new new(__FILE__, __LINE__)
#  define _DEFINED_DBGNEW // restrict the 'new' macro to this header file only
# endif
#endif


namespace nvsg
{
  //  forward declarations
  class GeoSet;

  //! Type to specify the type of texture coordinate to generate.
  /** There are three different types of texture coordinate generation available: cylindrical, planar, and spherical.*/
  typedef enum
  {
    TCT_CYLINDRICAL,      //!<  Puts a cylinder around the \c BoundingSphere with the z-axis as it's axis and projects the vertices of the \c GeoSet on it to get the texture coordinates.
    TCT_PLANAR,           //!<  Puts a plane in the x-y-plane and vertically projects the vertices of the \c GeoSet on that plane to the the texture coordinates.
    TCT_SPHERICAL,        //!<  Projects the vertices of the \c GeoSet radially on the \c BoundingSphere to get the texture coordinates.
  } TextureCoordType;

// VertexAttrib serves as a private helper class providing data management (so no need for docs)
# if ! defined( DOXYGEN_IGNORE )
  class VertexAttrib
  {
    public:
      // default constructs a VertexAttrib object
      VertexAttrib();
      // cleanup on destruction
      ~VertexAttrib();
      // Copies float data pointed to by fdata to the vertex attribute's internal data storage.
      // fdim specifies the dimension, in terms of float, of a single data element.
      // count specifies the number of data elements to be copied, so that always
      // count * fdim floats are copied to the internal data storage.
      // Behavior is undefined if fdata is an invalid pointer.
      void setData(unsigned int fdim, const float * fdata, size_t count);
      // Overrides the data element at position pos with the data element pointed to by fdata.
      // Behavior is undefined if either the specified position is invalid, or if fdata is an invalid pointer.
      // Behavior also is undefined if fdim doesn't match the dimension of the data element currently stored
      // at the specified position.
      void setData(unsigned int fdim, size_t pos, const float * fdata, size_t count);
      // Returns a pointer to the float data stored in this vertex attrib, 
      // or a NULL pointer if no data is currently strored
      const float * getData() const;
      // Returns the dimension, in terms of float, of the stored data.
      // If no data is stored in this vertex attrib, the return value might be random. 
      unsigned int getDataDim() const;
      // Returns the number of data elements currently stored in this vertex attribute
      size_t getNumberOfData() const;

    private:
      unsigned int m_fdim; // dimension of data elements in terms of float
      std::vector<float> m_fdata; // float data storage (contiguous memory)
      size_t m_count; // number of data elements (==m_fdata.size()/m_fdim)
                      // serves as a cache to improve performance when querying the current data count
  };

  inline const float * VertexAttrib::getData() const
  {
    __TRACE();
    return !m_fdata.empty() ? &m_fdata[0] : NULL; 
  }

  inline unsigned int VertexAttrib::getDataDim() const
  {
    __TRACE();
    __ASSERT(m_fdata.empty() || !(m_fdata.size()%m_fdim)); // this would be a serious error! 
    __ASSERT(m_fdata.empty() == !m_fdim);
    return m_fdim;
  }

  inline size_t VertexAttrib::getNumberOfData() const
  {
    __TRACE();
    __ASSERT(m_fdata.empty() == !m_fdim);
    __ASSERT(m_fdata.empty() || !(m_fdata.size()%m_fdim)); // this would be a serious error
    __ASSERT(m_fdata.empty() || m_count==m_fdata.size()/m_fdim); // count cache is not up to data
                                                               // --> serious, as well
    return m_count;
  }
#endif    //  DOXYGEN_IGNORE

  //! Base class for Geometries.
  /** This class only holds the vertex and attribute data. 
    * Any information about how they are connected has to be
    * provided by the derived topology classes (such as Triangles, etc ...). */
  class GeoSet : public Drawable
  {
    friend class CachedGeometry;

    public:
      //! Vertex Attribute Register Numbers
      /** These represent valid indices identifying a certain vertex attribute. */
      enum 
      {
        VERTEX_POSITION  = 0,
        VERTEX_WEIGHTS,
        NORMAL,
        PRIMARY_COLOR,
        SECONDARY_COLOR,
        FOG_COORDINATE,
        UNUSED_1,
        UNUSED_2,
        TEXTURE_COORD_0,
        TEXTURE_COORD_1,
        TEXTURE_COORD_2,
        TEXTURE_COORD_3,
        TEXTURE_COORD_4,
        TEXTURE_COORD_5,
        TEXTURE_COORD_6,
        TEXTURE_COORD_7,
        NUMBER_OF_VERTEX_ATTRIBUTES
      };

      //! Determine whether the data of this object is shared.
      /** \returns  \c true if the object's data is shared, \c false otherwise. */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  The DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Ask if the bounding sphere of this GeoSet is valid.
      /** If the bounding sphere isn't valid, it is tried to validate it.
        * \returns The valid state of the bounding sphere. */
      NVSG_API bool isBoundingSphereValid( void ) const;

      //! Get the bounding sphere of this GeoSet.
      /** The bounding sphere of a Node is a sphere that completely contains the Node and all its children.
        * \note It is assumed that the bounding sphere of this Node is valid.
        * \returns A constant reference to the bounding sphere. */
      NVSG_API const nvmath::Sphere3f & getBoundingSphere( void ) const;

      //! Get amount of data bytes currently stored in the vertex attribute specified by \a attrib.
      /** \returns The number of bytes used. 
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15. */
      NVSG_API size_t getNumberOfVertexAttributeBytes( size_t attrib    //!< Zero based index identifying the target vertex attribute. 
                                                     ) const;

      //! Get float data stored in the vertex attribute specified by \a attrib.  
      /** \returns A valid pointer to the vertex attribute's data if data was stored in the
        * vertex attribute, a NULL pointer otherwise.
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15. */
      NVSG_API const float * getVertexAttributeData( size_t attrib    //!< Zero based index identifying the target vertex attribute.
                                                   ) const;

      //! Determine the dimension, in terms of float, of the data stored in the vertex attribute specified by \a attrib.
      /** The native data format for data stored in a vertex attribute is float, thus the reason for expressing
        * a data element's dimension in terms of float. A single vertex position, for example, is of dimension 3, which is
        * one single-precision float value for representing a direction in a 3-dimensional space.
        * \returns The dimension, in terms of float, of the data elements stored in the specified vertex attribute.
        * If there is no data currently stored in that vertex attribute, the function returns 0.
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15. */
      NVSG_API unsigned int getVertexAttributeDataDim( size_t attrib    //!< Identifies the target vertex attribute.
                                                     ) const;

      //! Determine if data is available for a vertex attribute specified by \a attrib.
      /** \returns The number of stored data elements.
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15. */
      NVSG_API size_t getNumberOfVertexAttributeData( size_t attrib   //!< Identifies the target vertex attribute.
                                                    ) const;

      //! Set data for the vertex attribute specified by \a attrib.
      /** The function copies \a fdim * \a count float values from the location pointed to by \a fdata
        * to the vertex attribute specified by \a attrib. Pre-existing data of the specified vertex
        * attribute will be entirely erased.
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15.\n
        * The behavior is also undefined if either \a fdim doesn't match the correct dimension, in
        * terms of float, of the data pointed to by \a fdata, or if \a fdata is an invalid pointer,
        * or if count exceeds the actual number of source data elements. */
      NVSG_API void setVertexAttributeData( size_t attrib         //!< Identifies the target vertex attribute.
                                          , unsigned int fdim     //!< Dimension, in terms of float, of the source data elements.
                                          , const float * fdata   //!< Address of source data.
                                          , size_t count          //!< Number of source data elements.
                                          );

      //! Set data for the vertex attribute specified by \a attrib.
      /** The function copies \a fdim * \a count float values from the location pointed to by \a fdata
        * to the vertex attribute specified by \a attrib, starting at position \a pos of the pre-existing
        * vertex data. Pre-existing vertex data in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing data outside this range remains untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of elements currently stored for
        * the vertex attribute specified by \a attrib, the data pointed to by \a fdata will be appended to
        * the pre-existing vertex data. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of elements currently
        * stored for the specified vertex attribute nor does it specify a valid position inside the range
        * of pre-existing data, the behaviour is undefined!
        * 
        * \note The behavior is undefined when passing an invalid index for \a attrib. 
        * Valid indices are:\n
        * GeoSet::VERTEX_POSITION\n
        * GeoSet::VERTEX_WEIGHTS\n
        * GeoSet::NORMAL\n
        * GeoSet::PRIMARY_COLOR\n
        * GeoSet::SECONDARY_COLOR\n
        * GeoSet::FOG_COORDINATE\n
        * GeoSet::UNUSED_1\n
        * GeoSet::UNUSED_2\n
        * GeoSet::TEXTURE_COORD_0\n
        * GeoSet::TEXTURE_COORD_1\n
        * GeoSet::TEXTURE_COORD_2\n
        * GeoSet::TEXTURE_COORD_3\n
        * GeoSet::TEXTURE_COORD_4\n
        * GeoSet::TEXTURE_COORD_5\n
        * GeoSet::TEXTURE_COORD_6\n
        * GeoSet::TEXTURE_COORD_7\n
        * The indices listed here are a contiguous set of indices and range from 0 to 15.\n
        * The behavior is also undefined if either \a fdim doesn't match the correct dimension, in
        * terms of float, of the data element currently stored for the specified vertex attribute, or
        * if \a fdata is an invalid pointer. */
      NVSG_API void setVertexAttributeData( size_t attrib         //!< Identifies the target vertex attribute.
                                          , unsigned int fdim     //!< Dimension, in terms of float, of the source data element.
                                          , size_t pos            //!< Position in contigious memory of the vertex attribute's data set.
                                          , const float * fdata   //!< Address of source data.
                                          , size_t count          //!< Number of source data elements.
                                          );
      //! Get number of vertices.
      /** \returns The number of vertices. 
        * \note Calling this function has the same effect as calling getNumberOfVertexAttributeData(VERTEX_POSITION).*/
      NVSG_API size_t getNumberOfVertices( void ) const;

      //! Get vertices.
      /** \returns A pointer to the vertices. 
        * \note Calling this function has the same effect as calling getVertexAttributeData(VERTEX_POSITION).
        * The only difference is, that this function returns a const Vec3f* rather than a const float* to reflect a
        * vertex position's dimension. One must keep this in mind when it comes to pointer arithmetics! */
      NVSG_API const nvmath::Vec3f * getVertices( void ) const;

      //! Set vertices for this GeoSet.
      /** Copies \a count vertices from the location pointed to by \a verts into this GeoSet.
        * Pre-existing vertices will be entirely erased.
        * \note The behaviour is undefined if either \a verts is invalid, or if \a count exceeds the actual 
        * number of vertices stored at \a verts. */
      NVSG_API void setVertices( const nvmath::Vec3f * verts                    //!<  vertices to set
                               , size_t count                           //!<  number of vertices
                               );

      //! Set vertices for this GeoSet.
      /** Copies \a count vertices from the location pointed to by \a verts into this GeoSet, 
        * starting at position \a pos inside the range of pre-existing vertices. 
        * Pre-existing vertices in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing vertices outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of vertices currently stored
        * the vertices pointed to by \a verts will be appended to the pre-existing vertices. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of vertices currently
        * stored nor does it specify a valid position inside the range of pre-existing vertices, 
        * the behaviour is undefined!
        *
        * \note The behaviour is undefined if either \a verts is invalid, or if \a count exceeds 
        * the actual number of vertices stored at \a verts. */
      NVSG_API void setVertices( size_t pos                             //!<  Start position inside the range of pre-existing vertices.
                               , const nvmath::Vec3f * verts            //!<  Start address of vertices to set.
                               , size_t count                           //!<  Number of vertices to copy.
                               );
       
      //! Get number of normals.
      /** \returns The number of normals. 
        * \note Calling this function has the same effect as calling getNumberOfVertexAttributeData(NORMAL).*/
      NVSG_API size_t getNumberOfNormals( void ) const;

      //! Get normals.
      /** \returns A pointer to the normals. 
        * \note Calling this function has the same effect as calling getVertexAttributeData(NORMAL).
        * The only difference is, that this function returns a const Vec3f* rather than a const float* to reflect a
        * normal's dimension. One must keep this in mind when it comes to pointer arithmetics! */
      NVSG_API const nvmath::Vec3f * getNormals( void ) const;

      //! Set normals for this GeoSet.
      /** Copies \a count normals from the location pointed to by \a nor into this GeoSet. 
        * Pre-existing normals will be entirely erased.
        *
        * \note The behaviour is undefined if either \a nor is invalid, or if \a count exceeds 
        * the actual number of normals stored at \a nor. */
      NVSG_API void setNormals( const nvmath::Vec3f* nor    //!< Address of source normals.
                              , size_t count        //!< Number of normals to set.
                              );

      //! Set normals for this GeoSet.
      /** Copies \a count normals from the location pointed to by \a nor into this GeoSet, 
        * starting at position \a pos inside the range of pre-existing normals. 
        * Pre-existing normals in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing normals outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of normals currently stored,
        * the normals pointed to by \a nor will be appended to the pre-existing normals. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of normals currently
        * stored nor does it specify a valid position inside the range of pre-existing normals, 
        * the behaviour is undefined!
        *
        * \note The behaviour is undefined if either \a nor is invalid, or if \a count exceeds 
        * the actual number of normals stored at \a nor. */
      NVSG_API void setNormals( size_t pos                 //!<  Start position inside the range of pre-existing normals.
                              , const nvmath::Vec3f * nor  //!<  Start address of normals to set.
                              , size_t count               //!<  Number of normals to copy.
                              );

      //! Determine if at least one texture coordinate array has valid texture coordinates.
      /** \returns  true, if this GeoSet has texture coordinates, false otherwise. */
      NVSG_API bool hasTexCoords( void ) const;

      //! Get number of texture coordinates in a specified texture unit.
      /** \returns  The number of texture coordinates in texture unit \a texUnit. 
        * \note Calling the function has the same effect as calling 
        * getNumberOfVertexAttributeData(TEXTURE_COORD_0 + \a texUnit). 
        * \note The behavior is undefined if \a texUnit specifies an invalid index. A valid index ranges from 0 to 7.*/
      NVSG_API size_t getNumberOfTexCoords( size_t texUnit    //!< The zero based index of the texture unit.
                                          ) const;

      //! Get texture coordinates for a specified texture unit.
      /** \returns A pointer to the texture coords.
        * \note Texture coordinates can be either 1, 2, 3, or 4-dimensional in terms of float. 
        * Call either getTexCoordDim(\a texUnit) or getVertexAttributeDataDim(TEXTURE_COORD_0 + \a texUnit) 
        * to determine the dimension of the texture coordinates stored in texture unit \a texUnit.
        * \note The behavior is undefined if \a texUnit specifies an invalid index. A valid index ranges from 0 to 7.*/
      NVSG_API const float * getTexCoords( size_t texUnit   //!< The zero based index of the texture unit. 
                                         ) const;

      //! Get the dimension of the texture coordinates stored in the specified texture unit.
      /** The dimension of texture coordinates is meant in terms of floats. Texture coordinates
        * can be either 1, 2, 3, or 4-dimensional.
        * \returns  The dimension of the texture coordinates stored in texture unit \a texUnit. If there are no
        * texture coordinates stored in that texture unit, the function returns 0.
        * \note The behavior is undefined if \a texUnit specifies an invalid index. A valid index ranges from 0 to 7.*/
      NVSG_API unsigned int getTexCoordsDim( size_t texUnit   //!< The zero based index of the texture unit. 
                                           ) const;

      //! Set the texture coordinates for a specified texture unit.
      /** Copies \a count texture coordinates of dimension \a dim from the location pointed to by \a coords 
        * into the texture unit \a texUnit of this GeoSet.\n
        * Valid values for \a fdim are either 1, 2, 3, or 4. A valid index for \a texUnit ranges from 0 to 7.
        * \note The behavior is undefined if either \a texUnit specifies an invalid index, or \a fdim doesn't match 
        * the correct dimension, in terms of float, of the data pointed to by \a coords, or \a coords is an invalid
        * pointer, or count exceeds the actual number of source coordinates. */
      NVSG_API void setTexCoords( size_t texUnit          //!< The zero based index of the texture unit. 
                                , unsigned int fdim       //!< Dimension of coordinates, in terms of float
                                , const float * coords    //!< Address of source coordinates
                                , size_t count            //!< Number of coordinates to copy
                                );

      //! Set the texture coordinates for a specified texture unit.
      /** Copies \a count texture coordinates of dimension \a dim from the location pointed to by \a coords 
        * into the texture unit \a texUnit of this GeoSet, starting at position \a pos inside the range of
        * pre-existing texture coordinates. Valid values for \a fdim are either 1, 2, 3, or 4. 
        * A valid index for \a texUnit ranges from 0 to 7.
        *
        * Pre-existing coordinates in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing coordinates outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of coordinates currently stored
        * the coordinates pointed to by \a coords will be appended to the pre-existing coordinates. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of coordinates currently
        * stored nor does it specify a valid position inside the range of pre-existing coordinates, 
        * the behaviour is undefined!
        *
        * \note The behavior is undefined if either \a texUnit specifies an invalid index, or \a fdim doesn't match 
        * the correct dimension, in terms of float, of the data pointed to by \a coords, or \a coords is an invalid
        * pointer, or count exceeds the actual number of source coordinates. */
      NVSG_API void setTexCoords( size_t texUnit          //!< The zero based index of the texture unit. 
                                , unsigned int fdim       //!< Dimension of coordinates, in terms of float
                                , size_t pos              //!< Start position inside the range of pre-existing texture coordinates.
                                , const float * coords    //!< Address of source coordinates
                                , size_t count            //!< Number of coordinates to copy
                                );

      //! Generate texture coordinates into a texture unit.
      /** Calculates either spherical, cylindrical, or planar 
        * two-dimensional texture coordinates into texture unit \a
        * tu. The \a sphere can be either the bounding sphere of 
        * this GeoSet to calculate local texture coordinates or
        * some other for non-local texture coordinates.        */
      NVSG_API void generateTexCoords( TextureCoordType tct             //!<  type of texture coordinates to generate
                                      , const nvmath::Sphere3f &sphere  //!<  bounding sphere to map coordinates to
                                      , size_t tu = 0                   //!<  texture unit to generate the coordinates in
                                      );
      
      //! Get number of colors.
      /** \returns  The number of colors. */
      NVSG_API size_t getNumberOfColors( void ) const;

      //! Get constant pointer to colors.
      /** \note It is a mistake to call \c getColors on a \c GeoSet that has no colors available.
        * For this reason, the client code should either call \c hasColors or \c getNumberOfColors 
        * on that \c GeoSet to ensure that colors are available for that \c GeoSet. 
        * \returns  A constant pointer to The colors. */
      NVSG_API const float * getColors( void ) const;

      //! Get the dimension of the colors.
      /** The dimension of colors is conveyed in terms of floats. With this in mind, colors can
        * be either 3-dimensional - occupying 3 floating point values per color, or 4-dimensional -  
        * occupying 4 floating point values per color.
        * \returns The dimension of the colors stored. If there are no colors stored, the function return 0. */
      NVSG_API unsigned int getColorsDim( void ) const;

      //! Set colors for this GeoSet.
      /** Copies \a count colors of dimension \a fdim from the location pointed to by \a cols to this GeoSet.\n 
        * Valid values for \a fdim are 3 and 4.
        * \note The behavior is undefined if either \a fdim doesn't match the correct dimension, in terms of float, 
        * of the colors pointed to by \a cols, or \a cols is an invalid pointer, or count exceeds the actual number 
        * of source colors. */
      NVSG_API void setColors( unsigned int fdim    //!< Dimension of colors, in terms of float
                             , const float * cols   //!< Address of source colors
                             , size_t count         //!< Number of colors to copy
                             );

      //! Set colors for this GeoSet.
      /** Copies \a count colors of dimension \a fdim from the location pointed to by \a cols to this GeoSet,
        * starting at position \a pos inside the range of pre-existing colors.
        * Valid values for \a fdim are 3 and 4.
        *
        * Pre-existing colors in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing colors outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of colors currently stored
        * the colors pointed to by \a cols will be appended to the pre-existing colors. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of colors currently
        * stored nor does it specify a valid position inside the range of pre-existing colors, 
        * the behaviour is undefined!
        *
        * \note The behavior is undefined if either \a fdim doesn't match the correct dimension, in terms of float, 
        * of the colors pointed to by \a cols, or \a cols is an invalid pointer, or count exceeds the actual number 
        * of source colors. */
      NVSG_API void setColors( unsigned int fdim    //!< Dimension of colors, in terms of float
                             , size_t pos           //!< Start position inside the range of pre-existing colors.
                             , const float * cols   //!< Address of source colors
                             , size_t count         //!< Number of colors to copy
                             );

      //! Get number of secondary colors.
      /** \returns  The number of secondary colors. */
      NVSG_API size_t getNumberOfSecondaryColors( void ) const;

      //! Get constant pointer to secondary colors.
      /** \returns A pointer to the secondary colors. */
      NVSG_API const float * getSecondaryColors( void ) const;

      //! Get the dimension of the secondary colors.
      /** \returns The dimension of the secondary colors stored. If there are no 
        * secondary colors stored, the function returns 0. */
      NVSG_API unsigned int getSecondaryColorsDim( void ) const;

      //! Set secondary colors for this GeoSet.
      /** Copies \a count colors of dimension \a fdim from the location pointed to by \a cols to this GeoSet.\n 
        * Valid values for \a fdim are 3 and 4.
        * \note The behavior is undefined if either \a fdim doesn't match the correct dimension, in terms of float, 
        * of the colors pointed to by \a cols, or if \a cols is an invalid pointer, or if count exceeds the actual number 
        * of source colors. */
      NVSG_API void setSecondaryColors( unsigned int fdim     //!< Dimension of colors, in terms of float.
                                      , const float * cols    //!< Address of source colors.
                                      , size_t count          //!< Number of colors to copy.
                                      );

      //! Set secondary colors for this GeoSet.
      /** Copies \a count colors of dimension \a fdim from the location pointed to by \a cols to this GeoSet,
        * starting at position \a pos inside the range of pre-existing colors.
        * Valid values for \a fdim are 3 and 4.
        *
        * Pre-existing colors in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing colors outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of colors currently stored
        * the colors pointed to by \a cols will be appended to the pre-existing colors. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of colors currently
        * stored nor does it specify a valid position inside the range of pre-existing colors, 
        * the behaviour is undefined!
        *
        * \note The behavior is undefined if either \a fdim doesn't match the correct dimension, in terms of float, 
        * of the colors pointed to by \a cols, or \a cols is an invalid pointer, or count exceeds the actual number 
        * of source colors. */
      NVSG_API void setSecondaryColors( unsigned int fdim    //!< Dimension of colors, in terms of float
                                      , size_t pos           //!< Start position inside the range of pre-existing colors.
                                      , const float * cols   //!< Address of source colors
                                      , size_t count         //!< Number of colors to copy
                                      );

      //! Get number of fog coordinates.
      /** \returns  The number of fog coordinates. */
      NVSG_API size_t getNumberOfFogCoords( void ) const;

      //! Get fog coordinates.
      /** \returns A pointer to the fog coordiantes. */
      NVSG_API const float * getFogCoords( void ) const;

      //! Set the fog coordinates.
      /** Copies \a count fog coordinates from the location pointed to by \a coords to this GeoSet. 
        * \note The behaviour is undefined if either \a coords is an invalid pointer, or count exceeds the actual number
        * of source coordinates. */
      NVSG_API void setFogCoords( const float * coords    //!< Address of source coordinates. 
                                , size_t count            //!< Number of coordinates to copy.
                                );

      //! Set the fog coordinates.
      /** Copies \a count fog coordinates from the location pointed to by \a coords to this GeoSet,
        * starting at position \a pos inside the range of pre-existing coordinates. 
        *
        * Pre-existing coordinates in the range [\a pos, \a pos + \a count) will be replaced.
        * Pre-existing coordinates outside this range remain untouched.
        *
        * If you specify -1 for \a pos or if \a pos specifies the number of coordinates currently stored
        * the coordinates pointed to by \a coords will be appended to the pre-existing coordinates. 
        *
        * If you not specify -1 for \a pos or \a pos neither specifies the number of coordinates currently
        * stored nor does it specify a valid position inside the range of pre-existing coordinates, 
        * the behaviour is undefined!
        *
        * \note The behaviour is undefined if either \a coords is an invalid pointer, or count exceeds the actual number
        * of source coordinates. */
      NVSG_API void setFogCoords( size_t pos              //!< Start position inside the range of pre-existing coordinates
                                , const float * coords    //!< Address of source coordinates. 
                                , size_t count            //!< Number of coordinates to copy.
                                );

      //! Interface for calculating tangents and binormals.
      /** The texture coordinates out of texture unit \a tu are used as the parameterization for the calculation. 
        * The tangents are written to texture unit \a tg and the binormals to texture unit \a bn. */
      NVSG_API virtual void  calculateTangentSpace( size_t tu = 0   //!<  texture unit to get parameterizing coords
                                                  , size_t tg = 1   //!<  texture unit to write tangents to
                                                  , size_t bn = 2   //!<  texture unit to write binormals to
                                                  ) = 0;

      //! Will attach \a pObj as Highlight object for visualization.
      /** Pass NULL to remove any attached object. */
      NVSG_API void attachHighlightObject( const HighlightObject* pObj = NULL ) const;

      //! Get the currently attached \a pObj Highlight object for visualization.
      /** \returns A pointer to the highlight object. */
      NVSG_API const HighlightObject* getHighlightObject( void ) const;

    protected:
      //! Instantiate a default constructed GeoSet object.
      /** \note A GeoSet naturally will not be instantiated directly, but through
        * instantiating a derived topology object like Triangles or TriStrips for
        * example. */
      NVSG_API GeoSet( void );

      //! Instantiate a copy of the GeoSet refered to by \a rhs.
      /** \note A GeoSet naturally will not be instantiated directly, but through
        * instantiating a derived topology object like Triangles or TriStrips for
        * example. */
      NVSG_API GeoSet( const GeoSet& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~GeoSet( void );

      //! Vertex attribute data have been changed.
      /** This function gets called from the framework, whenever certain vertex attribute data has been changed. 
        * Classes that derive from GeoSet may overload this function to add some necessary work after the vertex
        * attribute data has been changed. */ 
      NVSG_API virtual void onSetVertexAttributeData( void );

      //! Invalidate the BoundingSphere (of the parent geo node also).
      NVSG_API virtual void invalidateBoundingSphere( void ) const;

    private:      
        // implementation details
      void generateTexCoordsCylinder( const nvmath::Sphere3f &sphere, size_t tu = 0 );
      void generateTexCoordsPlane( const nvmath::Sphere3f &sphere, size_t tu = 0 );
      void generateTexCoordsSphere( const nvmath::Sphere3f &sphere, size_t tu = 0 );
      void _validateBoundingSphere( void ) const;

      nvutil::RCPtr<nvutil::RCVector<VertexAttrib> >            m_vertexAttribs;    // shareable vertex attribs
      mutable nvmath::BoundingSphere                    m_boundingSphere;   // geometry's bounds
      mutable const HighlightObject *                  m_highlightObject;       
  };

  inline const float * GeoSet::getVertexAttributeData( size_t attrib ) const
  {
    __TRACE();
    __ASSERT(attrib < NUMBER_OF_VERTEX_ATTRIBUTES );
    return (*m_vertexAttribs)[attrib].getData();
  }

  inline size_t GeoSet::getNumberOfVertexAttributeData( size_t attrib ) const
  {
    __TRACE();
    __ASSERT(attrib < NUMBER_OF_VERTEX_ATTRIBUTES);
    return (*m_vertexAttribs)[attrib].getNumberOfData();
  }

  inline unsigned int GeoSet::getVertexAttributeDataDim( size_t attrib ) const
  {
    __TRACE();
    __ASSERT(attrib < NUMBER_OF_VERTEX_ATTRIBUTES );
    return (*m_vertexAttribs)[attrib].getDataDim();
  }

  inline size_t GeoSet::getNumberOfVertices( void ) const 
  {
    __TRACE();
    return getNumberOfVertexAttributeData(VERTEX_POSITION);
  }

  inline const nvmath::Vec3f * GeoSet::getVertices( void )  const
  {
    __TRACE();
    return (const nvmath::Vec3f *)getVertexAttributeData(VERTEX_POSITION);
  }

  inline void GeoSet::setVertices( const nvmath::Vec3f * verts, size_t count )
  {
    __TRACE();
    setVertexAttributeData(VERTEX_POSITION, 3, (const float*)verts, count);

    if(m_highlightObject)
    {
      // the highlight object needs to be recalculated
      HighlightObject * hlobj = beginEdit( m_highlightObject );
      hlobj->invalidate();
      endEdit( hlobj );
    }
  }

  inline void GeoSet::setVertices( size_t pos, const nvmath::Vec3f * verts, size_t count )
  {
    __TRACE();
    setVertexAttributeData(VERTEX_POSITION, 3, pos, (const float*)verts, count);

    if(m_highlightObject)
    {
      // the highlight object needs to be recalculated
      HighlightObject * hlobj = beginEdit( m_highlightObject );
      hlobj->invalidate();
      endEdit( hlobj );
    }
  }

  inline size_t GeoSet::getNumberOfNormals( void ) const 
  {
    __TRACE();
    return getNumberOfVertexAttributeData(NORMAL);
  }

  inline const nvmath::Vec3f * GeoSet::getNormals( void ) const 
  {
    __TRACE();
    return (const nvmath::Vec3f *)getVertexAttributeData(NORMAL);
  }

  inline void GeoSet::setNormals( const nvmath::Vec3f * nor, size_t count )
  {
    __TRACE();
    setVertexAttributeData(NORMAL, 3, (const float*)nor, count);
                                 //^ normals always are 3-dimensional
  }

  inline void GeoSet::setNormals( size_t pos, const nvmath::Vec3f * nor, size_t count )
  {
    __TRACE();
    setVertexAttributeData(NORMAL, 3, pos, (const float*)nor, count );
  }

  inline bool GeoSet::hasTexCoords( void ) const
  {
    __TRACE();
    return(  getNumberOfVertexAttributeData(TEXTURE_COORD_0)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_1)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_2)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_3)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_4)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_5)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_6)
          || getNumberOfVertexAttributeData(TEXTURE_COORD_7) );
  }

  inline size_t GeoSet::getNumberOfTexCoords( size_t texUnit ) const
  {
    __TRACE();
    __ASSERT(texUnit <= 7);
    return getNumberOfVertexAttributeData(TEXTURE_COORD_0+texUnit);
  }

  inline const float * GeoSet::getTexCoords( size_t texUnit ) const
  {
    __TRACE();
    __ASSERT(texUnit <= 7);
    return getVertexAttributeData(TEXTURE_COORD_0+texUnit);
  }

  inline unsigned int GeoSet::getTexCoordsDim( size_t texUnit ) const
  {
    __TRACE();
    __ASSERT(texUnit <= 7);
    return getVertexAttributeDataDim(TEXTURE_COORD_0+texUnit);
  }

  inline void GeoSet::setTexCoords( size_t texUnit, unsigned int fdim, const float* coords, size_t count )
  {
    __TRACE();
    __ASSERT(texUnit <= 7);
    setVertexAttributeData(TEXTURE_COORD_0+texUnit, fdim, coords, count);
  }

  inline void GeoSet::setTexCoords( size_t texUnit, unsigned int fdim, size_t pos, const float* coords, size_t count )
  {
    __TRACE();
    __ASSERT(texUnit <= 7);
    setVertexAttributeData(TEXTURE_COORD_0+texUnit, fdim, pos, coords, count);
  }

  inline size_t GeoSet::getNumberOfColors( void ) const
  {
    __TRACE();
    return getNumberOfVertexAttributeData(PRIMARY_COLOR);
  }

  inline const float * GeoSet::getColors( void ) const
  {
    __TRACE();
    return getVertexAttributeData(PRIMARY_COLOR);
  }

  inline unsigned int GeoSet::getColorsDim( void ) const
  {
    __TRACE();
    return getVertexAttributeDataDim(PRIMARY_COLOR);
  }

  inline void GeoSet::setColors( unsigned int fdim, const float * cols, size_t count )
  {
    __TRACE();
    setVertexAttributeData(PRIMARY_COLOR, fdim, cols, count);
  }

  inline void GeoSet::setColors( unsigned int fdim, size_t pos, const float * cols, size_t count )
  {
    __TRACE();
    setVertexAttributeData(PRIMARY_COLOR, fdim, pos, cols, count);
  }

  inline size_t GeoSet::getNumberOfSecondaryColors( void ) const
  {
    __TRACE();
    return getNumberOfVertexAttributeData(SECONDARY_COLOR);
  }

  inline const float * GeoSet::getSecondaryColors( void ) const
  {
    __TRACE();
    return getVertexAttributeData(SECONDARY_COLOR);
  }

  inline unsigned int GeoSet::getSecondaryColorsDim( void ) const
  {
    __TRACE();
    return getVertexAttributeDataDim(SECONDARY_COLOR);
  }

  inline void GeoSet::setSecondaryColors( unsigned int fdim, const float* cols, size_t count )
  {
    __TRACE();
    setVertexAttributeData(SECONDARY_COLOR, fdim, cols, count);
  }

  inline void GeoSet::setSecondaryColors( unsigned int fdim, size_t pos, const float* cols, size_t count )
  {
    __TRACE();
    setVertexAttributeData(SECONDARY_COLOR, fdim, pos, cols, count);
  }

  inline size_t GeoSet::getNumberOfFogCoords( void ) const
  {
    __TRACE();
    return getNumberOfVertexAttributeData(FOG_COORDINATE);
  }

  inline const float * GeoSet::getFogCoords( void ) const
  {
    __TRACE();
    return getVertexAttributeData(FOG_COORDINATE);
  }

  inline void GeoSet::setFogCoords( const float * coords, size_t count )
  {
    __TRACE();
    setVertexAttributeData(FOG_COORDINATE, 1, coords, count);
                                         //^ fog coords always are 1-dimensional
  }

  inline void GeoSet::setFogCoords( size_t pos, const float * coords, size_t count )
  {
    __TRACE();
    setVertexAttributeData(FOG_COORDINATE, 1, pos, coords, count);
                                         //^ fog coords always are 1-dimensional
  }

  inline const HighlightObject* GeoSet::getHighlightObject( void ) const
  {
    __TRACE();
    return m_highlightObject;
  }

} //  namespace nvsg

#ifdef _DEFINED_DBGNEW
# undef new
#endif
