// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/HighlightObject.h"
#include "nvmath/BoundingSphere.h"

namespace nvsg 
{
  //! Base class for Highlighting objects that can be attached to Geosets
  class HighlightBox : public HighlightObject
  {
    public:
      //! Create a HighlightBox.
      /** \returns a constant pointer to a HighlightBox. */
      NVSG_API static const HighlightBox * create( float width = 2.0f                                    //!< default line width
                                                 , nvmath::Vec3f color = nvmath::Vec3f(1.0f, 0.0f, 0.0f) //!< default line color
                                                 );

      //! Create a HighlightBox by copying from a HighlightObject.
      /** \returns a constant pointer to a HighlightBox. */
      NVSG_API static const HighlightBox * createFromBase( const HighlightObject &rhs                           //!< HighlightObject to copy from
                                                         , float width = 2.0f                                   //!< default line width
                                                         , nvmath::Vec3f color = nvmath::Vec3f(1.0f,0.0f,0.0f)  //!< default line color
                                                         );

      //! Create a HighlightBox by cloning this.
      /** \returns a constant pointer to a HighlightBox. */
      NVSG_API virtual const HighlightBox * clone( void ) const;

      //! Ask if the bounding sphere of this HighlightBox is valid.
      /** If the bounding sphere isn't valid, it is tried to validate it.
        * \returns The valid state of the bounding sphere. */
      NVSG_API virtual bool isBoundingSphereValid( void ) const;

      //! Get the bounding sphere of this HighlightObject.
      /** The bounding sphere of a Node is a sphere that completely contains the HighlightObject.
        * \note It is assumed that the bounding sphere of this HighlightObject is valid.
        * \returns A constant reference to the bounding sphere. */
      NVSG_API virtual const nvmath::Sphere3f & getBoundingSphere( void ) const;

      //! set the vertices (edges) of the bounding box
      NVSG_API void setVertices(nvmath::Vec3f* vertices);
      
      //! return the vertices (edges) of the bounding box
      NVSG_API const nvmath::Vec3f* getVertices() const;

      //! return the line width used for drawing the box
      NVSG_API float getLineWidth() const;

      //! return the line width used for drawing the box
      NVSG_API const nvmath::Vec3f & getColor() const;

    protected:
      //! Constructor with line width and color.
      NVSG_API HighlightBox( float width, nvmath::Vec3f color );

      //! Partial Constructor
      NVSG_API HighlightBox( const HighlightObject &rhs, float width, nvmath::Vec3f color );

      //! Copy Constructor
      NVSG_API HighlightBox( const HighlightBox &rhs );

    protected:
      nvmath::Vec3f m_vertices[8];             //!< the box's edges
      float m_lineWidth;               //!< line width used for drawing the bounding box
      nvmath::Vec3f m_color;                   //!< color used for drawing
      nvmath::BoundingSphere m_boundingSphere; //!< the box's bounding sphere

    private:
      HighlightBox( void );   //  private default constructor to prevent usage
  };

  inline const nvmath::Vec3f* HighlightBox::getVertices() const
  {
    __TRACE();
    return m_vertices;
  }

  inline float HighlightBox::getLineWidth() const 
  { 
    __TRACE();
    return m_lineWidth; 
  }       

  inline const nvmath::Vec3f & HighlightBox::getColor() const
  {
    __TRACE();
    return m_color;
  }

  inline bool HighlightBox::isBoundingSphereValid( void ) const
  {
    return( m_boundingSphere.isValid() );
  }

  inline const nvmath::Sphere3f & HighlightBox::getBoundingSphere( void ) const
  {
    __ASSERT( m_boundingSphere.isValid() );
    return( m_boundingSphere.getSphere() );
  }

}
