// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvsg/Object.h"

namespace nvmath
{
  class Sphere3f;
}

namespace nvsg
{
  //! Base class for Highlighting objects that can be attached to Geosets
  class HighlightObject : public Object
  {
    public:
      //! Ask if the bounding sphere of this HighlightObject is valid.
      /** If the bounding sphere isn't valid, it is tried to validate it.
        * \returns The valid state of the bounding sphere. */
      NVSG_API virtual bool isBoundingSphereValid( void ) const = 0;

      //! Get the bounding sphere of this HighlightObject.
      /** The bounding sphere of a Node is a sphere that completely contains the HighlightObject.
        * \note It is assumed that the bounding sphere of this HighlightObject is valid.
        * \returns A constant reference to the bounding sphere. */
      NVSG_API virtual const nvmath::Sphere3f & getBoundingSphere( void ) const = 0;

      //! Tells us whether the highlight object needs to be recalculated.
      /** Calculation is required before first rendering, and every time
          the associated geometry is changed. */
      NVSG_API virtual bool isValid( void ) const;
      
      //! Invalidate highlight geometry, so that it will be recalculated
      NVSG_API virtual void invalidate( void );

    protected:
      //! Constructor
      NVSG_API HighlightObject( void );
      
    protected:
      bool m_invalid;        //!< set if the highlight needs to be recalculated
  };

  inline HighlightObject::HighlightObject( void )
    : m_invalid(true)
  {
    __TRACE();
  }

  inline bool HighlightObject::isValid( void ) const 
  { 
    __TRACE();
    return !m_invalid; 
  }
    
  inline void HighlightObject::invalidate( void ) 
  { 
    __TRACE();
    m_invalid=true; 
  }    
}
