// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** @file */

#include "nvsgcommon.h"

#include  "nvsg/Animation.h"

namespace nvsg
{
  //! Base class for an interpolating \c Animation.
  /** An InterpolatedAnimation holds the information for a key-frame based animation. It provides the interface needed
    * for an interpolated animation, but does not perform any interpolations. Interpolations must be performed by a class that is
    * derived from this, such as \c LinearInterpolatedAnimation. */
  template<typename T> class InterpolatedAnimation : public Animation<T>
  {
    public:
      //--  Functions reimplemented from Animation  --
      virtual size_t getNumberOfFrames( void ) const;

      //--  Functions specific to InterpolatedAnimation --
      //! Get the number of key frames in the animation.
      /** \returns The number of animation key frames */
      size_t getNumberOfKeys( void ) const;

      //! Set the number of key frames in the animation.
      /** Resizes the array of key animation values. On growing this array, the new elements are not initialized. */
      void setNumberOfKeys( size_t count    //!<  number of keys
                          );

      //! Set the frame and value of key \a key.
      /** \note There is no non-constant operator[] for an InterpolatedAnimation!*/
      void setKeyFrame( size_t key            //!<  index of the key frame to set (not the frame index!)
                      , unsigned int frame    //!<  frame index to set
                      , const T & value       //!<  frame value to set
                      );

      //! Get the index of the current frame.
      /** \return The index of the current frame. */
      unsigned int getCurrentFrame( void ) const;

      //! Get the value at the current frame.
      /** \return The value of the current frame. */
      const T & getCurrentValue( void ) const;

      //! Get the index of key frame \a key.
      /** \return The frame index at the \a key th keyframe.  */
      unsigned int getFrame( size_t key ) const;

      //! Get the value on key \a key.
      /** \return The frame value at the \a key th keyframe.  */
      const T & getValue( size_t key ) const;

      //! Get the key above or equal to frame \a frame.
      /** \return The key index above or equal to the frame index \a frame. */
      size_t getKey( unsigned int frame ) const;

    protected:
      //! Default Constructor
      InterpolatedAnimation( void );

      //! Copy Constructor
      InterpolatedAnimation( const InterpolatedAnimation &rhs   //!<  InterpolatedAnimation to copy
                           );

      //! Protected destructor to prevent explicit creation on stack.
      virtual ~InterpolatedAnimation( void );

      //! Set the current frame and value on \a key.
      void  setCurrent( size_t key ) const;

      //! Set the current frame and value.
      void  setCurrent( unsigned int frame, const T &value ) const;

    private:
      std::vector<unsigned int>  m_keys;
      std::vector<T>             m_values;
      mutable unsigned int  m_currentFrame;
      mutable T             m_currentValue;
  };

  template<typename T> 
  inline InterpolatedAnimation<T>::InterpolatedAnimation( void )
  : m_currentFrame(-1)
  {
    __TRACE();
  }

  template<typename T> 
  inline InterpolatedAnimation<T>::InterpolatedAnimation( const InterpolatedAnimation &rhs )
    : m_keys( rhs.m_keys )
    , m_values( rhs.m_values )
    , m_currentFrame( rhs.m_currentFrame )
    , m_currentValue( rhs.m_currentValue )
  {
    __TRACE();
  }

  template<typename T> 
  inline InterpolatedAnimation<T>::~InterpolatedAnimation(void)
  {
    __TRACE();
  }

  template<typename T> 
  inline size_t InterpolatedAnimation<T>::getNumberOfFrames( void ) const
  {
    __TRACE();
    __ASSERT( m_keys.size() );                    //  the last key holds the index of the last frame,
    return( 1 + m_keys[m_keys.size()-1] );      //  but the number of frames is one greater
  }

  template<typename T> 
  inline size_t InterpolatedAnimation<T>::getNumberOfKeys( void ) const
  {
    __TRACE();
    return( m_keys.size() );
  }

  template<typename T> 
  inline void InterpolatedAnimation<T>::setNumberOfKeys( size_t count )
  {
    __TRACE();
    m_keys.resize( count );
    m_values.resize( count );
  }

  template<typename T> 
  inline void InterpolatedAnimation<T>::setKeyFrame( size_t key, unsigned int frame, const T &value )
  {
    __TRACE();
    __ASSERT( key < m_keys.size() );
    m_keys[key]   = frame;
    m_values[key] = value;
  }

  template<typename T> 
  inline unsigned int InterpolatedAnimation<T>::getCurrentFrame( void ) const
  {
    __TRACE();
    return( m_currentFrame );
  }

  template<typename T> 
  inline const T & InterpolatedAnimation<T>::getCurrentValue( void ) const
  {
    __TRACE();
    return( m_currentValue );
  }

  template<typename T> 
  inline unsigned int InterpolatedAnimation<T>::getFrame( size_t key ) const
  {
    __TRACE();
    return( m_keys[key] );
  }

  template<typename T> 
  inline const T & InterpolatedAnimation<T>::getValue( size_t key ) const
  {
    __TRACE();
    return( m_values[key] );
  }

  template<typename T> 
  inline size_t InterpolatedAnimation<T>::getKey( unsigned int frame ) const
  {
    __TRACE();
    std::vector<unsigned int>::const_iterator it = std::lower_bound( m_keys.begin(), m_keys.end(), frame );
    __ASSERT( it != m_keys.end() );
    return( it - m_keys.begin() );
  }

  template<typename T> 
  inline void InterpolatedAnimation<T>::setCurrent( size_t key ) const
  {
    __TRACE();
    m_currentFrame = m_keys[key];
    m_currentValue = m_values[key];
  }

  template<typename T> 
  inline void InterpolatedAnimation<T>::setCurrent( unsigned int frame, const T &value ) const
  {
    __TRACE();
    m_currentFrame = frame;
    m_currentValue = value;
  }
}
