// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvmath/BoundingSphere.h"
#include "nvsg/Object.h"

namespace nvsg
{
  class Group;
  
  //! Base class for all nodes of the scene graph.
  /** The Node is the base of all nodes of the scene graph, like GeoNode, Group, or LightSource.  */
  class Node : public Object
  {
    public:
      //! Get a clone of this \c Node object.
      /** \returns A pointer to the new created clone. */
      NVSG_API virtual const Node * clone( void ) const = 0;

      //! Determine whether the data of this object is shared.
      /** \returns \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void )  const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void )   const;

      //! Returns the number of \link nvsg::Group Group \endlink nodes that currently reference this node.
      /** \returns The actual number of Group nodes that have this node as child node. */
      NVSG_API size_t getNumberOfParents() const;

      //! Get the \a i-th parent of this Node.
      /** \note The behavior is undefined if called with an invalid index. 
        * Use \link nvsg::Node::getNumberOfParents getNumberOfParents \endlink to determine
        * the actual number of parents that have this node as child node. */
      NVSG_API const Group * getParent( size_t i       //!<  zero-based index of parent to get
                                      ) const;

      //! Ask if the bounding sphere of this Node is valid.
      /** If the bounding sphere isn't valid, it is tried to validate it.
        * \returns The valid state of the bounding sphere. */
      NVSG_API bool isBoundingSphereValid( void ) const;

      //! Get the bounding sphere of this Node.
      /** The bounding sphere of a Node is a sphere that completely contains the Node and all its children.
        * \note It is assumed that the bounding sphere of this Node is valid.
        * \returns A constant reference to the bounding sphere. */
      NVSG_API const nvmath::Sphere3f & getBoundingSphere( void ) const;

      //! Invalidate the BoundingSphere.
      NVSG_API void invalidateBoundingSphere( void ) const;

      //! Ask if this Node contains any CgFx attributes.
      /** \returns  \c false */
      NVSG_API virtual bool containsCgFx( void ) const;

      //! Invalidate the CgFx containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateCgFxContainment( void );

      //! Ask if this Node contains any LightSource.
      /** \returns  false */
      NVSG_API virtual bool containsLight( void ) const;

      //! Invalidate the LightSource containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateLightContainment( void );

      //! Ask if this Node contains any transparent Material attributes.
      /** \returns  \c false */
      NVSG_API virtual bool containsTransparentMaterial( void ) const;

      //! Invalidate the transparent Material containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateTransparentMaterialContainment( void );

      //! Ask if this Node contains any transparent Texture.
      /** \returns  \c false */
      NVSG_API virtual bool containsTransparentTexture( void ) const;

      //! Invalidate the transparent Texture containment cache.
      /** Invalidating the cache is carried to all parents of this Node.  */
      NVSG_API virtual void invalidateTransparentTextureContainment( void );

      //! Get the number of animation frames that are part of this Node.
      /** \returns 0  */
      NVSG_API virtual size_t getNumberOfFrames( void ) const;

      //! Invalidate the number of animation frames cache.
      /** The cache is invalidated and the invalidation of a Node is called.  */
      NVSG_API virtual void invalidateNumberOfFrames( void ) const;

      //! Set the annotation of this Node.
      NVSG_API void setAnnotation( const char *anno );

      //! Get the annotation of this Node.
      NVSG_API const char * getAnnotation( void ) const;

    protected:
      //! Constructor
      NVSG_API Node(void);

      //! Copy constructor
      NVSG_API Node(const Node& rhs);

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~Node(void);

      //! Interface for calculating the BoundingSphere of this Node.
      /*** \returns \c true if the calculated bounding sphere is valid, otherwise \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const = 0;

      //! Extend the bounding sphere of this Node by another sphere.
      /** When a Node contains a number of children or geometries, it can calculate the bounding sphere of all its
        * constituents and extend its bounding sphere by them. */
      NVSG_API bool extendBoundingSphere( const nvmath::Sphere3f &sphere //!< bounding sphere to extend with
                                        ) const;

      //! Transform the BoundingSphere by a matrix and a factor for the radius.
      NVSG_API void transformBoundingSphere( const nvmath::Mat44f &m      //!<  matrix to transform with
                                           , float factor = 1.0f  //!<  factor for radius
                                           ) const;

    private:
      // helpers with managing parent/child releationships
      friend class Group; // Group needs access to these helpers - no one else!
      void addParent(const Group * group);
      void removeParent(const Group * group);

      // helper with BS validation
      void validateBoundingSphere( void ) const;

    private:
      std::string                   * m_annotation;     //!<  annotation for this node
      std::vector<const Group*>       m_parents;        //!<  vector of parent group nodes
      mutable nvmath::BoundingSphere  m_boundingSphere; //!<  bounding sphere of this node (including all children)
  };

//<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

  inline const char * Node::getAnnotation( void ) const
  {
    return( m_annotation ? m_annotation->c_str() : NULL );
  }

  inline size_t Node::getNumberOfParents() const
  {
    return m_parents.size();
  }
  
  inline const Group * Node::getParent( size_t i ) const
  {
    __ASSERT(i < m_parents.size()); // undefined behavior for invalid index!
    return m_parents[i];
  }

  inline bool Node::isBoundingSphereValid( void ) const
  {
    validateBoundingSphere();
    return( m_boundingSphere.isValid() );
  }

  inline const nvmath::Sphere3f & Node::getBoundingSphere( void ) const
  {
    validateBoundingSphere();
    __ASSERT( m_boundingSphere.isValid() );
    return( m_boundingSphere.getSphere() );
  }

  inline bool Node::extendBoundingSphere( const nvmath::Sphere3f &sphere ) const
  {
    return( m_boundingSphere.extend( sphere ) );
  }

  inline void Node::transformBoundingSphere( const nvmath::Mat44f &m, float factor ) const
  {
    validateBoundingSphere();
    __ASSERT( m_boundingSphere.isValid() );
    m_boundingSphere.transform( m );
    m_boundingSphere.enlarge( factor );
  }

  inline void Node::validateBoundingSphere( void ) const
  {
    if ( ! m_boundingSphere.isValid() )
    {
      calcBoundingSphere();
    }
  }

} // namespace
