// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  "nvutil/RCObject.h"
#include  "nvutil/SWMRSync.h"
#include  "nvsg/DAL.h"
#include  <typeinfo> // using typeid requires this

namespace nvsg
{
  //!  64-bit save DataID.
  /** The DataID is used to identify an instance of an \c Object. */
  typedef unsigned long long DataID;

  //! Object Codes for NVCSG Object Type Identification 
  /** For a custom traverser it is essential to identify all actual scenegraph 
    * components while traversing the scenegraph, and to call the corresponding 
    * handler routine for a particular object. For this reason, all concrete 
    * NVSG object types must be assigned to a unique object code. Secure object 
    * codes for custom defined NVSG objects start at OC_CUSTOMOBJECT.\n\n
    * <B>See also:</B> \ref howtoderiveatraverser and \ref howtoderiveanvsgobject 
    * for further information.
    * \note Choosing an object code lower than OC_CUSTOMOBJECT for a 
    * custom-defined NVSG object results in undefined runtime behavior! */
  enum
  {
    OC_INVALID = 0xFFFFFFFF //!< invalid object code
  , OC_OBJECT = 0           //!< Object
  // nodes (0x00-0x1F)
  , OC_NODE = OC_OBJECT     //!< Node = 0x00 
  , OC_GEONODE              //!< GeoNode
  , OC_GROUP                //!< Group
  , OC_LOD                  //!< LOD
  , OC_SWITCH               //!< Switch
  , OC_TRANSFORM            //!< Transform
  , OC_ANIMATEDTRANSFORM    //!< AnimatedTransform
  , OC_DIRECTEDLIGHT        //!< DirectedLight
  , OC_POINTLIGHT           //!< PointLight
  , OC_SPOTLIGHT            //!< SpotLight
  // drawables (0x20-0x3F)
  , OC_DRAWABLES = 0x20     //!< Drawable = 0x20
  , OC_GEOSET               //!< GeoSet
  , OC_QUADS                //!< Quads
  , OC_ANIMATEDQUADS        //!< AnimatedQuads
  , OC_QUADSTRIPS           //!< QuadStrips
  , OC_TRIANGLES            //!< Triangles
  , OC_ANIMATEDTRIANGLES    //!< AnimatedTriangles
  , OC_SKINNEDTRIANGLES     //!< SkinnedTriangles
  , OC_TRISTRIPS            //!< TriStrips
  // states (0x40-0x5F)
  , OC_STATESET = 0x40      //!< StateSet = 0x40
  , OC_STATEATTRIBUTE       //!< StateAttribute
  , OC_CGFX                 //!< CgFX
  , OC_FACEATTRIBUTE        //!< FaceAttribute
  , OC_MATERIAL             //!< Material
  , OC_TEXTUREATTRIBUTE     //!< TextureAttribute
  , OC_TEXTUREATTRIBUTEITEM //!< TextureAttributeItem
  // animations (0x60-0x7F) 
  , OC_ANIMATION = 0x60                       //!< Animation = 0x60
  , OC_FRAMEDTRAFOANIMATION                   //!< FramedTrafoAnimation
  , OC_FRAMEDVNVECTORANIMATION                //!< FramedVectorAnimation
  , OC_LINEARINTERPOLATEDTRAFOANIMATION       //!< LinearAnimatedTrafoAnimation
  , OC_LINEARINTERPOLATEDVNVECTORANIMATION    //!< LinearInterpolatedVNVectorAnimation
  // cameras (0x80-0x9F)  
  , OC_CAMERA = 0x80                          //!< Camera = 0x80              
  , OC_JITTERCAMERA                           //!< JitterCamera
  , OC_PARALLELCAMERA                         //!< ParallelCamera
  , OC_PERSPECTIVECAMERA                      //!< PerspectiveCamera
  , OC_STEREOCAMERA                           //!< StereoCamera
  // highlight objects (0xA0-0xBF)
  , OC_HIGHLIGHTOBJECT = 0xA0                 //!< HighlightObject = 0xA0
  , OC_HIGHLIGHTBOX                           //!< HighlightBox
  , OC_HIGHLIGHTEDGES                         //!< HighlightEdges
  // custom (0xC0 and higher)
  , OC_CUSTOMOBJECT = 0xC0                    //!< CustomObjects = 0xC0
  };

  //! Base class for all objects.
  /** This class maintains a name and a user pointer for all objects that derive from it. */
  class Object : public nvutil::RCObject
               , public nvutil::ISWMRSync
               , public DALHost // an object can host a DAL
  {
  public:
    //! Identify an object at runtime by means of its object code.
    /** \returns The object code, which is unique per object type. */
    unsigned int getObjectCode() const;

    //! Determine whether the data of this object is shared.
    /** \note Every class deriving from Object should reimplement this function in the same way!
      * \returns  \c true if the object's data is shared, otherwise \c false. */
    NVSG_API virtual bool isDataShared( void ) const;

    //! Get the DataID of the data of this object.
    /** \note Every class deriving from Object should reimplement this function in the same way!
      * \returns  DataID of the object's data. */
    NVSG_API virtual DataID getDataID( void ) const;

    //! Set the name of the object using C++ style string.
    NVSG_API void setName( const std::string& name    //!<  name of the object
                          );

    //! Get the name of the object.
    /** \returns  name of the object  */
    NVSG_API const std::string& getName( void ) const;

    //! Acquire read access in a multithreaded environment.
    /** For safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
      * contemporaneous access by multiple readers but restricts write access to a single writer.  
      * \par
      * This function returns either when it obtains read access or when the time-out interval elapses. By default,
      * this function waits indefinitely.
      * \returns \c true, if read access could be acquired, \c false otherwise.
      */
    NVSG_API bool waitToRead( size_t milliseconds = 0xFFFFFFFF   //<! Time-out interval, in milliseconds. 
                            ) const;

    //! Signal when finished reading the object.
    /** Read access to this object, acquired by a previous call to waitToRead, is ended
      * by a call to this function. */
    NVSG_API void doneReading( void ) const;
    
    //! Acquire write access in a multithreaded enviroment.
    /** For safety in a multithreaded environment, an object maintains a read/write locking mechanism that allows 
      * contemporaneous access for multiple readers but restricts write access to a single writer. 
      * \par
      * This function returns either when it gets write access or when the time-out interval elapses. By default,
      * this function waits indefinitely.
      * \returns \c true, if write access could be acquired, \c false otherwise.
      */
    NVSG_API bool waitToWrite( size_t milliseconds = 0xFFFFFFFF   //! Time-out interval, in millisceconds.
                             ) const;
    
    //! Signal when finished writing the object.
    /** Write access to this object, acquired by a previous call to waitToWrite, is ended
      * by a call to this function. */
    NVSG_API void doneWriting( void ) const;

  protected: 
    //! Constructor
    NVSG_API Object( void );

    //! Copy constructor
    NVSG_API Object( const Object& rhs );

    //! Protected destructor to prevent explicit creation on stack.
    NVSG_API virtual ~Object();

    unsigned int      m_objectCode; //!< object code (unique per object type)
                                    // accessible for derived classes

  private:
    std::string       m_name;   //!< Every object derived class can have its personal name.
    nvutil::SWMRSync  m_lock;   //!< Access synchronization between different threads.
  };

  inline unsigned int Object::getObjectCode() const
  {
    __TRACE();
    return m_objectCode;
  }

  inline void Object::setName( const std::string& name )
  {
    __TRACE();
    m_name = name;
  }

  inline const std::string& Object::getName() const
  {
    __TRACE();
    return( m_name );
  }

  inline bool Object::waitToRead( size_t milliseconds ) const
  {
    __TRACE();
    size_t status;
    return m_lock.waitToRead(milliseconds, status);
  }

  inline void Object::doneReading( void ) const
  {
    __TRACE();
    m_lock.doneReading();
  }

  inline bool Object::waitToWrite( size_t milliseconds ) const
  {
    __TRACE();
    size_t status;
    return m_lock.waitToWrite(milliseconds, status);
  }

  inline void Object::doneWriting( void ) const
  {
    __TRACE();
    m_lock.doneWriting();
  }

  /*! \relates nvsg::Object
    * Initiates a safe edit operation on an NVSG Object.
    *
    * The function returns when the calling thread successfully obtained write access to the object.
    * If the object is currently locked for write or read access from a concurrent thread, the calling thread
    * enters the wait state. It uses no processor time while waiting to obtain write access.
    *
    * After getting write access to the object, the function returns a pointer to the non-const object.
    * Throught this pointer, the const and especially the non-const interface of the object can be accessed.
    *
    * When the function returns, the object is protected against any write or read access from concurrent threads.
    * After the desired edit operation has been completed, the write lock for the object must be released through
    * a corresponding call to \link nvsg::endEdit endEdit \endlink. Missing the endEdit call after completing the 
    * edit operation can cause the system to become deadlocked!
    *
    * \returns A pointer to the now writable object.
    *
    * \note The behavior is undefined if called with an invalid pointer. */ 
  template<typename T>
  inline T * beginEdit( const T * obj  //!< Specifies the object for which write access is requested.
                      )
  {
    __TRACE();
    __ASSERT(obj); // undefined behavior if this fires 
    obj->waitToWrite();
    return const_cast<T *>(obj);
  }

  /*! \relates nvsg::Object
    * Ends an edit operation on an NVSG Object.
    *
    * The function ends an edit operation that previously has been initiated through a call to 
    * \link nvsg::beginEdit beginEdit \endlink. 
    *
    * The function must be called with the pointer to the non-const object, that has been returned by
    * the matching call to beginEdit. The write lock for the object will be released and the function
    * returns a pointer to the const object, through which only the const interface of the object could 
    * be utilized.
    *
    * Immediately after the call, the passed pointer to the non-const object becomes invalid. Hence, using
    * the pointer to the non-const object after this call leads to undefined behavior! 
    *
    * \returns A pointer to the non-const object, which is not writable anymore.
    * 
    * \note The behavior is undefined if called with an invalid pointer. */
  template<typename T>
  inline const T * endEdit(T *& obj //!< Specifies the object for which write access should be released.
                          )
  {
    __TRACE();
    __ASSERT(obj); // undefined behavior if this fires
    const T* cobj = obj;
    cobj->doneWriting();
    obj = NULL; // force undefined behavior if accessed after endEdit
    return( cobj );
  }

  // this specialization detects wrong usage of endEdit at compile time
  template<typename T>
  inline const T * endEdit(const T *& obj)
  { // !!! must not call endEdit with a 'pointer to const' !!!

    // generate a compiler error
    Object * p=obj;
  }

  //! Detect if the \c Object pointed to by \a p is a pointer to a specified (templated) type.
  /** \returns \c true if \a p is a pointer to the specified type, \c false otherwise. */
  template<typename T> 
  inline bool isTypeOf( const Object *p )
  {
    return( dynamic_cast<const T *>( p ) != NULL );
  }

  //! Detect if the two \c Objects \a p0 and \a p1 are of the same type.
  /** \return \c true if \a p0 and \a p1 are of the same type, \c false otherwise. */
  inline bool areOfSameType( const Object &p0, const Object &p1 )
  {
    return( !! ( typeid( p0 ) == typeid( p1 ) ) );
  }

} //  namespace nvsg
