// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Camera.h"

namespace nvsg
{
  class MonoCamera;
  //! A stereoscopic camera
  /** With this camera stereoscopic vision is introduced. */
  class StereoCamera : public Camera
  {
    public:
      //! Create a StereoCamera using a MonoCamera as a base.
      /** \returns a constant pointer to a StereoCamera. */
      NVSG_API static const StereoCamera * create( const MonoCamera &base  //!< base of the StereoCamera
                                                 );

      //! Create a StereoCamera by copying from an other.
      NVSG_API static const StereoCamera * createFromBase( const Camera &rhs      //!< Camera to copy from
                                                         , const MonoCamera &base //!< base of the StereoCamera
                                                         );

      //! Create a clone of this StereoCamera.
      /** \returns  pointer to the new PerspectiveCamera as a pointer to a Camera */
      NVSG_API virtual const StereoCamera * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns \c true if the object's data is shared, otherwise \c false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      // 
      // stereo interface
      //

      //! Get the depth factor.
      /** The depth factor controls how much the two cameras are horizontally displaced from the base camera.
        * It is measured in multiples of the target distance. A value of 0.0 results in monoscopic vision. A value
        * of 0.03 (the default) gives a good stereo effect. You can imagine that the user is about one meter from
        * the monitor and his eyes are about 0.03 meters displaced to the right and left. The larger the depth factor, the
        * better the stereo effect, until the left and right images are too spread out for eyes to fuse. */
      NVSG_API float getDepthFactor( void ) const;

      //! Set the depth factor.
      /** The depth factor controls how much the two cameras for stereo vision are horizontally displaced from the base
        * camera. It is measured in multiples of the target distance. A value of 0.0 results in monoscopic vision. A value
        * of 0.03 (the default) gives a good stereo effect. You can imagine that the user is about one meter from
        * the monitor and his eyes are about 0.03 meters displaced to the right and left. The larger the depth factor, the
        * better the stereo effect, until the left and right images are too spread out for the eyes to fuse. */ 
      NVSG_API void setDepthFactor( float df    //!<  new depth factor
                                  );

      //! Get the focal distance.
      /** The focus distance controls where the focal plane is in the model. It is measured in the same units as the
        * world space. A value equal to the near distance (the default) makes the near plane the focal plane. This results
        * in a stereo image that seems to be completely inside of the monitor. A value equal to the far distance makes
        * the far plane the focal plane. This results in a stereo image that seems to be completely in front of the
        * monitor. */
      NVSG_API float getFocusDistance( void ) const;

      //! Set the focus distance.
      /** The focus distance controls where the focal plane is in the model. It is measured in the same units as the
        * world space. A value equal to the near distance (the default) make the near plane the focal plane. This results
        * in a stereo image that seems to be completely inside of the monitor. A value equal to the far distance makes
        * the far plane the focal plane. This results in a stereo image that seems to be completely in front of the
        * monitor. */
      NVSG_API void setFocusDistance( float fd    //!<  new focus distance
                                    );

      //! Ask if \a pass is the left pass.
      /** A stereo camera is a two pass camera. On rendering, the scene is traversed two times, once for the left eye
        * and once for the right eye. Depending on the synchronization between the monitor and the stereo glasses,
        * it might be important to know which pass is for the left eye and which pass for the right eye.
        * \returns  \c true if \a pass is the index of the left pass, otherwise \c false */
      NVSG_API bool isLeftPassIndex( size_t pass    //!<  pass index to ask for
                                   ) const;

      //! Reverse left and right pass order.
      /** A stereo camera is a two pass camera. On rendering, the scene is traversed two times, once for the left eye
        * and once for the right eye. Depending on the synchronization between the monitor and the stereo glasses,
        * it might be necessary to reverse this order to get the correct stereo effect. */
      NVSG_API void reverseLeftRight( void );

      //! Update the monoscopic cameras (right and left)
      /** Before traversing the monoscopic cameras (right and left) of a stereo camera, this function should be called
        * to update these cameras with any changes that might have applied to the base camera. */
      NVSG_API void updateMonoCameras( void ) const;

      //! Get the base camera.
      /** \returns a const poniter to the monoscopic base camera */
      const MonoCamera * getMonoCamera( void ) const;

      //! Get the camera for the left eye.
      /** \return a constant pointer to the monoscopic camera representing the left eye.  */
      const MonoCamera * getLeftCamera() const;

      //! Get the camera for the right eye.
      /** \return a constant pointer to the monoscopic camera representing the right eye. */
      const MonoCamera * getRightCamera() const;

      //
      // head light interface
      //

      //! Add a head light on this camera.
      /** A camera can have a number of headlights that are positioned relative to the camera, i.e. they move together
        * with the camera. These lights are not part of the scene, but part of the camera. The default position and
        * orientation of a new headlight is the camera position and orientation. */
      NVSG_API virtual void addHeadLight( 
        const LightSource * pLight                                              //!<  The light source to add
      , const nvmath::Vec3f &position = nvmath::Vec3f( 0.0f, 0.0f, 0.0f )       //!<  The position relative to the camera
      , const nvmath::Quatf &quat     = nvmath::Quatf( 0.0f, 0.0f, 0.0f, 1.0f ) //!<  The rotation relative to the camera
      );

      //! Get number of headlights attached to the camera.
      /** \returns  The number of headlights attached to camera */
      NVSG_API virtual size_t getNumberOfHeadLights( void ) const;

      //! Get the indexed head light of this camera
      /** \returns  A constant pointer to the \a indexed headlight*/
      NVSG_API virtual const LightSource * getHeadLight( size_t index   //!<  index of head light to get
                                                       ) const;

      //! Remove a headlight.
      /** \returns \c true if headlight was removed, otherwise \c false */
      NVSG_API virtual bool removeHeadLight( size_t index = 0   //!<  index of headlight to remove
                                           );

      //! Set the head light position and orientation relative to the camera
      /** \returns true if a headlight with this index exists, otherwise false */
      NVSG_API virtual bool setHeadLightTrafo( const nvmath::Vec3f &trans   //!<  position of headlight
                                             , const nvmath::Quatf &quat    //!<  orientation of headlight
                                             , size_t index = 0     //!<  index of headlight
                                             );

      //! Get the position/orientation of the indexed head light of this camera
      /** It is assumed that the camera holds at least \a index headlights. */
      NVSG_API virtual void getHeadLightTrafo( nvmath::Vec3f &trans       //!<  translation relative to the camera
                                             , nvmath::Quatf &quat        //!<  rotation relative to the camera
                                             , size_t index = 0   //!<  index of the headlight to get information from
                                             ) const;

      //
      // basic camera interface
      //

      //! Get the distance to the target point.
      /** \returns  The distance to the target point */
      NVSG_API virtual float getTargetDistance( void ) const;

      
      //! Set the distance to the target point.
      NVSG_API virtual void setTargetDistance( float td   //!<  distance to target point (or viewing plane)
                                             );

      //! Get the up direction.
      /** \returns  The up direction of camera */
      NVSG_API virtual nvmath::Vec3f getUpVector( void ) const;

      //! Get the camera position.
      /** \returns  The position of camera */
      NVSG_API virtual const nvmath::Vec3f & getPosition( void ) const;

      //! Set the camera position.
      NVSG_API virtual void setPosition( const nvmath::Vec3f& pos   //!<  position
                                       );

      //! Set the up direction.
      NVSG_API virtual void setUpVector( const nvmath::Vec3f &up    //!<  up direction (might change viewing direction)
                                       );

      //! Get viewing direction
      /** The default viewing direction of a SimpleCamera is along the negative z-axis.
        * \returns  The viewing direction */
      NVSG_API virtual nvmath::Vec3f getDirection( void )  const;

      //! Set the viewing direction.
      NVSG_API virtual void setDirection( const nvmath::Vec3f& dir    //!<  new viewing direction (might change upVector)
                                        );

      //! Set the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Quatf &quat   //!<  quaternion that determines direction and up vector
                                          );

      //! Set the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &dir    //!<  direction to look at
                                          , const nvmath::Vec3f &up     //!<  up vector
                                          );

      //! Interface for setting the orientation of the camera.
      NVSG_API virtual void setOrientation( const nvmath::Vec3f &axis   //!<  axis to rotate about
                                          , float angle         //!<  angle to rotate
                                          );
      //! Get the current orientation.
      /** \returns  quaternion that describes the orientation */
      NVSG_API virtual nvmath::Quatf getOrientation( void ) const;

      //! Get the inverse of the camera specific world-to-view transformation
      /** \returns  The inverse of the world-to-view transformation (i.e. view to world) */
      NVSG_API virtual nvmath::Mat44f getViewToWorldMatrix( void ) const;

      //! Get the world to view transformation.
      /** \returns  The world-to-camera transformation */
      NVSG_API virtual nvmath::Mat44f getWorldToViewMatrix( void ) const;

      //! Get the window offset.
      /** \returns  The window offset */
      NVSG_API virtual const nvmath::Vec2f & getWindowOffset( void ) const;

      //! Set the window offset.
      NVSG_API virtual void setWindowOffset( const nvmath::Vec2f &offset    //!<  window ofset to viewing reference point
                                            );

      //! Get the relative window region to view
      NVSG_API virtual void getWindowRegion( nvmath::Vec2f &lowerLeft     //!< relative lower left position
                                           , nvmath::Vec2f &upperRight    //!< relative upper right position
                                           ) const;

      //! Set the relative window region to view
      NVSG_API virtual void setWindowRegion( const nvmath::Vec2f &lowerLeft     //!< relative lower left position
                                           , const nvmath::Vec2f &upperRight    //!< relative upper right position
                                           );

      //! Get the window size.
      /** \returns  The window size */
      NVSG_API virtual const nvmath::Vec2f & getWindowSize( void ) const;

      //!  Set the window size.
      NVSG_API virtual void setWindowSize( const nvmath::Vec2f &size    //!<  window size (world coordinates)
                                         );

      //! Get the current aspect ratio
      /** \returns  The width versus height of the camera window */
      NVSG_API virtual float getAspectRatio( void ) const;

      //! Set the aspect ratio.
      NVSG_API virtual void setAspectRatio( float ar                  //!<  new aspect ratio
                                          , bool keepWidth = false    //!<  keep width or height constant
                                          );

      //! Enable/Disable automatic clip plane determination.
      NVSG_API virtual void setAutoClipPlanes( bool on    //!<  true to enable, false to disable
                                             );
      
      //! Get auto clipping state.
      /** If auto clip is on, the camera determines the optimal near and front clipping distances out of the scene. If
        * the user has additional information about the scene that might help clipping, it should be turned off.
        * \returns \c true if auto clip is on, otherwise \c false */
      NVSG_API virtual bool isAutoClipPlanes( void ) const;

      //! Get distance to the far clipping plane
      /** Near and far distances can be used internally to the camera to improve depth resolution.
        * \returns  The distance to the far clipping plane */
      NVSG_API virtual float getFarDistance( void ) const;

      //! Set the far clipping distance.
      NVSG_API virtual void setFarDistance( float fd    //!<  far clipping distance
                                          );

      //! Get distance to the near clipping plane
      /** Near and far distances can be used internally to the camera to improve depth resolution.
        * \returns  The distance to the near clipping plane */
      NVSG_API virtual float getNearDistance( void ) const;

      //! Set the near clipping distance.
      NVSG_API virtual void setNearDistance( float nd   //!<  near clipping distance
                                           );

      //! Calculate the distances to the near and far clipping planes.
      /** Near and far distances can be used internally to the camera to improve depth resolution. This function
        * should determine these values such that the sphere given by \a center and \a radius just fits between them. */
      NVSG_API virtual void calcNearFarDistances( const nvmath::Sphere3f &sphere  //!<  bounding sphere of the world to look at
                                                );

      //
      // operations to perform on a camera
      //

      //! Move the camera.
      NVSG_API virtual void move( const nvmath::Vec3f& delta    //!<  delta to move the camera (camera coordinates)
                                );

      //! Orbit the camera.
      NVSG_API virtual void orbit( const nvmath::Vec3f& axis    //!<  axis to orbit around (camera coordinates)
                                 , float rad            //!<  angle to orbit (radians)
                                 );

      //! Interface for rotating the camera.
      NVSG_API virtual void rotate( const nvmath::Vec3f& axis             //!<  axis to rotate around 
                                  , float rad                     //!<  angle to rotate (radians)
                                  , bool  cameraRelative = true   //!<  flag to indicate if the axis is camera or world relative
                                  );

      //! Zoom by a factor.
      NVSG_API virtual void zoom( float factor    //!< zoom in/out by this factor; changes windowSize
                                );

      //! Zoom to fit view to a sphere.
      NVSG_API virtual void zoom( const nvmath::Sphere3f &sphere  //!<  sphere to zoom on
                                , float fovy = -1.0f    //!<  field of view; if negative, keep the current
                                );

      
    protected:
      //! Constructor
      /** A StereoCamera uses an arbitrary MonoCamera as its base camera. This base camera is used and slightly adjusted
        * to get stereo vision. */
      NVSG_API StereoCamera( const MonoCamera& base   //!<  base camera to use
                           );

      //! Partial Constructor
      NVSG_API StereoCamera( const Camera &rhs, const MonoCamera &base );

      //! Copy constructor
      NVSG_API StereoCamera( const StereoCamera& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~StereoCamera( void );

    private:
      // assignment not permitted
      StereoCamera& operator=( const StereoCamera& rhs ) {}

    private:
      const MonoCamera  * m_baseCamera;
      float         m_depthFactor;
      float         m_focusDistance;
      int           m_leftPassIndex;

      mutable const MonoCamera  * m_cameras[2];
  };

  // - - - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - - - -

  inline float StereoCamera::getDepthFactor( void ) const
  {
    __TRACE();
    return( m_depthFactor );
  }

  inline float StereoCamera::getFocusDistance( void ) const
  {
    __TRACE();
    return( m_focusDistance );
  }

  inline const MonoCamera *  StereoCamera::getMonoCamera( void ) const
  {
    __TRACE();
    return( m_baseCamera );
  }

  inline bool StereoCamera::isLeftPassIndex( size_t pass ) const
  {
    __TRACE();
    return( pass == m_leftPassIndex );
  }

  inline void StereoCamera::reverseLeftRight( void )
  {
    __TRACE();
    m_leftPassIndex = ! m_leftPassIndex;
  }

  inline void StereoCamera::setDepthFactor( float df )
  {
    __TRACE();
    m_depthFactor = df;
  }

  inline void StereoCamera::setFocusDistance( float fd )
  {
    __TRACE();
    m_focusDistance = fd;
  }

  inline const MonoCamera * StereoCamera::getLeftCamera() const
  {
    __TRACE();
    return m_cameras[0];
  }

  inline const MonoCamera * StereoCamera::getRightCamera() const
  {
    __TRACE();
    return m_cameras[1];
  }

} // namespace
