// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include <set>
#include "nvsg/Group.h"

namespace nvsg
{
  //! Switch node
  /** A switch node is a special kind of grouping node 
    * where one can define which children will be traversed or not (active/inactive children).\n
    * See LightSource documentation on how to use a switch for light sources.*/
  class Switch : public Group
  {
    public:
      //! Create Switch.
      /** \returns a constant pointer to a Switch. */
      NVSG_API static const Switch * create( void );

      //! Create Switch by copying from a Group.
      /** \returns a constant pointer to a Switch. */
      NVSG_API static const Switch * createFromBase( const Group &rhs );

      //! Get a clone of this \c Switch object.
      /** \returns a constant pointer to a Switch. */
      NVSG_API virtual const Switch * clone( void ) const;

      //! Set all children to active
      NVSG_API void setActive( void );

      //! Set one child to active.
      /** \note Currently active children will stay active!*/
      NVSG_API void setActive( size_t index   //!< Index of the child.
                             );

      //! Set all children to inactive
      NVSG_API void setInactive( void );

      //! Set one child to inactive.
      /** \note Currently inactive children will stay inactive!*/
      NVSG_API void setInactive( size_t index   //!< Index of the child.
                               );

      //! Get the number of active children.
      /** \returns  The number of active children */
      NVSG_API size_t getNumberOfActive( void ) const;

      //! Get indices to active child nodes.
      /** Copies the indices of all active child nodes to the vector /a indices.
        * \returns  The number of indices copied to /a indices.
        * \note The internal format a Switch uses for storing indices is size_t, which expands to
        * 32-bit for 32-bit environments and 64-bit for 64-bit environments. 
        */
      NVSG_API size_t getActive( std::vector<size_t>& indices ) const;

      //! Determine if any of the children is active.
      /** \return \c true - at least one child is active. \n
        *         \c false - no child is active. */
      NVSG_API bool isActive( void ) const;

      //! Determine if the child is active.
      /** \return \c true - the child is active. \n
        *         \c false - the child is active. */
      NVSG_API bool isActive( size_t index    //!< Index of the child.
                            ) const;

      NVSG_API virtual bool isDataShared( void ) const;
      NVSG_API virtual DataID getDataID( void ) const;

    protected:
      //! Default constructor.
      NVSG_API Switch( void );

      //! Partial Constructor
      NVSG_API Switch( const Group &rhs );

      //! Copy Constructor
      NVSG_API Switch( const Switch &rhs );

      //! Prohibit explicit creation on stack by making the destructor protected.
      NVSG_API virtual ~Switch( void );

      //! Calculate the bounding sphere of this switch.
      /** Calculate the bounding sphere for the active children only.
        * If this group doesn't contain any real geometry (
        * for example only light sources without a bounding sphere),
        * it doesn't have a valid bounding sphere.
        * \returns \c true if the calculated bounding sphere is valid, otherwise \c false. */
      NVSG_API virtual bool calcBoundingSphere( void ) const;

    private:
      /** Index set (ordered) into the vector of children that's currently in use */
      std::set<size_t> m_activeChildren;
  };
  
} //  namespace nvsg
