// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvutil/RCObject.h"
#include "nvsg/DAL.h"

namespace nvsg
{


  //! Class to hold a texture
  /** This class is used to represent a texture of any format in the scenegraph. It is defined by the filename and hides
    * all the format specific information behind this common interface.
    * \par
    * A Texture might contain several images (for example a texture can be a cube map, consisting of six images). Each
    * image might be mipmapped. */
  class Texture : public nvutil::RCObject
                , public DALHost // an object can host a DAL
  {
    public:   
      //! Enumeration of the supported texture file formats.
      typedef enum
      {
        TFT_UNKNOWN = -1  //!< Unknown format.
      , TFT_DDS           //!< DDS format.
      , TFT_JPG           //!< JPG format.
      , TFT_TGA           //!< TGA format.
      , TFT_BMP           //!< BMP format.
      , TFT_TIF           //!< TIF format.
      , TFT_RGB           //!< RGB format.
      , TFT_GIF           //!< GIF format.
      , TFT_PNG           //!< PNG format.
      , TFT_HDR           //!< HDR format.
      // to be continued
      } eTexFileType;
   
    public:
      //! Create a Texture out of a file.
      NVSG_API static const Texture * create( const std::string &fileName //!<  file name of texture
                                            , bool flipImage = true       //!<  flip image horizontally if true
                                            );

      //! Get the components of the mipmap level of an image of the texture.
      /** \returns  The bytes per pixel of this image */
      NVSG_API int getComponents( unsigned int image = 0    //!<  image to get bpp from
                                , unsigned int mipmap = 0   //!<  mipmap to get bpp from
                                ) const;

      //! Get the type of the mipmap level of an image of the texture.
      /** \returns The data type of this image */
      NVSG_API int getType( unsigned int image = 0    //!<  image to get bpp from
                          , unsigned int mipmap = 0   //!<  mipmap to get bpp from
                          ) const;

      //! Get the depth of the mipmap level of an image of the texture.
      /** \returns  The depth, if it is a 3D-texture, otherwise 1  */
      NVSG_API int getDepth( unsigned int image = 0     //!<  image to get depth from
                           , unsigned int mipmap = 0    //!<  mipmap to get depth from
                           ) const;

      //! Get the format of the mipmap level of an image of the texture.
      /** \returns  The format of the texture (equals GL-format like GL_RGBA)  */
      NVSG_API int getFormat( unsigned int image = 0    //!<  image to get format from
                            , unsigned int mipmap = 0   //!<  mipmap to get format from
                            )  const;
      //! Get the filename of the texture.
      /** \returns The filename of the texture */
      std::string getFileName( void ) const;

      //! Get the handle of the texture.
      /** Each texture can have an associated handle that can be queried here.
        * \returns  The texture handle of the texture */
      unsigned int getHandle( void ) const;

      //! Get the height of the mipmap level of an image of the texture.
      /** \returns  The height of the mipmap level of this image */
      NVSG_API int getHeight( unsigned int image = 0    //!<  image to get height from
                            , unsigned int mipmap = 0   //!<  mipmap to get height from
                            ) const;

      //! Get the number of image in this texture.
      /** \returns  The number of images in the texture (e.g. 6 for a cubemap)  */
      unsigned int getNumberOfImages( void ) const;

      //! Get the number of mipmaps of an image of the texture.
      /** \returns  The number of mipmaps of \a image */
      NVSG_API unsigned int getNumberOfMipmaps( unsigned int image = 0    //!<  image to get information from
                                              ) const;

      //! Get pointer to pixels of the mipmap level of an image of the texture.
      /** \returns  A constant pointer to image data  */
      NVSG_API const void * getPixels( unsigned int image = 0     //!<  image to get data from
                                     , unsigned int mipmap = 0    //!<  mipmap to get data from
                                     )  const;

      //! Get the size of the mipmap level of an image of the texture.
      /** \returns  The size of the image data mipmap level */
      NVSG_API int getSize( unsigned int image = 0    //!<  image to get image size from
                          , unsigned int mipmap = 0   //!<  mipmap to get image size from
                          ) const;

      //! Get the target of the texture.
      /** The target is an arbitrary ID that can be used to classify this texture.
        * \returns  The target of the texture */
      NVSG_API unsigned long getTarget( void ) const;

      //! Get the width of the mipmap-level of an image of the texture.
      /** \returns  The width of the mipmap-level of this image  */
      NVSG_API int getWidth( unsigned int image = 0     //!<  image to get width from
                           , unsigned int mipmap = 0    //!<  mipmap to get width from
                           ) const;

      //! Ask if this texture is a cubemap.
      /** \returns  \c true, if this texture is a cubemap, otherwise \c false */
      NVSG_API bool isCubeMap( void ) const;

      //! Ask if this texture is a floating point texture (4 bpp per color channel).
      /** \returns  \c true, if this texture is a floating point texture, otherwise \c false */
      NVSG_API bool isFloatingPoint( void ) const;

      //! Ask if this texture is of a specific file type
      /** \returns  \c true, if the file type is \a tft, otherwise \c false */
      NVSG_API bool isTexFileType( eTexFileType tft   //!<  texture file type to test for
                                 ) const;

      //! Load the texture.
      /** \returns  \c true, if the texture was successfully loaded, otherwise \c false */
      NVSG_API bool load( void )  const;

      //! Scale an image of the texture to the specified size.
      NVSG_API void scale( unsigned int image = 0       //!<  The image to scale - 2D textures conatain one image, 
                                                        //!   so 0 is a reasonable value;\n 
                                                        //!   Cube maps contain 6 images so the index runs from 0-4 
                         , unsigned int width = 128     //!<  The width to achieve
                         , unsigned int height = 128    //!<  The height to achieve
                         , unsigned int depth = 1       //!<  The depth to achieve - only usefull for 3D textures.
                         ) const;

      //! Set the handle of the texture.
      NVSG_API void setHandle( unsigned int handle    //!<  handle to use with this Texture
                             ) const;

      //! Set the target of the texture.
      /** The target is an arbitrary ID that can be used to classify this texture.  */
      NVSG_API void setTarget( unsigned long target   //!<  target for this texture
                             ) const;

      //! Unload the texture.
      NVSG_API void unload( void ) const;

    protected: 
      //! Constructor
      /** Creates a texture out of the file \a fileName. If requested, it flips the image horizontally. */
      NVSG_API Texture( const std::string& fileName, bool flipImage = true );

      //! Prohibit explicit creation on stack by making the destructor protected.
      NVSG_API virtual ~Texture( void );

    protected:
      mutable bool          m_cube;         //!< true for cube maps
              std::string   m_fileName;     //!< file name of the texture
              bool          m_flipImage;    //!< true for horizontally flipped images
      mutable unsigned int  m_handle;       //!< opaque handle of the texture
      mutable unsigned int  m_imageID;      //!< image id of the texture
      mutable unsigned int  m_numImages;    //!< number of images in the texture
      mutable unsigned long m_target;       //!< target type of the texture (GL_TEXTURE_1D,...)
      mutable eTexFileType  m_texFileType;  //!< texture file type, like TFT_DDS, TFT_JPG (see eTexFileType)

    private:
      // copying is not permitted
      Texture( const Texture& rhs );
  };

  inline std::string Texture::getFileName( void ) const
  {
    __TRACE();
    return( m_fileName );
  }
  inline unsigned int Texture::getHandle( void ) const
  {
    __TRACE();
    return( m_handle );
  }

  inline bool Texture::isCubeMap() const 
  {
    __TRACE();
    return m_cube; 
  }

  inline bool Texture::isTexFileType( eTexFileType tft ) const
  {
    __TRACE();
    return( m_texFileType == tft );
  }

  inline unsigned int Texture::getNumberOfImages( void ) const
  {
    __TRACE();
    return m_numImages;
  }

  inline void Texture::setHandle( unsigned int handle ) const
  {
    __TRACE();
    m_handle = handle;
  }

  inline void Texture::setTarget( unsigned long target ) const
  {
    __TRACE();
    m_target = target;
  }

  inline unsigned long Texture::getTarget( void ) const
  {
    __TRACE();
    return( m_target );
  }

} //  namespace nvsg
