// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include <map>
#include "nvmath/Plane3f.h"
#include "nvmath/Trafo.h"
#include "nvsg/StateAttribute.h"
#include "nvsg/Texture.h"

namespace nvsg
{
  //! Texture types
  typedef enum
  {
    TT_INVALID     = 0,       //!< invalid texture type
    TT_TRANSPARENT = 1,       //!< transparent texture type - blending on
    TT_DIFFUSE     = 3,       //!< diffuse texture
  } TextureType;

  //! Texture coordinate axis
  typedef enum
  {
    TCA_S = 0
  , TCA_T
  , TCA_R
  , TCA_Q
  } TexCoordAxis;

  //! Texture environment mode
  typedef enum
  {
    TEM_REPLACE = 0  //!< C = C_t; A = A_t
  , TEM_MODULATE     //!< C = C_f*C_t; A = A_f*A_t
  , TEM_DECAL        //!< C = C_f*(1-A_t)+C_t*A_t; A = A_f
  , TEM_BLEND        //!< C = C_f*(1-C_t)+C_c*C_t; A = A_f*A_t
  , TEM_ADD          //!< C = min(1,C_f+C_t); A = A_f*A_t
  } TextureEnvMode;

  //! Texture coordinate generation linear mode
  typedef enum
  {
    TGML_OBJECT = 0
  , TGML_EYE
  } TexGenLinearMode;

  //! Texture coordinate generation mode
  typedef enum
  {
    TGM_OFF           = -1
  , TGM_OBJECT_LINEAR = TGML_OBJECT
  , TGM_EYE_LINEAR    = TGML_EYE
  , TGM_SPHERE_MAP
  //, TGM_REFLECTION_MAP
  //, TGM_NORMAL_MAP
  } TexGenMode;

  //! Texture wrap mode
  typedef enum
  {
    TWM_REPEAT = 0    //!< Corresponds to GL_REPEAT in OpenGL
  , TWM_CLAMP         //!< Corresponds to GL_CLAMP in OpenGL
  , TWM_MIRROR        //!< Corresponds to GL_MIRRORED_REPEAT_ARB in OpenGL
  } TextureWrapMode;

  //! Texture minification mode
  typedef enum
  {
    TFM_MIN_NEAREST = 0     //!< Corresponds to GL_NEAREST in OpenGL
  , TFM_MIN_BILINEAR        //!< Corresponds to GL_LINEAR in OpenGL
  , TFM_MIN_TRILINEAR       //!< Corresponds to GL_LINEAR_MIPMAP_LINEAR in OpenGL
  } TextureMinFilterMode;

  //! Texture magnification mode
  typedef enum
  {
    TFM_MAG_NEAREST = 0     //!< Corresponds to GL_NEAREST in OpenGL
  , TFM_MAG_BILINEAR        //!< Corresponds to GL_LINEAR in OpenGL
  } TextureMagFilterMode;

  //! Data of a texture attribute item that is worth sharing
  class TextureAttributeItemData : public nvutil::RCObject
  {
      friend class TextureAttributeItem;
    public:
      //! Copy constructor.
      /** Initialize a new copy from \a rhs */
      NVSG_API TextureAttributeItemData(const TextureAttributeItemData& rhs);

    private:
      // instantiation on stack is not permitted.
      TextureAttributeItemData();
      ~TextureAttributeItemData();
      // assignment also not permitted
      TextureAttributeItemData& operator=(const TextureAttributeItemData& rhs) {}

      std::string           m_fileName;           // texture source file
      nvmath::Vec4f         m_borderColor;
      TextureEnvMode        m_envMode;            // enum environment mode to use with rasterization
      TextureMagFilterMode  m_magFilterMode;
      TextureMinFilterMode  m_minFilterMode;
      TextureType           m_type;               // enum type of the texture
      nvmath::Trafo         m_trafo;              // texture coordinate transformation
      TextureWrapMode       m_wrapMode[3];
      TexGenMode            m_texGenMode[4];
      nvmath::Plane3f       m_texGenPlane[4];
  };

  //! Texture attribute item
  /** This class serves as an abstract layer for a single texture object. The use of
    * a \c TextureAttributeItem is meaningful only in conjunction with a \c TextureAttribute, 
    * which provides an interface to bind a \c TextureAttributeItem to a specified texture
    * unit. */
  class TextureAttributeItem : public Object
  {
    friend class TextureAttribute;

  public:
    //! Create a TextureAttributeItem
    /** \returns a constant pointer to a TextureAttributeItem. */
    NVSG_API static const TextureAttributeItem * create( void );

    //! Create a TextureAttributeItem by copying from an Object.
    /** \returns a constant pointer to a TextureAttributeItem. */
    NVSG_API static const TextureAttributeItem * createFromBase( const Object &rhs  //!<  Object to copy from
                                                               );

    //! Create a TextureAttributeItem by cloning an other.
    /** \returns a constant pointer to a TextureAttributeItem. */
    NVSG_API virtual const TextureAttributeItem * clone( void ) const;

    //! Determine whether the data of this object is shared.
    /** \returns  \c true if the object's data is shared, otherwise \c false.  */
    NVSG_API virtual bool isDataShared(void) const;

    //! Get the DataID of the data of this object.
    /** \returns  DataID of the object's data.  */
    NVSG_API virtual DataID getDataID(void) const;

    //! Set texture file name.
    /** Use this to specify the name of the image file from which to load the texture */
    NVSG_API void setFileName( const std::string& fname  //!<  Name of the image file
                             );

    //! Get texture file name.
    /** \returns The name of the image file from which to load the texture. */
    NVSG_API const std::string& getFileName( void ) const;

    //! Get texture.
    /** \returns  A constant pointer to the device dependent texture object.*/
    NVSG_API const Texture * getTexture( void ) const;

    //! Set texture.
    /** This is used to attach the device dependent texture object to this texture attribute item.
      * \note The device dependent texture will be generated while rendering the scenegraph. As 
      * rendering is a const traverser step working on a const texture attribute item, this function
      * needs to be a const member function.*/ 
    NVSG_API void setTexture( const Texture *p //!< Pointer to the Texture object to be set. 
                            ) const;

    //! Sets the texture border color
    NVSG_API void setBorderColor( const nvmath::Vec4f &color );

    //! Gets the texture border color
    NVSG_API const nvmath::Vec4f& getBorderColor( void ) const;

    //! Get texture environment mode.
    /** \returns The texture environment mode used with rasterizing */
    NVSG_API TextureEnvMode getEnvMode( void ) const;

    //! Set texture environment mode
    /** Use this to specify the texture environment mode to be used with rasterizing. */
    NVSG_API void setEnvMode( TextureEnvMode mode //!< Environment mode.
                            );

    //! Set the texture magnification filter mode
    /** The texture magnification function is used when the pixel being textured maps to an area less than or equal to one
      * texture element. It sets the texture magnification function to either TFM_MAG_NEAREST or TFM_MAG_BILINEAR. */
    NVSG_API void setMagFilterMode( TextureMagFilterMode mode    //!<  magnification filter mode to set
                                  );

    //! Get the texture magnification filter mode
    /** \return The magnification filter mode.  */
    NVSG_API TextureMagFilterMode getMagFilterMode( void ) const;

    //! Set the texture minification filter mode
    /** The texture minifying function is used whenever the pixel being textured maps to an area greater than one texture
      * element. There are three defined minifying functions. TFM_MIN_NEAREST uses the nearest texture element to compute
      * the texture value. TFM_MIN_BILINEAR uses the nearest four texture elements. TFM_MIN_TRILINEAR uses mipmaps. */
    NVSG_API void setMinFilterMode( TextureMinFilterMode mode    //!<  minification filter mode to set
                                  );

    //! Get the texture minification filter mode
    /** \return The minification filter mode. */
    NVSG_API TextureMinFilterMode getMinFilterMode( void ) const;

    //! Get the texture wrap mode
    /** The texture wrap mode can be either TWM_REPEAT (default), TWM_CLAMP, or TWM_MIRROR. Each of the up to three
      * texture dimensions may have it's own wrap mode.
      * \return TextureWrapMode of the texture target (TCA_S, TCA_T, or TCA_R) */
    NVSG_API TextureWrapMode getWrapMode( TexCoordAxis target   //!<  Target dimension to get wrap mode for
                                        ) const;

    //! Set the texture wrap mode
    /** The texture wrap mode can be set independently for each of up to three texture dimensions to either TWM_REPEAT
      * (default), TWM_CLAMP or TWM_MIRROR.
      * \note If the underlying hardware doesn't support mirroring textures, it falls back to the repeat mode. */
    NVSG_API void setWrapMode( TexCoordAxis target  //!<  Target dimension to set wrap mode for
                             , TextureWrapMode mode      //!<  Wrap mode to set
                             );

    //! Get texture type.
    /** \note Currently only diffuse textures are supported.
      * \returns  The type of the embedded texture object. */
    NVSG_API TextureType getType( void ) const;

    //! Set texture type.
    /** Use this to specify the texture type.
      * \note Only diffuse textures are currently supported. */
    NVSG_API void setType( TextureType tt //!< Texture type.
                         );

    //! Get texture matrix.
    /** The texture matrix is composed of the cropping and the rotation matrix.
      * \returns  The matrix used to transform the embedded texture object. */
    NVSG_API nvmath::Mat44f getMatrix( void ) const;

    //! Get orientation.
    /** Get the rotation part of the texture transformation.
      * \returns  A quaternion that describes the rotational part. */
    NVSG_API const nvmath::Quatf & getOrientation( void ) const;

    //! Set orientation.
    /** Sets the orientation of the texture transformation. */
    NVSG_API void setOrientation( const nvmath::Quatf & orientation //!< Rotational part of the texture transformation.
                                );

    //! Get scaling.
    /** Get the scaling part of the texture transformation.
      * \returns A three component vector that describes the scaling in u, v, and w.  */
    NVSG_API const nvmath::Vec3f & getScaling( void ) const;

    //! Set scaling.
    /** Sets the scaling of the texture transformation. */
    NVSG_API void setScaling( const nvmath::Vec3f &scale //!< Scaling factors in u, v, w.
                            );

    //! Get translation. 
    /** Get the translation part of the texture transformation.
      * \returns  The translation in u, v, and w  */
    NVSG_API const nvmath::Vec3f & getTranslation( void ) const;

    //! Set translation.
    /** Sets the translation of the texture transformation. */
    NVSG_API void setTranslation( const nvmath::Vec3f &trans //!< Translation in u, v, w.
                                );

    //! Determine if a texture lookup is required with the very next rendering step.
    /** \returns \c true if a texture lookup is required with the very next rendering step,
      * \c false otherwise. */
    NVSG_API bool isLookupRequired( void ) const;

    //! Set the texture coordinate generation of the s- and t-axis to TGM_SPHERE_MAP.
    NVSG_API void setTexGenEnvironmentMapping( bool sAxis = true  //!< enable TGM_SPHERE_MAP on the s-axis
                                             , bool tAxis = true  //!< enable TGM_SPHERE_MAP on the t-axis
                                             );

    //! Set the texture coordinate generation of \a axis to object or eye linear.
    NVSG_API void setTexGenLinearMapping( TexCoordAxis            axis    //!< select the axis to enable
                                        , TexGenLinearMode        mode    //!< select the linear mode to set
                                        , const nvmath::Plane3f & plane   //!< set the plane equation
                                        );

    //! Disable the texture coordinate generation of \a axis.
    NVSG_API void disableTexGen( TexCoordAxis axis  //!< select the axis to disable
                               );

    //! Get the texture coordinate generation mode.
    NVSG_API TexGenMode getTexGenMode( TexCoordAxis axis  //!< select the axis to query
                                     ) const;

    //! Get the base plane for linear texture coordinate generation.
    NVSG_API const nvmath::Plane3f &getTexGenPlane( TexCoordAxis axis   //!< select the axis to query
                                                  ) const;

  protected:
    //! Default constructor
    NVSG_API TextureAttributeItem( void );

    //! Partial Constructor
    NVSG_API TextureAttributeItem( const Object &rhs );
    
    //! Copy constructor
    NVSG_API TextureAttributeItem( const TextureAttributeItem& rhs );
    
    // prohibit explicit creation on stack by making the destructor protected
    NVSG_API virtual ~TextureAttributeItem();


  private:
    void addOwner( const TextureAttribute * pAttrib );
    void removeOwner( const TextureAttribute * pAttrib );

  private:
    std::vector<const TextureAttribute*>  m_owners;       // owners of this texture attribute item are collected here

    nvutil::RCPtr<TextureAttributeItemData>   m_sharedData; // reference to embedded data that might be shared between different items
    mutable const Texture * m_texture; // device dependent texture object
    mutable bool      m_lookupRequired; // assigns wether a texture lookup is required with the very next rendering step
  };

  //! Texture attribute
  /** This maintains bindings between abstract texture units and \c TextureAttributeItem objects.
    * For this purpose it provides an interface to bind a certain \c TextureAttributeItem to a 
    * specified texture unit or to unbind a \c TextureAttributeItem from a certain texture unit. 
    * It also provides an interface to iterate through all bounded \c TextureAttributeItem objects.
    * \par
    * As a \c StateAttribute, a \c TextureAttribute can be added to a \c StateSet of a certain 
    * \c GeoNode for texturing purposes. */ 
  class TextureAttribute : public StateAttribute
  {
    public:

      enum { 
        INVALID_ITEM = 0xFFFFFFFF  
      };

      //! Create a TextureAttribute.
      /** \returns a constant pointer to a TextureAttribute. */
      NVSG_API static const TextureAttribute * create( void );

      //! Create a TextureAttribute by copying from a StateAttribute.
      /** \returns a constant pointer to a TextureAttribute. */
      NVSG_API static const TextureAttribute * createFromBase( const StateAttribute &rhs );

      //! Get a clone of this texture attribute.
      /** \returns a constant pointer to a TextureAttribute. */
      NVSG_API virtual const TextureAttribute * clone( void ) const;

      //! Determine whether the data of this object is shared.
      /** \returns  true if the object's data is shared, otherwise false.  */
      NVSG_API virtual bool isDataShared( void ) const;

      //! Get the DataID of the data of this object.
      /** \returns  DataID of the object's data.  */
      NVSG_API virtual DataID getDataID( void ) const;

      //! Bind texture attribute item.
      /** This binds the texture attribute item pointed to by \a item to the texture 
        * unit specified by \a unit. A texture attribute that was bound to that
        * texture unit prior to this call, will be unbound. */
      NVSG_API void bindTextureAttributeItem( const TextureAttributeItem* item //!< Texture attribute item to be bound.
                                            , size_t unit //!< Zero based index of the texture unit the item should be bound to.               
                                            );
      
      //! Unbind texture attribute item.
      /** Unbind the texture attribute item currently bound to the texture unit 
        * specified by \a unit. Immediately after this call, no texture attribute 
        * item is bound to that unit.*/
      NVSG_API void unbindTextureAttributeItem( size_t unit //!< Zero based index of the texture unit.
                                              );

      //! Get first item.
      /** This is thought to initiate iteration of bounded texture attribute items. 
        * Use this in conjunction with \c getNextItem to iterate over all bounded 
        * texture attribute items.
        * \returns The zero based index of the first texture unit a texture attribute 
        * item is actually bound to or \c INVALID_ITEM if no texture attribute item is 
        * currently bound. */
      NVSG_API size_t getFirstItem( void ) const;

      //! Get next item.
      /** This is thought to pursue iteration of bounded texture attribute items.
        * \note An iteration must be started by a call to \c getFirstItem to ensure that 
        * the iteration starts at the very first texture unit a texture attribute item is
        * actually bound to.
        * \returns The zero based index of the very next texture unit a texture attribute 
        * item is actually bound to or \c INVALID_ITEM if no further texture attribute item
        * is currently bound. */
      NVSG_API size_t getNextItem( void ) const;

      //! Get number of item.
      /** \returns The number of texture attribute items currently bound. 
        * \note Do not use this to iterate through the items. If you want to iterate use
        * this construct:
        * \code
        * for ( size_t tu = pAttrib->getFirstItem()
        *     ; tu != TextureAttribute::INVALID_ITEM
        *     ; tu = pAttrib->getNextItem())
        * {
        * }
        * \endcode */
      NVSG_API size_t getNumberOfItems( void ) const;

      //! Get texture attribute item. 
      /** \returns The texture attribute item that is currently bound to the texture unit 
        * specified by \a unit or a null pointer if no texture attribute item is currently
        * bound to texture unit \a unit. */
      NVSG_API const TextureAttributeItem * getTextureAttributeItem( size_t unit //!< Zero based index of queried texture unit.
                                                             ) const;

    protected:
      //! Default Constructor
      NVSG_API TextureAttribute( void );

      //! Partial Constructor
      NVSG_API TextureAttribute( const StateAttribute &rhs );
      
      //! Copy constructor
      NVSG_API TextureAttribute( const TextureAttribute& rhs );

      //! Protected destructor to prevent explicit creation on stack.
      NVSG_API virtual ~TextureAttribute( void );

    private:
      std::map<size_t, const TextureAttributeItem*> m_bindings; // item bindings 
      mutable size_t                     m_currItem; // used with iterating items
  };


  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - TextureAttribute
  inline size_t TextureAttribute::getFirstItem( void ) const
  {
    __TRACE();
    m_currItem = !m_bindings.empty() ? m_bindings.begin()->first : INVALID_ITEM;
    return m_currItem;
  }

  inline size_t TextureAttribute::getNextItem( void ) const
  {
    __TRACE();
    if (m_currItem!=INVALID_ITEM)
    {
      std::map<size_t, const TextureAttributeItem*>::const_iterator ci = m_bindings.upper_bound(m_currItem);
      m_currItem = ci!=m_bindings.end() ? ci->first : INVALID_ITEM;
    }
    return m_currItem;
  }

  inline size_t TextureAttribute::getNumberOfItems( void ) const
  {
    __TRACE();
    return m_bindings.size();
  }

  inline const TextureAttributeItem* TextureAttribute::getTextureAttributeItem( size_t unit ) const
  {
    __TRACE();
    std::map<size_t, const TextureAttributeItem*>::const_iterator ci = m_bindings.find(unit);
    return ci!=m_bindings.end() ? (*ci).second : NULL;
  }

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - TextureAttributeItem
  inline const std::string& TextureAttributeItem::getFileName() const
  {
    __TRACE();
    return ( m_sharedData->m_fileName );
  }

  inline void TextureAttributeItem::setFileName( const std::string& fname )
  {
    __TRACE();

    if ( fname == m_sharedData->m_fileName )
    { // nothing to do if filename doesn't change
      return; 
    }

    // nvutil::writeAccess performs a deep copy if necessary
    nvutil::writeAccess( m_sharedData );   
    
    // unreference old texture object
    setTexture(NULL);

    m_sharedData->m_fileName = fname;
    m_lookupRequired = true; // new image file requires new texture upload
  }

  inline const Texture * TextureAttributeItem::getTexture( void ) const
  {
    __TRACE();
    return( m_texture );
  }

  inline void TextureAttributeItem::setTexture( const Texture *p ) const
  {
    __TRACE();
    if ( m_texture )
    {
      m_texture->removeRef();
    }
    m_texture = p;
    if ( m_texture )
    {
      m_texture->addRef();
    }
    m_lookupRequired = ! m_texture;
  }

  inline void TextureAttributeItem::setBorderColor( const nvmath::Vec4f &color )
  {
    m_sharedData->m_borderColor = color;
  }

  inline const nvmath::Vec4f& TextureAttributeItem::getBorderColor( void ) const
  {
    return( m_sharedData->m_borderColor );
  }

  inline TextureEnvMode TextureAttributeItem::getEnvMode( void ) const
  {
    __TRACE();
    return( m_sharedData->m_envMode );
  }

  inline void TextureAttributeItem::setEnvMode( TextureEnvMode mode )
  {
    __TRACE();
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if neccessary
    m_sharedData->m_envMode = mode;
  }

  inline TextureWrapMode TextureAttributeItem::getWrapMode( TexCoordAxis target ) const
  {
    __TRACE();
    return( m_sharedData->m_wrapMode[target] );
  }

  inline void TextureAttributeItem::setWrapMode( TexCoordAxis target, TextureWrapMode mode )
  {
    __TRACE();
    m_sharedData->m_wrapMode[target] = mode;
  }

  inline TextureMinFilterMode TextureAttributeItem::getMinFilterMode( void ) const
  {
    __TRACE();
    return( m_sharedData->m_minFilterMode );
  }

  inline void TextureAttributeItem::setMinFilterMode( TextureMinFilterMode mode )
  {
    __TRACE();
    m_sharedData->m_minFilterMode = mode;
  }

  inline TextureMagFilterMode TextureAttributeItem::getMagFilterMode( void ) const
  {
    __TRACE();
    return( m_sharedData->m_magFilterMode );
  }

  inline void TextureAttributeItem::setMagFilterMode( TextureMagFilterMode mode )
  {
    __TRACE();
    m_sharedData->m_magFilterMode = mode;
  }

  inline  TextureType TextureAttributeItem::getType( void ) const
  {
    __TRACE();
    return( m_sharedData->m_type );
  }

  inline void TextureAttributeItem::setType( TextureType tt )
  {
    __TRACE();
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if neccessary
    if (tt == TT_TRANSPARENT)
    {
      std::vector<const TextureAttribute*>::const_iterator ci = m_owners.begin();
      while (ci != m_owners.end())
      {
        TextureAttribute * pAttrib = beginEdit(*ci);
        pAttrib->invalidateTransparentTextureContainment();
        endEdit(pAttrib);

        ci++;
      }
    }
    m_sharedData->m_type = tt;
  }

  inline  nvmath::Mat44f  TextureAttributeItem::getMatrix( void ) const
  {
    __TRACE();
    return( m_sharedData->m_trafo.getMatrix() );
  }

  inline const nvmath::Quatf & TextureAttributeItem::getOrientation( void ) const
  {
    __TRACE();
    return( m_sharedData->m_trafo.getOrientation() );
  }

  inline void TextureAttributeItem::setOrientation( const nvmath::Quatf &orientation )
  {
    __TRACE();
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if neccessary
    m_sharedData->m_trafo.setOrientation(orientation);
  }

  inline const nvmath::Vec3f & TextureAttributeItem::getScaling( void ) const
  {
    __TRACE();
    return( m_sharedData->m_trafo.getScaling() );
  }

  inline void TextureAttributeItem::setScaling( const nvmath::Vec3f &scale )
  {
    __TRACE();
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if neccessary
    m_sharedData->m_trafo.setScaling(scale);
  }

  inline const nvmath::Vec3f & TextureAttributeItem::getTranslation( void ) const
  {
    __TRACE();
    return( m_sharedData->m_trafo.getTranslation() );
  }

  inline void TextureAttributeItem::setTranslation( const nvmath::Vec3f &trans )
  {
    __TRACE();
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if neccessary
    //  just translate in x and y, don't translate in z
    m_sharedData->m_trafo.setTranslation(trans);
  }

  inline bool TextureAttributeItem::isLookupRequired( void ) const
  {
    __TRACE();
    __ASSERT((m_lookupRequired==false)==(m_texture!=NULL));
    return m_lookupRequired || !m_texture->getHandle();
  }

  inline void TextureAttributeItem::setTexGenEnvironmentMapping( bool sAxis, bool tAxis )
  {
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if necessary
    if ( sAxis )
    {
      m_sharedData->m_texGenMode[TCA_S] = TGM_SPHERE_MAP;
    }
    if ( tAxis )
    {
      m_sharedData->m_texGenMode[TCA_T] = TGM_SPHERE_MAP;
    }
  }

  inline void TextureAttributeItem::setTexGenLinearMapping( TexCoordAxis axis, TexGenLinearMode mode, const nvmath::Plane3f &plane )
  {
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if necessary
    m_sharedData->m_texGenMode[axis] = (TexGenMode) mode;
    m_sharedData->m_texGenPlane[axis] = plane;
  }

  inline void TextureAttributeItem::disableTexGen( TexCoordAxis axis )
  {
    nvutil::writeAccess(m_sharedData);   //  nvutil::writeAccess performs a deep copy if necessary
    m_sharedData->m_texGenMode[axis] = TGM_OFF;
  }

  inline TexGenMode TextureAttributeItem::getTexGenMode( TexCoordAxis axis ) const
  {
    return( m_sharedData->m_texGenMode[axis] );
  }

  inline const nvmath::Plane3f &TextureAttributeItem::getTexGenPlane( TexCoordAxis axis ) const
  {
    return( m_sharedData->m_texGenPlane[axis] );
  }
}
