// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include  <map>
#include  "nvsg/CullData.h"
#include  "nvtraverser/ModelViewTraverser.h"

// forward declarations
namespace nvsg
{
  class Drawable;
  class Node;
  class GeoNode;
  class Group;
  class LOD;
  class Quads;
  class QuadStrips;
  class Switch;
  class Transform;
  class Triangles;
  class TriStrips;
  class SimpleCamera;
  class ViewState;
  class Scene;
}

namespace nvtraverser
{
  //! Base class for different culling implementations.
  /** Use this class as the base class for cull traversers.
  /** \note A CullTraverser doesn't change anything in the scene graph, but might be used by a modifying traverser.
    * therefore, it gets the readOnly flag as a parameter to pass to ModelViewTraverser. */
  class CullTraverser : public ModelViewTraverser
  {
    public:

    protected:
      //! Constructor (protected to prevent explicit instantiation)
      NVSG_API CullTraverser();

      //! Destructor
      NVSG_API virtual ~CullTraverser(void);

      //! determine the \c CullCode of a \c Drawable.
      /** The default implementation returns CC_IN.
        * \return \c CullCode for \c Drawable \a p. */
      NVSG_API virtual nvsg::CullCode cullCode( const nvsg::Drawable *p   //!< pointer to \c Drawable for determining the \c CullCode
                                              ) const;

      //! Determine the \c CullCode of a Node.
      /** The default implementaion returns CC_IN.
        * \return \c CullCode for \c Node \a p. */
      NVSG_API virtual nvsg::CullCode cullCode( const nvsg::Node *p   //!< pointer to \c Node for determining the \c CullCode
                                              ) const;

      //! Determine the CullCode of each Drawable of \a p.
      /** If the GeoNode \a p is partially visible, the CullCode of each of it's Drawable is determined and then
        * traversed. */
      NVSG_API virtual void handleGeoNode( const nvsg::GeoNode * p  //!<  GeoNode to traverse
                                         );

      //! Determine the CullCode of all children of \a p.
      /** If the Group \a p is partially visible, the CullCode of each of it's children is determined and then
        * traversed. */
      NVSG_API virtual void handleGroup( const nvsg::Group * p    //!<  Group to traverse
                                       );

      //! Determine the CullCode of all children of \a p.
      /** If the LOD \a p is partially visible, the CullCode of each of it's children is determined and then traversed. */
      NVSG_API virtual void handleLOD( const nvsg::LOD * p    //!<  LOD to traverse
                                     );

      //! Determine the CullCode of all children of \a p.
      /** If the Switch \a p is partially visible, the CullCode of each of it's children is determined and then
        * traversed. */
      NVSG_API virtual void handleSwitch( const nvsg::Switch * p  //!<  Switch to traverse
                                        );

      //! Determine the CullCode of all children of \a p.
      /** If the Transform \a p is partially visible, the CullCode of each of it's children is determined and then
        * traversed. */
      NVSG_API virtual void handleTransform( const nvsg::Transform * p  //!<  Transform to traverse
                                           );

      //! Determine the CullCode of the Quads \a p.
      /** The CullCode is determined and added to the Object to CullData map on top of the ObjectCullDataMapStack. */
      NVSG_API virtual void handleQuads( const nvsg::Quads * p    //!<  Quads to determine CullData for
                                       );

      //! Determine the CullCode of the QuadStrips \a p.
      /** The CullCode is determined and added to the Object to CullData map on top of the ObjectCullDataMapStack. */
      NVSG_API virtual void handleQuadStrips( const nvsg::QuadStrips * p    //!<  QuadStrips for determining the CullData 
                                            );

      //! Determine the CullCode of the Triangles \a p.
      /** The CullCode is determined and added to the Object to CullData map on top of the ObjectCullDataMapStack. */
      NVSG_API virtual void handleTriangles( const nvsg::Triangles * p    //!<  Triangles for determining the CullData 
                                           );

      //! Determine the CullCode of the TriStrips \a p.
      /** The CullCode is determined and added to the Object to CullData map on top of the ObjectCullDataMapStack. */
      NVSG_API virtual void handleTriStrips( const nvsg::TriStrips * p    //!<  TriStrips for determining the CullData 
                                           );

      //! Set the CullCode of the LightSource \a p.
      /** All lights are always considered to be in. */
      NVSG_API virtual void traverseLight(
        const nvsg::LightSource *p //!< Points to the LightSource currently traversed.
      );

    protected:
      // doApply override
      NVSG_API virtual void doApply( const nvsg::ViewState *pViewState, const nvsg::Scene *pScene );

    private:
      //  hold a pointer to a map of Node pointers to CullData pointers on each level
      std::stack<std::map<const nvsg::Object*,nvsg::CullData*>*>    m_objectCullMapStack;
      nvsg::CullData                                              * m_cullTree;
  };
}
