// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#include <list>

#include "nvsgcommon.h"

#include "nvtraverser/Traverser.h"

namespace nvtraverser
{
  //! Traverser that combines multiple GeoNodes under one Transform to a single one.
  /** All geometries of all GeoNodes unter one Transform are reparented to a single GeoNode. */
  class GeoNodeUnifyTraverser : public Traverser
  {
    public:
      //! Constructor
      NVSG_API GeoNodeUnifyTraverser( void );

    public:
      //! Get the 'ignore names' flag.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored.
        * \return true if the names will be ignored, otherwise false */
      NVSG_API bool getIgnoreNames( void ) const;

      //! Set the 'ignore names' flags.
      /** If the 'ignore names' flag is set, the names of the GeoNodes are ignored. */
      NVSG_API void setIgnoreNames( bool ignore   //!<  set true to ignore names on comparison
                                  );

    protected:
      //! Protected destructor to prevent instantiation of a GeoNodeUnifyTraverser on stack.
      NVSG_API virtual ~GeoNodeUnifyTraverser( void );

      //! Routine to handle a \link nvsg::GeoNode GeoNode \endlink while traversing the scene graph.
      /** This routine just records all GeoNodes beneath a Group for reparenting. */
      NVSG_API virtual void handleGeoNode( const nvsg::GeoNode * p //!< Points to the currently visited GeoNode object.
                                         );
                                          
      //! This function will be called from the framework immediately before a Group is to be traversed.
      /** Before traversing the Group, a list of the contained GeoNodes is created. */
      NVSG_API virtual void preTraverseGroup( const nvsg::Group *p  //!< Points to the Group that will be traversed immediately after this call.
                                            );

      //! This function will be called from the framework immediately after a Group has been traversed.
      /** After traversing the Group, all it's children that are GeoNodes are combined. */
      NVSG_API virtual void postTraverseGroup( const nvsg::Group *p  //!< Points to the Group that was traversed immediately before this call.
                                             );

    private:
      bool                                          m_ignoreNames;
      std::stack<std::list<const nvsg::GeoNode*>*>  m_geoNodes;
  };

  inline bool GeoNodeUnifyTraverser::getIgnoreNames( void ) const
  {
    return( m_ignoreNames );
  }

  inline void GeoNodeUnifyTraverser::setIgnoreNames( bool ignore )
  {
    m_ignoreNames = ignore;
  }

}
