// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Drawable.h"
#include "nvsg/Path.h"
#include "nvtraverser/ModelViewTraverser.h"

namespace nvsg
{
  class Camera;
}

namespace nvtraverser
{
  //! Intersection class.
  /** This class is used to store intersection results generated by the RayIntersectTraverser. 
    * It should only be used with classes derived from RayIntersectTraveser because some 
    * information like the distance makes sense only when knowing the ray data. */
  class Intersection
  {
    public:
      //! Contructor.
      /** This contructor creates an intersection object from the given data.
        * It also performs an addRef() on the given path and drawable. */
      NVSG_API Intersection( const nvsg::Path * pPath          //!< Path to the intersected drawable.
                           , const nvsg::Drawable * pDrawable  //!< Intersected drawable.
                           , const nvmath::Vec3f & isp         //!< Intersection point on the drawable.
                           , const float & dist                //!< Distance from the ray origin to the drawable.
                           );

      //! Copy constructor.
      /** Creates a new Intersection object from the given one. */
      NVSG_API Intersection( const Intersection & rhs    //!< Intersection object to copy from.
                           );

      //! Default destructor.
      NVSG_API virtual ~Intersection(void);

      //! Assignment operator.
      NVSG_API Intersection & operator = (const Intersection & rhs);

      //! Get the Path to the intersected drawable.
      /** \return Path to the drawable. */
      NVSG_API const nvsg::Path * getPath() const;

      //! Get the intersected drawable.
      /** \return Intersected drawable. */
      NVSG_API const nvsg::Drawable * getDrawable() const;

      //! Get the intersection point on the intersected drawable.
      /** \return Intersection Point. */
      NVSG_API const nvmath::Vec3f & getIsp() const;

      //! Get the distance from the ray origin to the intersection point.
      /** \return Distance from ray origin to intersection point. */
      NVSG_API const float & getDist() const;

    protected:
      //! Default contructor.
      /** The user has to use the other constructors! */
      NVSG_API Intersection(void){};

      //! Clone the whole object.
      /** Use this, for example, for the copy constructor. */
      NVSG_API void clone( const Intersection & rhs    //!< Object to clone.
                ); 
      
      const nvsg::Drawable  * m_pDrawable;    //!< Intersected drawable.
      const nvsg::Path      * m_pPath;        //!< Path to the intersected drawable.
      nvmath::Vec3f           m_isp;          //!< %Intersection point on the intersected drawable.
      float                   m_dist;         //!< Distance from the ray origin to the intersection point.
  };

  inline Intersection::~Intersection(void) 
  {
    if (m_pDrawable)
    {
      m_pDrawable->removeRef();
      m_pDrawable = NULL;
    }

    if (m_pPath)
    {
      m_pPath->removeRef();
      m_pPath = NULL;
    }
  }

  inline Intersection::Intersection (const nvsg::Path * pPath, const nvsg::Drawable * pDrawable, const nvmath::Vec3f & isp, const float & dist)
  {
    __ASSERT(pDrawable);
    __ASSERT(pPath);

    m_isp       = isp; 
    m_dist      = dist;
    
    m_pDrawable = pDrawable;
    m_pDrawable->addRef();

    m_pPath     = pPath;
    m_pPath->addRef();
  }

  inline Intersection::Intersection (const Intersection & rhs)
  {
    clone(rhs);
  }

  inline Intersection & Intersection::operator = (const Intersection & rhs) 
  {
    clone(rhs);
    return (*this);
  }

  inline void Intersection::clone(const Intersection & rhs)
  {
    m_isp       = rhs.m_isp; 
    m_dist      = rhs.m_dist; 
    
    m_pDrawable = rhs.m_pDrawable; 
    m_pDrawable->addRef();
    
    m_pPath     = rhs.m_pPath; 
    m_pPath->addRef();
  }

  inline const nvmath::Vec3f & Intersection::getIsp() const
  {
  return m_isp;
  }

  inline const float & Intersection::getDist() const
  {
    return m_dist;
  }

  inline const nvsg::Drawable * Intersection::getDrawable() const
  {
    return m_pDrawable;
  }

  inline const nvsg::Path * Intersection::getPath() const
  {
    return m_pPath;
  }


  //! RayIntersectTraverser
  /** A \a RayIntersectTraverser is a specialized traverser that calculates 
    * the intersections of a given ray that passes through the scene.
    * \note If you add new geometry objects classes to NVSG you must also derive from this class and 
    *  add an operator for this object to handle the intersection calculation.
    * \note If you create a special kind of transformation node you also have to 
    *  take care of the transformation stack in this class. */
  class RayIntersectTraverser : public ModelViewTraverser
  {
    public:
      //! Default constructor.
      NVSG_API RayIntersectTraverser(void);

      //! Release all resources.
      /** By releasing all resources we can use this traverser for the 
        * next intersection calculation. */
      NVSG_API void release();

      //! Set the ray for the intersection test.
      /** Sets a world-space ray along which to pick. The ray is defined as a world space 
        * \a origin point and \a direction vector. The direction vector must be normalized. */
      NVSG_API void setRay( const nvmath::Vec3f &origin      //!< World space origin
                          , const nvmath::Vec3f &dir         //!< Direction of the ray - must be normalized!
                          );

      //! Enable/disable camera plane clipping. 
      /** Enabling camera plane clipping forces the traverser to calculate only 
        * intersections that lie in between the camera's clip planes. */
      NVSG_API void setCamClipping( bool b    //!< true = enable clipping 
                                  );

      //! Get the nearest intersection.
      /** \returns The nearest intersection regarding the rays origin. */
      NVSG_API const Intersection & getNearest( void ) const;
      
      //! Get intersections.
      /** \returns All intersections along the ray. */
      NVSG_API const Intersection * getIntersections( void ) const;
      
      //! Get number of intersections.
      /** \returns The number of intersections along the ray. */
      NVSG_API size_t getNumberOfIntersections( void ) const;
     
      //! Enable / disable back-face culling.
      /** Take back-faces culling into account. */
      NVSG_API void setBackFaceCulling( bool enable    //! true enables back-face culling
                                      );

      //! Check if back-face culling is enabled
      /** Check if the ray intersection also test on back-faces. 
        * \return true, if back-face culling is enabled, meaning no back-face will be returned.*/
      NVSG_API bool isBackFaceCullingEnabled();

    protected:
      //! Default destructor.
      NVSG_API virtual ~RayIntersectTraverser(void);

      //! Apply the traverser to the scene.
      /** Start the intersection calculation of the given 
        * ray with the scene objects. */
      NVSG_API virtual void  doApply( const nvsg::ViewState *pViewState   //!< Current view state
                                    , const nvsg::Scene *pScene           //!< Scene to intersect the ray with
                                    );

      // Nodes in the tree:
      NVSG_API virtual void handleGeoNode( const nvsg::GeoNode *p );
      
      // Groups in the tree:
      NVSG_API virtual void handleGroup( const nvsg::Group * p);
      NVSG_API virtual void handleLOD( const nvsg::LOD * p);
      NVSG_API virtual void handleSwitch( const nvsg::Switch * p);
      NVSG_API virtual void handleTransform( const nvsg::Transform * p);

      // Other operators:
      NVSG_API virtual void handleQuads( const nvsg::Quads *p );
      NVSG_API virtual void handleQuadStrips( const nvsg::QuadStrips *p );
      NVSG_API virtual void handleTriangles( const nvsg::Triangles *p );
      NVSG_API virtual void handleTriStrips( const nvsg::TriStrips *p );

      //! Handles actions to take between transform stack adjustment and traversal.
      /** Adjusts the ray origin and direction and the camera direction in model space.
        * \return result of the base class (ModelViewTraverser). */
      NVSG_API virtual bool preTraverseTransform( 
        const nvsg::Transform *p          //!< Transform node to traverse next 
      );

      //! Handles stuff to do between traversal and transform stack adjustment.
      /** Readjusts the ray origin and direction and the camera direction in model space. */
      NVSG_API virtual void postTraverseTransform(
        const nvsg::Transform *p          //!< Transform node that was traversed immediately before this call. 
      );

      //! %Intersection type (bounding sphere).
      /** These are the possible types of intersections between the ray 
        * and the boundingsphere of an object in the scene. These types 
        * are used for the preintersection calculation with the objects 
        * bounding sphere. Regarding the intersection type we can skip 
        * objects. */
      enum ISType { _NONE = 0       //!< No intersection
                  , _BSTANGENT      //!< Tangential intersection
                  , _BSNEAR         //!< Intersection with the front part of the bounding sphere
                  , _BSFAR          //!< Intersection with the back part of the bounding sphere
                  , _BSNEARFAR      //!< Intersection with the front and back part of the bounding sphere 
                  };

      //! Check the current ray for intersection with a \c Sphere3f.
      /** \return true, if the ray intersects \a sphere, otherwise false. */
      NVSG_API bool checkIntersection( const nvmath::Sphere3f &sphere
                                    );

      //! Intersect with triangle.
      /** This method tells if it intersects the triangle or not. 
        * It also calculates the intersection point between triangle and ray and the distance 
        * of the intersection from the origin of the ray.
        * \returns true: intersection in front of the ray origin, false: no such intersection */
      NVSG_API bool intersectTriangle( const nvmath::Vec3f & v0  //!< First vertex of the triangle
                                    , const nvmath::Vec3f & v1  //!< Second vertex of the triangle
                                    , const nvmath::Vec3f & v2  //!< Third vertex of the triangle
                                    , nvmath::Vec3f & isp       //!< Result: intersection point (world space)
                                    , float & dist              //!< Result: distance from camera to intersection
                                    );    

      //! store the intersection point \a isp.
      /** If the point isn't clipped, it is stored to the intersection list.  */
      NVSG_API void storeIntersection( const nvsg::GeoSet *p     //!< GeoSet that holds the intersection
                                    , const nvmath::Vec3f &isp    //!< intersection point
                                    , float dist                 //!<  distance from camera to intersection
                                    );

#ifdef _DEBUG
      /** Dump intersection map - for Debug*/
      void dumpIntersectionList();
#endif
       
      static const nvmath::Vec3f _RAY_ORIGIN_DEFAULT;    //!< Default value: (0.f, 0.f,  0.f)
      static const nvmath::Vec3f _RAY_DIRECTION_DEFAULT; //!< Default value: (0.f, 0.f, -1.f)

      bool                        m_backFaceCulling;  //!< true: consider backface culling
      bool                        m_ccw;              //!< true: winding is supposed to be counterclockwise
      bool                        m_camClipping;      //!< true: use camera far/near clipping planes
      nvmath::Vec3f               m_rayOrigin;        //!< Origin point of the ray - default (0,0,0)
      nvmath::Vec3f               m_rayDir;           //!< Direction of the ray - default (0,0,-1)
      const nvsg::Camera              * m_camera;           //!< Camera of the scene that came in via apply (we ref/unref it)
      std::stack<nvmath::Vec3f>   m_msRayOrigin;      //!< stack of Origin point of the ray in model space
      std::stack<nvmath::Vec3f>   m_msRayDir;         //!< stack of Direction of the ray in model space
      std::stack<nvmath::Vec3f>   m_msCamDir;         //!< stack of camera direction in model space
      std::vector<Intersection>   m_intersectionList; //!< Intersection list containing all intersections.
      size_t                      m_nearestIntIdx;    //!< Index of the nearest z intersection.
      nvsg::Path                * m_curPath;          //!< Current path.
      size_t                      m_currentFrame;     //!< This frame is to be used in an animation sequence.
      float                       m_scaleFactor;      //!< Adjusts distances by the scalings of the Transform nodes
  };

  inline void RayIntersectTraverser::setCamClipping(bool flag)
  {
    m_camClipping = flag;
  }

  inline size_t RayIntersectTraverser::getNumberOfIntersections() const
  {
    return m_intersectionList.size();
  }

  inline const Intersection & RayIntersectTraverser::getNearest() const
  {
    __ASSERT(!m_intersectionList.empty());
    return m_intersectionList[m_nearestIntIdx];
  }
   
  inline const Intersection * RayIntersectTraverser::getIntersections() const
  {
    __ASSERT(!m_intersectionList.empty());
    return &m_intersectionList[0];
  }

  inline void RayIntersectTraverser::setBackFaceCulling( bool enable)
  {
    m_backFaceCulling = enable;
  }
      
  inline bool RayIntersectTraverser::isBackFaceCullingEnabled()
  {
    return m_backFaceCulling;
  }
}
