// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvtraverser/ModelViewTraverser.h"
#include "nvsg/DAL.h"
#include "nvutil/Incarnation.h"

namespace nvsg
{
  class StateAttribute;
}

namespace nvtraverser
{
// renderer type IDs
enum 
{
  RT_GL  = 0x00010000
, RT_DX8 = 0x00020000
, RT_DX9 = 0x00040000
};

//! Pure virtual render traverser class
/** Use this pure virtual class as a base class to derive your own 
  * render traverser. Render traverser are not allowed to apply changes 
  * to the given scenegraph.
  * This base class containes the model matrix and inverse model matrix stack.
  * This is very handy, because you do not have to ask e.g. OpenGL which 
  * would be normally very expensive. So your special render traverser has to
  * take care of these stacks.
  * \note The render traverser object must not survive the render area it works on.
      For e.g. RenderArea and GLTraverser make sure to make the render area 
      current before destroying the traverser.*/
class RenderTraverser : public ModelViewTraverser
                      , public nvsg::DALDataCreator
{
  public:

    enum JitterTechnique
    {
      JT_ACCUM        = 0x01,   //!< Render to accumulation buffer.
      JT_READ_PIXELS  = 0x02,   //!< Render to backbuffer and use e.g. OpenGL read pixels to read it back.
      JT_PBUFFER      = 0x04    //!< Render to PBuffer.
    };    

    enum RenderTechnique
    {
      RT_POINTS       = 0x01,   //!< Render scene as points only.
      RT_WIREFRAME    = 0x02,   //!< Render scene as wireframe only.
      RT_SHADED       = 0x04,   //!< Render scene in simple shading mode.
      RT_TEXTURED     = 0x10,   //!< Render scene in simple shading mode plus textures.
      RT_EFFECTED     = 0x20    //!< Render scene fully effected and shaded.
    };      

    enum TransparencyTechnique
    {
      TT_NONE               = 0x01,   //!< Ignore transparency, draw opaque. 
      TT_STIPPLED           = 0x02,   //!< Use stipple patterns to simulate transparency.
      TT_BLENDED            = 0x04,   //!< Use blending to render transparent objects.
      TT_ORDER_INDEPENDENT  = 0x08    //!< Use the "Order Independent Transparency" approach to render transparent objects.
    };          //!< Technique for rendering transparent objects.

    enum OITTechnique
    {
      OT_TEXTURE  = 0x01,   //!< Use textures for OIT.
      OT_PBUFFER  = 0x02    //!< Use PBuffers for OIT.
    };         

    enum GeometryCacheTechnique
    {
      GCT_NONE    = 0x01,   //!<  no geometry caching
      GCT_DL      = 0x02,   //!<  cache geometries in display lists
      GCT_VBO     = 0x04    //!<  cache geometries in vertex buffer objects
    };

    //! Get the current jitter technique.
    /** \return The currently used technique.*/
    NVSG_API virtual JitterTechnique getJitterTechnique( void ) const = 0;

    //! Test on supported jitter techniques.
    /** Use this method to test if this special jitter 
      * technique is currently available.
      * \return true - supported, false - not supported */
    NVSG_API virtual bool isJitterTechniqueSupported( JitterTechnique jt               //!< Jitter technique.
                                                    ) const = 0;

    //! Set the jitter technique the render should use.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setJitterTechnique( JitterTechnique jt                       //!< Jitter technique
                                            ) = 0;
    
    //! Get the current render technique.
    //** \return The currently used render technique. */
    NVSG_API virtual RenderTechnique getRenderTechnique( void ) const = 0;

    //! Test on supported render techhniques.
    /** Use this technique to test if this special render 
      * technique is currently available.
      * \return true - supported, false - not supported.*/
    NVSG_API virtual bool isRenderTechniqueSupported( RenderTechnique rt               //! Render technique.
                                                    ) const = 0;

    //! Set the render technique the render should use.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setRenderTechnique( RenderTechnique rt                       //! Render technique.
                                            ) = 0;
    
    //! Get the current transparency technique.
    //** \return The currently used transparency technique. */
    NVSG_API virtual TransparencyTechnique getTransparencyTechnique( void ) const = 0;

    //! Test on supported transparency techhniques.
    /** Use this technique to test if this special transparency 
      * technique is currently available.
      * \return true - supported, jalse - not supported.*/
    NVSG_API virtual bool isTransparencyTechniqueSupported( TransparencyTechnique tt   //! Transparency technique.
                                                          ) const = 0;

    //! Set the transparency technique the render should use.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setTransparencyTechnique( TransparencyTechnique tt           //! Transparency technique.
                                                  ) = 0;
    
    //! Get the current OIT (Order Independent Transparency) technique.
    //** \return The currently used OIT technique. */
    NVSG_API virtual OITTechnique getOITTechnique( void ) const = 0;

    //! Test on supported OIT techhniques.
    /** Use this technique to test if this special OIT 
      * technique is currently available.
      * \return true - supported, jalse - not supported.*/
    NVSG_API virtual bool isOITTechniqueSupported( OITTechnique ot                     //! OIT technique 
                                                  ) const = 0;

    //! Set the OIT technique the render should use.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setOITTechnique( OITTechnique ot                             //! OIT technique 
                                          ) = 0;

    //! Get maximum OIT passes.
    /** \return Maximum rendering passes to archive the desired effect. */ 
    NVSG_API virtual size_t getOITMaxPasses( void ) const = 0;
    //! Set the number of maxumum passes to archieve the desired transparency effect.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setOITMaxPasses( size_t maxPasses                            //! Maximum passes to render, even when threshold is not reached.
                                          ) = 0;

    //! Get the current threshold.
    /** \return threshold [0,1]*/
    NVSG_API virtual float getOITThreshold( void ) const = 0;

    //! Set the threshold.
    /** The thereshold tells the refinement step of the OIT algorithm if it needs 
      * another pass or if it is good enough to show the rendered result.
      * The value lies between 0 and 1 (0%-100%). The refinement step compares 
      * the percentage of drawn pixels with the threshold and stops if it is 
      * smaller than the user defined value. 
      * \return true - everything went fine, false - not set */     
    NVSG_API virtual bool setOITThreshold( float threshold  //!< Threshold [0,1]
                                          ) = 0;

    //! Get the current geoemtry caching technique.
    //** \return The currently used geometry caching technique. */
    NVSG_API virtual GeometryCacheTechnique getGeometryCacheTechnique( void ) const = 0;

    //! Test on supported geometry caching techhniques.
    /** \return true - supported, false - not supported.*/
    NVSG_API virtual bool isGeometryCacheTechniqueSupported( GeometryCacheTechnique gct  //! GeometryCache technique.
                                                            ) const = 0;

    //! Set the geometry caching technique the render should use.
    /** \return true - everything went fine, false - not set */
    NVSG_API virtual bool setGeometryCacheTechnique( GeometryCacheTechnique gct          //! GeometryCache technique.
                                                    ) = 0;

    //! Get the texture size limits.
    /** Textures are downscaled to those limits if they are larger. */
    NVSG_API virtual void getTextureSizeLimits( size_t &width     //!<  texture width limit
                                              , size_t &height    //!<  texture height limit
                                              ) const = 0;

    //! Set the texture size limits.
    /** Textures are downscaled to those limits if they are larger.
      * \note Setting width or height to 0 (zero), disables the respective texture size limit. */
    NVSG_API virtual void setTextureSizeLimits( size_t width      //!<  texture width limit
                                              , size_t height     //!<  texture height limit
                                              ) = 0;

    //! Set the anti-aliasing mode for wire frame drawings.
    NVSG_API void  setLineAntiAliasing( bool onOff    //!<  enables/disables line anti-aliasing
                                      );

    //! Get the anti-aliasing mode for wire frame drawings.
    /** \returns \c true, if line anti-aliasing is on, otherwise \c false. */
    NVSG_API bool  getLineAntiAliasing( void ) const;

    //! Get the line width for wire frame drawings.
    /** \returns the line width for wire frame drawings. */
    NVSG_API float getLineWidth( void ) const;

    //! Set the line width for wire frame drawings.
    NVSG_API void  setLineWidth( float width          //!<  line width to set
                               );

    //! Set the anti-aliasing mode for point mode drawings.
    NVSG_API void  setPointAntiAliasing( bool onOff   //!<  enables/disables point anti-aliasing
                                       );

    //! Get the anti-aliasing mode for point mode drawings.
    /** \returns \c true, if point anti-aliasing is on, otherwise \c false. */
    NVSG_API bool  getPointAntiAliasing( void ) const;

    //! Get the point size for point mode drawings.
    /** \returns the point size for point mode drawings. */
    NVSG_API float getPointSize( void ) const;

    //! Set the point size for point mode drawings.
    NVSG_API void  setPointSize( float size           //!<  point size to set
                               );

  protected:
    //! Default constructor
    NVSG_API RenderTraverser(void);
    //! Default destructor
    NVSG_API virtual ~RenderTraverser(void);

    // Object handler overrides
    NVSG_API virtual void handleParallelCamera(const nvsg::ParallelCamera * p, const nvsg::Node * root);
    NVSG_API virtual void handlePerspectiveCamera(const nvsg::PerspectiveCamera * p, const nvsg::Node * root);
 
    // postTraverseGroup override:
    // pop lights off the light stack after a group has been visited
    NVSG_API virtual void postTraverseGroup(const nvsg::Group *p);

    // traverseLight override to handle a general LightSource
    NVSG_API virtual void traverseLight(const nvsg::LightSource * p);

    //! This function is called from the RenderTraverser framework whenever a LightSource is to be disabled.
    NVSG_API virtual void disableLightSource(const nvsg::LightSource *p) = 0;

    //! Get the zero based index of the current light.
    /** This index can be used to refer to the device dependent light in order
      * to switch the light on, when a specific light handler routine is called,
      * or off, after leaving the light's parent Group node. */
    NVSG_API size_t getCurrentLightIndex() const;
    
    //! Map between \c StateAttribute and \c Incarnation.
    std::map<const nvsg::StateAttribute*, nvutil::Incarnation> m_perStateAttributeIncarnationMap;

  private:
    void traverse(const nvsg::LightSource * p); // helper for general light source handling 
    void popLights( const nvsg::Group *p ); // pop lights from light stack after a group has been visited 

    std::stack<std::pair<const nvsg::Group *,const nvsg::LightSource *> > m_lightStack; // light stack
    size_t m_currentLightIndex;
    bool  m_lineAntiAliasing;
    float m_lineWidth;
    bool  m_pointAntiAliasing;
    float m_pointSize;
};

inline size_t RenderTraverser::getCurrentLightIndex() const 
{ 
  return m_currentLightIndex; 
}

inline void RenderTraverser::setLineAntiAliasing( bool onOff )
{
  m_lineAntiAliasing = onOff;
}

inline bool RenderTraverser::getLineAntiAliasing() const
{
  return( m_lineAntiAliasing );
}

inline float RenderTraverser::getLineWidth( void ) const
{
  return( m_lineWidth );
}

inline void RenderTraverser::setLineWidth( float width )
{
  m_lineWidth = width;
}

inline void RenderTraverser::setPointAntiAliasing( bool onOff )
{
  m_pointAntiAliasing = onOff;
}

inline bool RenderTraverser::getPointAntiAliasing() const
{
  return( m_pointAntiAliasing );
}

inline float RenderTraverser::getPointSize( void ) const
{
  return( m_pointSize );
}

inline void RenderTraverser::setPointSize( float size )
{
  m_pointSize = size;
}

}
