// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvui/CameraManipulator.h"

//  forward declarations
namespace nvmath
{
  class Trafo;
}
namespace nvsg
{
  template<typename T> class Animation;
}

namespace nvui
{  
  //! Animation Camera Manipulator Class
  /** This manipulator controls the camera while an \c Animation<Trafo> is playing. 
    * It converts any mouse input to camera control.
    * A typical usage of an AnimationCamManip looks like this:
    * \code
    * animManip.setAnimation( &myTrafoAnimation );
    * animManip.attach( &myCamera );
    * while ( isAnimating )
    * {
    *   curFrame = ((++curFrame) % myTrafoAnimation.getNumberOfFrames()) ? curFrame : 0;
    *   animManip.setAnimationFrame( curFrame );
    *   animManip.apply();
    *   // initiate a redraw here
    * }
    * \endcode
    * For further details on using an AnimationCamManip, see the AnimationViewer tutorial.
    */
  class AnimationCamManip : public CameraManipulator
  {
    public:
      //! Default Constructor
      NVSG_API AnimationCamManip(void);

      //! Default Destructor
      /** If the AnimationCamManip holds an \c Animation<Trafo>, it is dereferenced. */
      NVSG_API virtual ~AnimationCamManip(void);

      //! Apply the \c Trafo of the current animation step to the camera.
      NVSG_API virtual void apply();

      //! Set the current animation frame.
      NVSG_API void  setAnimationFrame( size_t nFrame  //!< index of the animation frame
                                      );

      //! Set the \c Animation to use with this AnimationCamManip.
      /** If this AnimationCamManip currently holds an \c Animation<Trafo>, it's dereferenced.
        * \a pAnimation is set as the current \c Animation<Trafo>. If it's not NULL, it is referenced.
        * To "reset" the \c Animation, use NULL as the argument. */
      NVSG_API void  setAnimation( const nvsg::Animation<nvmath::Trafo> * pAnimation  //! pointer to an \c Animation
                                 ); 
      
    protected:
      const nvsg::Animation<nvmath::Trafo>  * m_animation;    //!< animation information
      size_t              m_currentFrame; //!< current animationFrame
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  // inlines
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  inline void AnimationCamManip::setAnimationFrame( size_t nFrame )
  {
    __TRACE();
    m_currentFrame = nFrame;
  }

} // namespace nvui
