// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once

#include "nvsgcommon.h"

#include "nvsg/Camera.h"

/** \file */

namespace nvui
{  

  //! Camera manipulator base class
  /** This class should be used as a base class for your own camera manipulators. 
    * A camera manipulater is an object that controls the camera in a very specific way 
    * Trackball, walk, flight, and animation are examples of camera controls.
    * You can attach a camera to the manipulator and the manipulator converts actions, such as
    * the movement of the mouse or button events, into camera data. 
    * You can also write manipulators that work with other kinds of inputs and input devices.*/
  class CameraManipulator
  {
    public:
      //! Default Constructor
      NVSG_API CameraManipulator(void);
      //! Destructor
      NVSG_API virtual ~CameraManipulator(void);

      //! Key and button definitions.
      /** These values are used for the button/key state handling of the manipulator. */
      enum
      { _NO_BTN_KEY         = 0x0000 //!< No key or button.
      , _LEFT_MOUSE_BTN     = 0x0001 //!< Left mouse button.
      , _MIDDLE_MOUSE_BTN   = 0x0002 //!< Middle mouse button.
      , _RIGHT_MOUSE_BTN    = 0x0004 //!< Right mouse button.
      , _SHIFT_KEY          = 0x0100 //!< Shift key.
      , _CONTROL_KEY        = 0x0200 //!< Control key.
      };
      
      //! Apply the states to the camera.
      /** Recalculate the camera with respect to the manipulator states. */
      NVSG_API virtual void apply() = 0;
      
      //! Handle key down.
      /** Adapt key pressed events (shift, ...) 
        * \note Calls keyReapeat()*/
      NVSG_API virtual void keyDown( UINT key  //!< Pressed key (e.g. _SHIFT_KEY or _CONTROL_KEY) 
                                    );
      //! Handle key up.
      /** Adapt key release events (shift, ...) */
      NVSG_API virtual void keyUp( UINT key   //!< Released key (e.g. _SHIFT_KEY or _CONTROL_KEY)
                                 );
      //! Handle key repeats.
      /** Adapt key repeats. */
      NVSG_API virtual void keyRepeat();

      //! Handle mouse button down.
      /** Adapt mouse button pressed events. */
      NVSG_API virtual void mouseButtonDown( UINT btn  //!< Pressed button (e.g. _LEFT_MOUSE_BTN|_MIDDLE_MOUSE_BTN)
                                           , int x     //!< Position of the mouse.
                                           , int y     //!< Position of the mouse.
                                           );
      //! Handle mouse button up.
      /** Adapt mouse button release events. */
      NVSG_API virtual void mouseButtonUp( UINT btn    //!< Released buttons (e.g. _LEFT_MOUSE_BTN|_MIDDLE_MOUSE_BTN)
                                         , int x       //!< Position of the mouse.
                                         , int y       //!< Position of the mouse.
                                         );
      //! Handle mouse motion.
      /** Adapt mouse motion events while mouse buttons are pressed. */
      NVSG_API virtual void mouseMotion( int x  //!< Position of the mouse.
                                       , int y  //!< Position of the mouse.
                                       );       
      //! Determine the key state.
      /** Return the current key state mask.
        * \return Concatenation of the current key states */
      UINT getKeyState( void ) const;

      //! Reset the current button state.
      void resetKeyState( void );

      //! Determine the button state.
      /** Return the current button state mask.
        * \return Concatenation of the current button states */
      UINT getButtonState( void ) const;

      //! Reset the current button state.
      void resetButtonState( void );

      //! Attach camera to manipulator.
      /** Attach a camera to this manipulator. Attach to NULL to detach.  */
      void attach( const nvsg::Camera * pCamera  //!< camera to work with. Manipulator takes care of ref/deref-ing the camera object. 
                 );      

      //! Get the attached camera.
      /** Return the currently attached camera object. 
        * \return Pointer to the attached camera. This pointer could be NULL. 
        * If you want to use this camera, make sure to ref/deref the camera. */
      const nvsg::Camera * getAttachedCamera( void ); 
      
      //! Set the viewport size.
      /** Adapt resize of the render area.
        * \note Call this method on every resize, or if you create/initialize the manipulator.*/
      void setViewportSize( size_t width   //!< viewport width
                          , size_t height  //!< viewport height
                          ); 
      
    protected:
      bool      m_dragAndThrowAnim;  //!< Drag & Throw animation flag (Not implemented yet)
      
      UINT      m_buttonState;       //!< Current button mask.
      UINT      m_keyState;          //!< Current key mask.
      
      const nvsg::Camera  * m_camera;            //!< Attached camera. Could be NULL.
      
      int       m_mouseX;            //!< Current mouse x position.
      int       m_mouseY;            //!< Current mouse y position.

      int       m_mouseXlast;        //!< Last mouse x position.
      int       m_mouseYlast;        //!< Last mouse y position.
      
      int       m_mouseXstart;       //!< Mouse start x position. Set in mouse button pressed handler.
      int       m_mouseYstart;       //!< Mouse start y position. Set in mouse button pressed handler.
      
      size_t    m_vpWidth;           //!< Current viewport width.
      size_t    m_vpHeight;          //!< Current viewport height.

      //! Convert the key.
      /** Translate a key to the CameraManipulator key.
        * \return Corresponding key id. */
      NVSG_API virtual UINT convertKey(UINT key  //!< Currently pressed mouse button (e.g. MK_LBUTTON)
                                      );

      //! Convert the mouse button.
      /** Translate a button to the CameraManipulator mouse button.
        * \return Corresponding mouse button id. */
      NVSG_API virtual UINT convertMouseButton( UINT btn  //!< Currently pressed mouse button (e.g. MK_LBUTTON)
                                              );

  };

  inline void CameraManipulator::attach(const nvsg::Camera * pCamera) 
  { 
    __TRACE();

    if (m_camera)
    {
      m_camera->removeRef();
    }

    m_camera = pCamera;
    
    if (pCamera)
    {
      m_camera->addRef();
    }
  }

  inline void CameraManipulator::setViewportSize( size_t width, size_t height )
  {
    m_vpWidth = width;
    m_vpHeight = height;
  }

  inline const nvsg::Camera * CameraManipulator::getAttachedCamera( void )
  {
    return( m_camera );
  }

  inline UINT CameraManipulator::getButtonState( void ) const
  {
    return( m_buttonState );
  }

  inline UINT CameraManipulator::getKeyState( void ) const
  {
    return( m_keyState );
  }

  inline void CameraManipulator::resetButtonState( void )
  {
    m_buttonState = _NO_BTN_KEY;
  }

  inline void CameraManipulator::resetKeyState( void )
  {
    m_keyState = _NO_BTN_KEY;
  }
} //  namespace nvui
