// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvui/CameraManipulator.h"

namespace nvui
{
  //!Flight Camera Manipulator Class.
  /** This manipulator is a special CameraManipulator that interpretes/converts mouse movement
    * into fly-through camera movements. \n
    * \n
    * <b>Keys & Buttons </b>                  \n
    * Left mouse button     : Fly forward.    \n
    * Middle mouse button   : Fly backward.   \n
    * W-key                 : Increase speed. \n
    * S-key                 : Decrease speed. \n
    * \n
    * \note If you use this manipulator, then your application must use a mechanism (such as a small thread   
    * that triggers redraw, etc.) that provides continuous camera updates when walking forward or backward. \n
    * \n
    * \note If you implement your own manipulator, refer to the button state configuration to see how to  
    * add additional button combinations. 
    */
  class FlightCamManip : public CameraManipulator
  {
    public:
      //! Default constructor
      NVSG_API FlightCamManip(void);

      //! Default destructor
      NVSG_API ~FlightCamManip(void);

      //! Additional keys for flying.
      enum
      { _W_KEY        = 0x0400 //!< Fly forward.
      , _S_KEY        = 0x0800 //!< Fly backward.
      };

      //! Apply the states to the camera.
      /** Recalculate the camera regarding the manipulator states. */
      NVSG_API virtual void apply(void);

      //! Handle key repeats.
      /** Adapt key repeats. */
      NVSG_API virtual void keyRepeat();

      //! Signals if the manipulator is in \e fly mode
      /** This method compares the current button/key state 
        * with the configured button states 
        * of this object. e.g. FLYFORWARD
        * \return true: manipulator is in walk mode
        */
      bool isFly(void);

      //! Set the screen size.
      /** Set the screen resolution. The resolution is needed for the 
        * sensitivity of the manipulator. Default is 1280x1024. */
      void setScreenSize( UINT x  //!< width of the screen (pixels) -  must not be 0 
                        , UINT y  //!< height of the screen (pixels) -  must not be 0
                        );

      //! Set the world up-vector.
      /** Define where the 'sky' is. */
      void setWorldUpVector( nvmath::Vec3f up //!< World up-vector
                           );
      //! Get the world up-vector.
      /** \return Up-vector of the world pointing to the 'sky' */
      const nvmath::Vec3f& getWorldUpVector();

      //! Define speed increase factor.
      /** The current speed will be multiplied or divided by this factor */
      void setSpeedIncreaseFactor(float factor   //!< Speed increase factor - default 2.f \note This factor must not be 0!
                                 );
      
      //! Get the speed increase factor.
      /** \return  Speed increase factor.*/
      float getSpeedIncreaseFactor();

      //! Define horizontal sensitivity.
      /** The horizontal sensitivity is a factor that influences how 
        * sensitive the manipulator reacts to horizontal rotation changes */
      void setHSensitivity(float sensitivity //!< Sensitivity factor - default 1.f 
                          );
      
      //! Get the horizontal sensitivity.
      /** \return  Horizontal sensitivity factor.*/
      float getHSensitivity();

      //! Define vertical sensitivity.
      /** The vertical sensitivity is a factor that influences how 
        * sensitive the manipulator reacts to vertical rotation changes */
      void setVSensitivity(float sensitivity //!< Sensitivity factor - default 1.f
                          );

      //! Get the vertical sensitivity.
      /** \return  Vertical sensitivity factor.*/
      float getVSensitivity();
      
      //! Set the speed.
      /** This parameter controls the speed of the camera movement */
      void setSpeed(float speed //!< Speed value - default 1.f
                   );

      //! Get the movement speed.
      /** \return  Speed value.*/
      float getSpeed();

    protected:
      UINT  m_ScreenSizeX;    //!< Width of the screen (pixels)
      UINT  m_ScreenSizeY;    //!< Height of the screen (pixels)
      
      nvmath::Vec3f m_worldUpVector;  //!< Up-vector of the world.

      float m_speedIncreaseFactor;  //!< Factor that influences the stepping. Default: 2.0 NOTE: This factor must not be 0!
      float m_hSensitivity;         //!< Factor that influences the horizontal rotation. Default: 1.0
      float m_vSensitivity;         //!< Factor that influences the vertical rotation. Default: 1.0

      float m_speed; //!< Flight speed. Controlled via key or button input. Default: 1.0
      
      static NVSG_API const UINT FLYFORWARD;     //!< Key/button configuration flying.
      static NVSG_API const UINT FLYBACKWARD;    //!< Key/button configuration flying.
      static NVSG_API const UINT INCREASESPEED;  //!< Key/button configuration flying.
      static NVSG_API const UINT DECREASESPEED;  //!< Key/button configuration flying.

      static NVSG_API const float AREAOFPEACE;   //!< Size if the area where the mouse does not trigger the manipulator.
      
       //! Map your keys.
      /** \return Manipulator key code */
      NVSG_API virtual UINT convertKey(UINT key  //!< External key code.
                                      );
      //! Do the camera fly.
      /** Fly forward or backward. You can simply 
        * replace it with your own way to handle this. */
      void fly(void);

      //! Signals if the manipulator is in \e forward flying mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. FLYFORWARD
        * \return true: manipulator is in forward walking mode
        */
      bool isFlyForward(void);

      //! Signals if the manipulator is in \e backward flying mode
      /** This method compares the current button or key state 
        * with the configured button states 
        * of this object. e.g. FLYFORWARD
        * \return true: manipulator is in forward walking mode
        */
      bool isFlyBackward(void);

  };

  inline bool FlightCamManip::isFly(void)
  {
    return (isFlyForward() || isFlyBackward());
  }

  inline bool FlightCamManip::isFlyForward(void)
  {
    return(FLYFORWARD == m_buttonState && (  m_keyState == 0 
                                          || m_keyState == INCREASESPEED 
                                          || m_keyState == DECREASESPEED));
  }

  inline bool FlightCamManip::isFlyBackward(void)
  {
    return(FLYBACKWARD == m_buttonState && (  m_keyState == 0
                                           || m_keyState == INCREASESPEED 
                                           || m_keyState == DECREASESPEED));
  }

  inline void FlightCamManip::setScreenSize(UINT x, UINT y)
  {
    __ASSERT(x>0);
    __ASSERT(y>0);
    m_ScreenSizeX = x;
    m_ScreenSizeX = y;
  }

  inline void FlightCamManip::setWorldUpVector(nvmath::Vec3f up)
  {
    m_worldUpVector = up; 
  }
  
  inline const nvmath::Vec3f& FlightCamManip::getWorldUpVector()
  {
    return m_worldUpVector; 
  }

  inline void FlightCamManip::setSpeedIncreaseFactor(float factor)
  {
    __ASSERT(fabs(factor) > FLT_EPSILON);
    m_speedIncreaseFactor = factor; 
  }
  
  inline float FlightCamManip::getSpeedIncreaseFactor()
  {
    return m_speedIncreaseFactor; 
  }

  inline void FlightCamManip::setHSensitivity(float sensitivity)
  {
    m_hSensitivity = sensitivity; 
  }
  
  inline float FlightCamManip::getHSensitivity()
  {
    return m_hSensitivity; 
  }

  inline void FlightCamManip::setVSensitivity(float sensitivity)
  {
    m_vSensitivity = sensitivity; 
  }

  inline float FlightCamManip::getVSensitivity()
  {
    return m_vSensitivity;
  }

  inline void FlightCamManip::setSpeed(float speed)
  {
    m_speed = speed; 
  }

  inline float FlightCamManip::getSpeed()
  {
    return m_speed;
  }

} //  namespace nvui
