// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once
/** \file */

#include "nvsgcommon.h"

#include "nvsg/Scene.h"
#include "nvsg/ViewState.h"
#include "nvtraverser/AppTraverser.h"
#include "nvtraverser/CullTraverser.h"
#include "nvtraverser/RenderTraverser.h"
#include "nvui/CameraManipulator.h"

#include <vector>

namespace nvui
{    
  //! The render area is the place where the actual drawing takes place.
  /** This class provides a generic interface to a render target view that is
    * independent of device, hardware, and operating system specifics.\n 
    * \note By default the render area is created with a FrustumCullTraverser for 
    * the culling step. You need to set the desired AppTraverser and RenderTraverse. */    
  class RenderArea
  {
    public:
 
      //! Destructor.
      NVSG_API virtual ~RenderArea(void);

      //! Create the render area.
      /** This method creates and initializes the render area.
        * \return true - everything went fine */
      NVSG_API virtual bool init( RenderArea * shareArea = NULL  //!< Pointer to a render area to share display lists, textures, vertex programs, etc. with
                                );  

      //! Destroy the render area.
      /** This method destroys and cleans up the render area. */
      NVSG_API virtual void destroy( void );

      //! Handle key down.
      /** React on a key down event. (For example, report it to the camera manipulator.) */
      NVSG_API virtual void keyDown( UINT key  //!< Pressed key.
                                   , bool keyRepeated = false
                                   );
      //! Handle key up.
      /** React on a key down event. (For example, report it to the camera manipulator.) */
      NVSG_API virtual void keyUp( UINT key   //!< Released key.
                                 );

      //! Reset the current key state.
      /** This method also resets the camera manipulator key states.*/
      NVSG_API virtual void resetKeyState( void );

      //! Handle mouse button down.
      /** React on mouse button down events. (For example, report it to the current camera manipulator.) */
      NVSG_API virtual void mouseButtonDown( UINT btn    //!< Pressed button (Windows: MK_LBUTTON, MK_MBUTTON, MK_RBUTTON
                                           , int x       //!< Current x position of the mouse
                                           , int y       //!< Current y position of the mouse
                                           );
      //! Handle mouse button up.
      /** React on mouse button up events. (For example, report it to the current camera manipulator.) */
      NVSG_API virtual void mouseButtonUp( UINT btn    //!< Released buttons (Windows: MK_LBUTTON, MK_MBUTTON, MK_RBUTTON
                                         , int x       //!< Current x position of the mouse
                                         , int y       //!< Current y position of the mouse
                                         );
      //! Reset the current button state.
      /** This method also resets the camera manipulator button states.*/
      NVSG_API virtual void resetButtonState( void );

      //! Handle mouse motion.
      /** React on mouse motion events. (For example, report it to the current camera manipulator.) */
      NVSG_API virtual void mouseMotion( int x  //!< Current x position of the mouse
                                       , int y  //!< Current y position of the mouse
                                       );
    
      //! Render the scene.
      /** Render the entire scene. (app, cull, draw the whole tree) 
        * If you want to trigger this method, use triggerRedraw() to force a redraw 
        * of the scene. */
      NVSG_API virtual void renderScene( void );  

      //! Handle the viewport resizing
      /** Resize the viewport. React on parent window size changes. 
        * Report it to OpenGL, to the render window, or to the camera manipulator. */
      NVSG_API virtual void setViewportSize( size_t width   //!< Width of the viewport
                                           , size_t height  //!< Height of the viewport
                                           );

      //! Pick the cameras' target point.
      /** Determine the target point and report it to the current camera.
        * This method takes the mouse coordinates and sends a ray into the scene 
        * to calculate the nearest intersection with an object from the scene. 
        * It then takes the intersection and reports it to the camera, where the camera 
        * can use it as the rotation point, etc.
        * \return false - no intersection */
      NVSG_API bool pickCameraTargetPoint( int x  //!< Current x position of the mouse
                                         , int y  //!< Current x position of the mouse
                                         );

      //! Get the current render traverser.
      /** \return Current render traverser.
        * \note Render traverser must not live longer than the render area
                since they are device dependent! (See RenderTraverser)*/
      NVSG_API nvtraverser::RenderTraverser * getRenderTraverser( void ) const;
      
      //! Set the current render traverser.
      /** The old render traverser will be unreferenced 
        * and the new one will be referenced. 
        * \note Render traverser must not live longer than the render area
                since they are device dependent! (See RenderTraverser) */
      NVSG_API void setRenderTraverser( nvtraverser::RenderTraverser * tr   //!< New render traverser.
                             );
      
      //! Set the current app traverser.
      /** The old app traverser will be unreferenced 
        * and the new one will be referenced. */
      NVSG_API void setAppTraverser( nvtraverser::AppTraverser * tr   //!< New app traverser.
                                   );

      //! Get the current app traverser.
      /** \return Current app traverser. */
      NVSG_API nvtraverser::AppTraverser * getAppTraverser() const;

      //! Add a cull traverser to the cull pipeline.
      /** This method appends an additional cull traverser to the culling stage 
        * of this render area. Every additional culler will be added to the end 
        * of the culling pipeline.
        * \note Internally we increment the refcount.
        * \note Do not add the same traverser more than once. */
      NVSG_API void pushCullTraverser( nvtraverser::CullTraverser * tr       //!< New cull traverser.
                                     );

      //! Remove a cull traverser from the cull pipeline.
      /** This method removes the last cull traverser from the end of the cull pipeline.
        * \note Internally we decrement the refcount. */
      NVSG_API void popCullTraverser();
      
      //! Get the number of culling stages in the culling pipeline.
      /** \return Number of active culling stages. */
      NVSG_API size_t getNumberOfCullTraverser() const;

      //! Get the specified cull traverser
      /** Retrieve the cull traverser from the specified position 
        * of the culling pipeline.
        * \note Position 0 is the first entry in the culling pipeline.
        * \return Pointer to the specified cull traverser */
      NVSG_API nvtraverser::CullTraverser * getCullTraverser(size_t position) const;

      //! Get the current view state.
      /** \return Current view state. */
      NVSG_API const nvsg::ViewState * getViewState( void ) const;
            
      //! Set the current view state.
      /** This method retrieves the viewstate from the given and 
        * copies the data into the local viewstate. */
      NVSG_API void setViewState( const nvsg::ViewState * viewState );
      
      //! Get the current aspect ratio of the render area.
      /** \return Aspect ratio. (width / height). 
        * 0.f if the height and / or width is not valid.*/
      NVSG_API float getAspectRatio( void ) const;

      //! Check if stereo is available for the current render area.
      /** \note The render area must be created and valid.
        * \return true - stereo is available */
      NVSG_API bool isStereoAvailable( void );
      
      //! Request full-scene antialiasing (FSAA) support. 
      /** \note Call prior to init() */
      NVSG_API void requestFSAA( bool bOn                  //!< true - FSAA on
                               , unsigned int numBuffers   //!< FSAA level / number of multisample buffers
                               );
      //! Request stereo support. 
      /** \note Call prior to init() */
      NVSG_API void requestStereoSupport( bool bStereo     //! true - stereo support required
                                        );

      //! Sets the Scene to be rendered together with render-specific view data.
      /** The function replaces the RenderArea's current Scene by the Scene passed through \a pScene. 
        * The behavior is undefined if a NULL pointer is passed through \a pScene!
        * 
        * The CameraManipulator passed through \a pCamManip replaces the RenderArea's current CameraManipulator. 
        * If a NULL pointer is passed through \a pCamManip, there will be no CameraManipulator attached to the 
        * RenderArea after this call!
        *
        * If a valid CameraManipulator is passed through \a pCamManip, the RenderArea's internal ViewState will
        * always adopt the camera attached to that manipulator, regardless of whether the ViewState passed 
        * through \a pViewState is valid or not. 
        *
        * If the ViewState passed through \a pViewState is valid, it will replace the RenderArea's current ViewState.
        * If a NULL pointer is passed through \a pViewState, the RenderArea's current ViewState remains unchanged. */
      NVSG_API void setSceneData( nvui::CameraManipulator * pCamManip  //!< CameraManipulator to replace the RenderArea's current CameraManipulator.
                                , const nvsg::Scene * pScene           //!< Scene to replace the RenderArea's current Scene.
                                , const nvsg::ViewState * pViewState = NULL  /*!< ViewState to replace the RenderArea's current ViewState.
                                                                                  Will be ignored if a NULL pointer is passed here. */
                                );  

      NVSG_API static size_t createNewRenderAreaID();
      
      //!< Trigger a redraw of the scene
      /** \note Use this method to trigger a redraw 
        * instead of calling the drawing method directly! */
      NVSG_API virtual void triggerRedraw( void );                

    protected:
      //! Default constructor.
      NVSG_API RenderArea(void);
      
      //! Possible features of the render area. (mainly influencing the pixelformat - OGL)
      enum
      { _UNDEFINED    = 0x00000000  //!< No special feature requested.
      , _STEREO       = 0x00000001  //!< Stereo requested.
      , _MULTISAMPLE  = 0x00000002  //!< Multisample buffer requested.
      };

      std::vector<nvtraverser::CullTraverser *>  m_cullTraverser;    //!< These traversers cull objects in the tree.
      nvtraverser::RenderTraverser             * m_renderTraverser;  //!< This traverser renders / displays the tree.
      nvtraverser::AppTraverser                * m_appTraverser;     //!< This traversers applie all pending changes to the tree.

      CameraManipulator * m_camManip;         //!< Camera manipulator. Could be NULL!
            
      
      const nvsg::Scene * m_scene;            //!< Pointer to the root node of the scene. m_scene could be NULL !!
      const nvsg::ViewState * m_viewState;        //!< Pointer to the view specific controls (camera, animation state,...)

      unsigned long       m_features;         //!< All requested features for this render area stereo, multisample, ...
      UINT                m_numMSBuffers;     //!< Number of multi sample buffers
      bool                m_raCreated;        //!< Signal the Create() status
      size_t              m_raHeight;         //!< Height of the render area
      size_t              m_raWidth;          //!< Width of the render area
 
      static size_t       m_idRenderArea;     //!< The child window ID for the different CWnd derived 
                                              //!< render area windows. (e.g. GLMFCRAWnd)
                                              //!< Use this ID in the create call - see CWnd->Create() -
                                              //!< (don't forget to increment the ID before using it).
  };

  // - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // inlines 
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - 

  inline float RenderArea::getAspectRatio( void ) const
  {
    __ASSERT( m_raHeight > 0 && m_raWidth > 0 );

    if (m_raHeight > 0 && m_raWidth > 0)
    {
      return( (float) m_raWidth / m_raHeight );
    }
    
    return 0.f; // error
  }

  inline const nvsg::ViewState * RenderArea::getViewState( void ) const
  {
    return( m_viewState );
  }

  inline void RenderArea::triggerRedraw()
  {    
    if (m_camManip)
    {
      m_camManip->apply();
    }    
  }

  inline nvtraverser::RenderTraverser * RenderArea::getRenderTraverser( void ) const
  {
    return( m_renderTraverser );
  }

  inline void RenderArea::setRenderTraverser( nvtraverser::RenderTraverser * tr )
  {
    if (m_renderTraverser)
    {
      m_renderTraverser->removeRef();
    }

    m_renderTraverser = tr;
    
    if (m_renderTraverser)
    {
      m_renderTraverser->addRef();
    }
  }

  inline void RenderArea::setAppTraverser(nvtraverser::AppTraverser * tr)
  {
    if (m_appTraverser)
    {
      m_appTraverser->removeRef();
    }

    m_appTraverser = tr;

    if (m_appTraverser)
    {
      m_appTraverser->addRef();
    }
  }

  inline nvtraverser::AppTraverser * RenderArea::getAppTraverser() const
  {
    return m_appTraverser;
  }

  inline void RenderArea::pushCullTraverser(nvtraverser::CullTraverser * tr)
  {
    __ASSERT(tr);

    m_cullTraverser.push_back(tr);
    tr->addRef();
  }

  inline void RenderArea::popCullTraverser()
  {
    __ASSERT(!m_cullTraverser.empty());

    m_cullTraverser.back()->removeRef();
    m_cullTraverser.pop_back();
  }
  
  inline size_t RenderArea::getNumberOfCullTraverser() const
  {
    return m_cullTraverser.size();
  }

  inline nvtraverser::CullTraverser * RenderArea::getCullTraverser(size_t position) const
  {
    __ASSERT(m_cullTraverser.size() > position);

    return m_cullTraverser[position];
  }

  inline size_t RenderArea::createNewRenderAreaID()
  {
    return (++m_idRenderArea);
  }

} //  namespace nvui
