// Copyright NVIDIA Corporation 2002-2004
// TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
// *AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
// BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES 

#pragma once

#include "nvsgcommon.h"

#include "nvui/CameraManipulator.h"
#include "nvutil/Trackball.h"

/** \file */

namespace nvui
{    

  //!Trackball Camera Manipulator
  /** This manipulator is a special CameraManipulator that interpretes mouse movement
    * as trackball input. Think of a big sphere centered inside of the viewport
    * where you can lay your hand on and rotate it.
    * The manipulator also allows you to dolly and move the objects of the scene.
    * \note If you implement your own manipulator, refer to the button state to see  
    * how to add additional button combinations.
    */
  class TrackballCamManip : public CameraManipulator
  {
    public:
      //! Default constructor
      NVSG_API TrackballCamManip(void);

      //! Default destructor
      NVSG_API virtual ~TrackballCamManip(void);
      
      //! Apply the states to the camera.
      /** Recalculate the camera regarding the manipulator states. */
      NVSG_API virtual void apply();
      
      //! Set Viewport size.
      /** Adapt resize of the render area and update the size of the trackball.
        * Call this method whenever the viewport size changes or if you create / initialize the manipulator.
        */
      NVSG_API virtual void  setViewportSize( size_t x  //!< Specifies the width of the viewport (pixel).
                                            , size_t y  //!< Specifies the height of the viewport (pixel).
                                            );  
      
      //! Signals if manipulator is in \e dolly mode
      /** This method compares the current button state 
        * with the configured button states 
        * of this object. e.g. DOLLY1, DOLLY2
        * \return true: manipulator is in dolly mode
        */
      bool  isDolly( void );
      
      //! Signals if manipulator is in \e pan mode
      /** This method compares the current button state 
        * with the configured button states 
        * of this object. e.g. PAN1, PAN2
        * \return true: manipulator is in pan mode
        */
      bool  isPan( void );
      
      //! Signals if manipulator is in \e rotate mode
      /** This method compares the current button state 
        * with the configured button states 
        * of this object. e.g. ROTATE1
        * \return true: manipulator is in rotate mode
        */
      bool  isRotate( void );

    protected:
      static NVSG_API const UINT DOLLY1;  //!< Button configuration for dolly.
      static NVSG_API const UINT DOLLY2;  //!< Button configuration for dolly.
      static NVSG_API const UINT PAN1;    //!< Button configuration for pan.
      static NVSG_API const UINT PAN2;    //!< Button configuration for pan.
      static NVSG_API const UINT ROTATE1; //!< Button configuration for pan.
      
      nvutil::Trackball m_trackball;  //!< Trackball object that does all the calculations

      //! Do camera rotation.
      /** Trackball movement of the camera. You can simply 
        * replace it with your own way to handle this. */
      NVSG_API virtual void rotate(void);
      
      //! Do camera panning.
      /** Move the camera in the camera plane. (left, right up, down)
        * You can simply replace it with your own way to handle this. */
      NVSG_API virtual void pan(void);
      
      //! Do the camera dolly.
      /** Dolly in and out. You can simply 
        * replace it with your own way to handle this. */
      NVSG_API virtual void dolly(void);

  };

  inline  bool  TrackballCamManip::isRotate( void )
  {
    return( ROTATE1 == (m_buttonState | m_keyState));
  }

  inline  bool  TrackballCamManip::isPan( void )
  {
    return( PAN1 == (m_buttonState | m_keyState) || PAN2 == (m_buttonState | m_keyState) );
  }

  inline  bool  TrackballCamManip::isDolly( void )
  {
    return( DOLLY1 == (m_buttonState | m_keyState) || DOLLY2 == (m_buttonState | m_keyState) );
  }

} //  namespace nvui
