/***************************************************************************
 *   Copyright (C) 2004 by dsievers@NVIDIA
 *   
 ***************************************************************************/
#include <wx/wx.h>

#include "nvsg/nvsgapi.h"
#include "nvutil/Assert.h"

#include "GLwxRAWnd.h"
#include "nvtraverser/GLTraverser.h"
#include "nvui/CameraManipulator.h"

// enable memory leak detection
#include "nvutil/DbgNew.h"

using namespace nvgl;

using nvtraverser::GLTraverser;
using nvtraverser::RenderTraverser;
using nvui::CameraManipulator;


BEGIN_EVENT_TABLE(GLwxRAWnd, wxGLCanvas)
  EVT_PAINT(GLwxRAWnd::OnDraw)
  EVT_SIZE(GLwxRAWnd::OnSize)
  EVT_LEFT_DOWN(GLwxRAWnd::OnLButtonDown)
  EVT_LEFT_UP(GLwxRAWnd::OnLButtonUp)
  EVT_MIDDLE_DOWN(GLwxRAWnd::OnMButtonDown)
  EVT_MIDDLE_UP(GLwxRAWnd::OnMButtonUp)
  EVT_RIGHT_DOWN(GLwxRAWnd::OnRButtonDown)
  EVT_RIGHT_UP(GLwxRAWnd::OnRButtonUp)
  EVT_MOTION(GLwxRAWnd::OnMouseMove)
  EVT_KEY_DOWN(GLwxRAWnd::OnKeyDown)
END_EVENT_TABLE()


GLwxRAWnd::GLwxRAWnd(wxWindow* parent, int* attribList, wxWindowID id)
 : wxGLCanvas(parent, id, wxPoint(0,0), wxSize(0,0), wxSUNKEN_BORDER, "GLView",  attribList)
 , m_drawable(NULL)
 , m_ctx(NULL)
 , m_dpy(NULL)
 , m_viewMode(VM_VIEWING)
{    
}


GLwxRAWnd::~GLwxRAWnd()
{
}

bool GLwxRAWnd::init( RenderArea * shareArea )
{
  // Use wxWidget call for MakeCurrent() and then
  // retrieve the GLXContext and Drawable handles
  SetCurrent();
  GLXContext ctx = m_ctx = GLXGetCurrentContext();
  GLXDrawable drw = m_drawable = GLXGetCurrentDrawable();
  Display* dpy = m_dpy = (Display*)wxGetDisplay();  
  
  __ASSERT( dpy && drw && ctx );  

  
  // introduce the GL render context to nvsg!
  GLXAttachContext(dpy, ctx);  // this must be called if you do not create the OpenGL context through our GLXCreateContext() wrapper!!
                               // (e.g. when a GUI library calls the regular glXCreateContext() in its implementation)
  
  // must be current for the creation of the render traverser
  // (see remarks in makeCurrent() routine!)
  makeCurrent();
   
  // init non OGL stuff
  if ( dpy && drw && ctx && RenderArea::init(shareArea))
  {         
    // create traverser render traverser
    __ASSERT(m_renderTraverser == NULL);
    __VERIFY(m_renderTraverser = new GLTraverser);
    m_renderTraverser->addRef();    

    return true;
  }          
  
  return false;
}
   
void GLwxRAWnd::setViewportSize( size_t width, size_t height )
{  
  // (see remarks in makeCurrent() routine!)
  makeCurrent();
  GLViewport( 0, 0, (GLsizei)width, (GLsizei)height );

  RenderArea::setViewportSize(width, height);
}
      
void GLwxRAWnd::OnDraw(wxPaintEvent& event)
{  
  renderScene();
}

bool GLwxRAWnd::Destroy()
{
  destroy();

  return Destroy();  // wxWindow::Destroy()
}

void GLwxRAWnd::destroy( void )
{      
 
  // destroy non-OGL stuff like traversers    
  RenderArea::destroy();  
   
  // We need this for the cleanup of e.g. the render traverser
  // (see remarks in makeCurrent() routine!)
  makeCurrent();
  
  if (m_renderTraverser)
  {
    m_renderTraverser->removeRef();
    m_renderTraverser = NULL;
  }
    
  GLXDetachContext(m_ctx);  // this must be called if you do not destroy the OpenGL context through our GLXDestroyContext() wrapper!!
                            // (e.g. when a GUI library calls the regular glXDestroyContext() in its implementation)

  m_drawable=NULL;
  m_ctx=NULL;
  m_dpy=NULL;
  
}
      
void GLwxRAWnd::triggerRedraw()
{
  // apply camera manipulator...
  RenderArea::triggerRedraw();
    
  Refresh(FALSE);
}

void GLwxRAWnd::renderScene()
{  
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if ( m_scene ) 
  {
    GLClearColor(m_scene->getBackColor()[0], m_scene->getBackColor()[1], m_scene->getBackColor()[2], 0.0f);
  }
  GLClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  // Then initiate cull/app/render traversal through base class.
  RenderArea::renderScene();

  SwapBuffers();
}

void GLwxRAWnd::makeCurrent()
{
  __ASSERT(m_ctx);
  
  wxGLCanvas::SetCurrent();

  // Because of internal nvsg data structures, we have to call our GLXMakeCurrent() wrapper every time,
  // even if wxWidgets::SetCurrent() above has already made the context current.  
  // (Also take a look at the GXLAttachContext()/GLXDetachContext() call in init() and destroy()!)
  GLXMakeCurrent(m_dpy, m_drawable, m_ctx);
  
}

void GLwxRAWnd::OnSize(wxSizeEvent& event)
{
  int width, height;
  GetClientSize(&width, &height);
  
  if((width>0) && (height>0))
  {
    setViewportSize((size_t)width, (size_t)height);
  }
}

void GLwxRAWnd::OnLButtonDown(wxMouseEvent& event)
{
  if(!event.m_controlDown)
  {
    mouseButtonDown(CameraManipulator::_LEFT_MOUSE_BTN, event.m_x, event.m_y);
  }  
}

void GLwxRAWnd::OnLButtonUp(wxMouseEvent& event)
{   
  if(event.m_controlDown || (m_viewMode==VM_SET_ROTATION_POINT))
  {
    pickCameraTargetPoint(event.m_x, event.m_y);
    m_viewMode=VM_VIEWING;    
    SetCursor(wxNullCursor);  // set back to default

    triggerRedraw();
  }

  // always release mouse button to be sure
  mouseButtonUp(CameraManipulator::_LEFT_MOUSE_BTN, event.m_x, event.m_y);
}

void GLwxRAWnd::OnMButtonDown(wxMouseEvent& event)
{
  mouseButtonDown(CameraManipulator::_MIDDLE_MOUSE_BTN, event.m_x, event.m_y);
}

void GLwxRAWnd::OnMButtonUp(wxMouseEvent& event)
{
  mouseButtonUp(CameraManipulator::_MIDDLE_MOUSE_BTN, event.m_x, event.m_y);
}

void GLwxRAWnd::OnRButtonDown(wxMouseEvent& event)
{
  mouseButtonDown(CameraManipulator::_RIGHT_MOUSE_BTN, event.m_x, event.m_y);
}

void GLwxRAWnd::OnRButtonUp(wxMouseEvent& event)
{
  mouseButtonUp(CameraManipulator::_RIGHT_MOUSE_BTN, event.m_x, event.m_y);
}

void GLwxRAWnd::OnMouseMove(wxMouseEvent& event)
{
  mouseMotion(event.m_x, event.m_y);
  
  if( m_camManip->getButtonState() != nvui::CameraManipulator::_NO_BTN_KEY
      || m_camManip->getKeyState() != nvui::CameraManipulator::_NO_BTN_KEY)
  {
    triggerRedraw();
  }

}

void GLwxRAWnd::OnKeyDown(wxKeyEvent& event)
{
  if(event.m_controlDown)
  {
    switch(event.GetKeyCode())
    {
    case 'V': // View All
      if(m_scene->getRootNode())
      {
        if(m_camManip->getAttachedCamera())
        {
          m_camManip->getAttachedCamera()->zoom(m_scene->getRootNode()->getBoundingSphere());
          triggerRedraw();
        }
      }
      break;

    default:
      break;
    }
  }
}

void GLwxRAWnd::setViewMode(ViewMode mode)
{  
  if(mode==VM_SET_ROTATION_POINT)
  {
    SetCursor(wxCursor(wxCURSOR_CROSS));
  }
  else if(mode==VM_VIEWING)
  {
    SetCursor(wxNullCursor);  // set back to default
  }

  m_viewMode = mode;
}
