/***************************************************************************
 *   Copyright (C) 2004 by dsievers@NVIDIA
 *   
 ***************************************************************************/

#include <wx/wx.h>

#include "nvsg/nvsgapi.h"
#include "nvutil/Assert.h"

#include "vwrdocument.h"
#include "vwrview.h"

#include "nvsg/Node.h"
#include "nvsg/Triangles.h"
#include "nvsg/Transform.h"
#include "nvsg/GeoNode.h"
#include "nvsg/PointLight.h"
#include "nvsg/ViewState.h"

#include "nvsg/PlugInterface.h"
#include "nvsg/PlugInterfaceID.h"
#include "nvutil/PlugIn.h"
#include "nvutil/Tools.h"

// enable memory leak detection
#include "nvutil/DbgNew.h"

using namespace nvsg;
using namespace nvmath;
using std::string;
using std::vector;
using nvutil::GetDirFromPath;

IMPLEMENT_DYNAMIC_CLASS(VwrDocument, wxDocument)

// define a unique plug-interface ID for SceneLoader 
const nvutil::UPITID PITID_SCENE_LOADER(UPITID_SCENE_LOADER, UPITID_VERSION);
const nvutil::UPITID PITID_SCENE_SAVER(UPITID_SCENE_SAVER, UPITID_VERSION);

VwrDocument::VwrDocument()
 : wxDocument()
 , m_pScene(NULL)
 , m_pViewState(NULL)
{
}


VwrDocument::~VwrDocument()
{
}

bool VwrDocument::OnCreate(const wxString& path, long flags)
{   
  return wxDocument::OnCreate(path, flags);
}

bool VwrDocument::OnCloseDocument()
{
  DeleteContents(); // default behaviour
}


bool VwrDocument::OnOpenDocument(const wxString& filename)
{
  __ASSERT(m_pScene == NULL);

  printf("Loading %s...\n", filename.GetData());
  
  // Set appropriate search paths for the 
  // loader dll and the sample file.
  string modulePath;
  string curDir;
  string dir;
  vector<string> searchPaths;

  nvutil::GetCurrentDir(curDir);  
  searchPaths.push_back(curDir);
  
  nvutil::GetModulePath(modulePath);
  GetDirFromPath(modulePath, dir);
  searchPaths.push_back(dir);
  
  // also add the scene path  
  string fullpath=filename.GetData();
  GetDirFromPath(fullpath, dir);
  searchPaths.push_back(dir);

  // Search for the right loader dll. In this case
  // we want to load nbf files.
  bool foundAppropriate = false;
  vector<nvutil::UPIID> piids;
  if ( nvutil::queryInterfaceType(searchPaths, PITID_SCENE_LOADER, piids) )
  {
    // extract file extension from filename
    string ext;
    nvutil::GetFileExtFromPath(fullpath, ext);

    nvutil::UPIID plugID;
    vector<nvutil::UPIID>::iterator it = piids.begin();
    for ( ; it != piids.end(); ++it )
    {
      if ( !_stricmp((*it).getPlugSpecificIDString(), ext.c_str()) )
      {
        // found, copy the ID
        plugID = *it;
        foundAppropriate = true;
        break; // look no further
      }
    }
  
    // If there is a loader available, 
    // get the interface, load the file and 
    // release the interface when done.
    if ( foundAppropriate )
    {
      nvutil::PlugIn * plug; 
      if ( nvutil::getInterface(searchPaths, plugID, plug) )
      {
        nvsg::SceneLoader * loader = reinterpret_cast<SceneLoader*>(plug);
 
        __ASSERT(m_pViewState==NULL);
        string fname=filename.GetData();
        m_pScene = loader->load( (const string)fname, searchPaths, m_pViewState );
        if( m_pScene ) 
        { 
          m_pScene->addRef();
        }
        if(m_pViewState)
        {
          m_pViewState->addRef();
        }
        
        nvutil::releaseInterface(plugID);
      }
    }
  }
  
  if (!foundAppropriate)
  {
    wxMessageBox("No scene loader found!", "Error", wxOK | wxICON_EXCLAMATION );
  }
  
  if( m_pScene )
  {
    ((VwrView*)GetFirstView())->init();
    return true;
  }
  
  return false;
}

bool VwrDocument::OnSaveDocument(const wxString& filename)
{
  if ( m_pScene ) 
  {
    // Set search paths to search for an appropriate saver plug-in
    string modulePath;
    string curDir;
    string dir;
    vector<string> searchPaths;

    // ...add the cwd
    nvutil::GetCurrentDir(curDir);  
    searchPaths.push_back(curDir);
    // ...add the module path
    nvutil::GetModulePath(modulePath);
    GetDirFromPath(modulePath, dir);
    searchPaths.push_back(dir);

    // turn filename into a stl string
    string fname(filename.GetData());
    // Search for the right saver dll.
    bool foundAppropriate = false;
    vector<nvutil::UPIID> piids;
    if ( nvutil::queryInterfaceType(searchPaths, PITID_SCENE_SAVER, piids) )
    {
      // extract file extension from filename
      string ext;
      nvutil::GetFileExtFromPath(fname, ext);

      nvutil::UPIID plugID;
      vector<nvutil::UPIID>::iterator it = piids.begin();
      for ( ; it != piids.end(); ++it )
      {
        if ( !_stricmp((*it).getPlugSpecificIDString(), ext.c_str()) )
        {
          // found, copy the ID
          plugID = *it;
          foundAppropriate = true;
          break; // look no further
        }
      }

      // If there is a loader available, 
      // get the interface, load the file and 
      // release the interface when done.
      if ( foundAppropriate )
      {
        nvutil::PlugIn * plug; 
        if ( nvutil::getInterface(searchPaths, plugID, plug) )
        {
          nvsg::SceneSaver * saver = reinterpret_cast<SceneSaver*>(plug);

          saver->save(*m_pScene, m_pViewState, fname );
          nvutil::releaseInterface(plugID);
        }
      }
    }
  }
}

bool VwrDocument::DeleteContents()
{  
  wxDocument::DeleteContents();
  
  if ( m_pScene )
  {
    m_pScene->removeRef();
    m_pScene = NULL;
  }
  if(m_pViewState)
  {
    m_pViewState->removeRef();
    m_pViewState = NULL;
  }
}

