#include "./glutRenderArea.h"

#include "nvmath/nvmath.h"
#include "nvmath/Vec3f.h"
#include "nvsg/Material.h"
#include "nvsg/PointLight.h"
#include "nvsg/PerspectiveCamera.h"
#include "nvsg/Transform.h"
#include "nvsg/Node.h"
#include "nvsg/Triangles.h"
#include "nvsg/GeoNode.h"
#include "nvsg/FaceAttribute.h"
#include "nvgl/GLContext.h"
#include "nvsg/Scene.h"
#include "nvtraverser/GLTraverser.h"

#include <GL/glut.h>

#include "nvutil/DbgNew.h" // this must be the last include

using namespace nvgl;
using namespace nvsg;
using nvtraverser::GLTraverser;
using nvui::RenderArea;

glutRenderArea::glutRenderArea(void)
: m_initDone(false)
, m_scene(NULL)
, m_hglrc(NULL)
#if defined(_WIN32)
, m_hdc(NULL)
#else if defined(LINUX)
, m_dpy(NULL)
#endif
{
}

glutRenderArea::~glutRenderArea(void)
{
}

bool glutRenderArea::init( RenderArea * shareArea )
{       

  // The GL context has to be current, once you call in here!!

#if defined(_WIN32)
  m_hglrc=WGLGetCurrentContext();
  m_hdc=wglGetCurrentDC();  
  
  // introduce the GL render context to nvsg!
  WGLAttachContext(m_hdc, m_hglrc);   // this must be called if you do not create the OpenGL context through our WGLCreateContext() wrapper!!
                                      // (e.g. when a GUI library calls the regular wglCreateContext() in its implementation)
#else if defined(LINUX)
  GLXContext ctx = m_hglrc = GLXGetCurrentContext();
  GLXDrawable drw = m_drawable = GLXGetCurrentDrawable();
  Display* dpy = m_dpy = glXGetCurrentDisplay(); 

  // introduce the GL render context to nvsg!
  GLXAttachContext(m_dpy, m_hglrc);   // this must be called if you do not create the OpenGL context through our GLXCreateContext() wrapper!!
                                    // (e.g. when a GUI library calls the regular glXCreateContext() in its implementation)
#endif

  // must be current for the creation of the render traverser
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if ( RenderArea::init(shareArea))
  {         
    // create traverser render traverser
    __ASSERT(m_renderTraverser == NULL);
    __VERIFY(m_renderTraverser = new GLTraverser);   // have to create a GLTraverser
    m_renderTraverser->addRef();

    m_initDone = true;
    return true;
  }          

  return false;
}

void glutRenderArea::destroy( void )
{
  // destroy non-OGL stuff like traversers    
  RenderArea::destroy();  

  // must be current for the cleanup of e.g. the render traverser  
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if (m_renderTraverser)
  {
    m_renderTraverser->removeRef();
    m_renderTraverser = NULL;
  }

#if defined(_WIN32)
  WGLDetachContext(m_hglrc);  // this must be called if you do not destroy the OpenGL context through our WGLDeleteContext() wrapper!!
                              // (e.g. when a GUI library calls the regular wglDeleteContext() in its implementation)
#else if defined(LINUX)
  GLXDetachContext(m_hglrc);   // this must be called if you do not destroy the OpenGL context through our GLXDestroyContext() wrapper!!
                             // (e.g. when a GUI library calls the regular glXDestroyContext() in its implementation)
#endif
}

void glutRenderArea::setViewportSize( size_t width, size_t height )                                  
{
  RenderArea::setViewportSize(width, height);
  
  if(m_hglrc!=NULL)
  {    
    // see remarks in makeCurrent() routine!
    makeCurrent();
    GLViewport( 0, 0, (GLsizei)width, (GLsizei)height );
  } 
}

void glutRenderArea::triggerRedraw()
{
  // apply camera manipulator etc.
  RenderArea::triggerRedraw();

  // invalidate glut window
  glutPostRedisplay();
}

void glutRenderArea::renderScene( void )
{
  if(!m_initDone)
  {
    init();   // do initialization from here, because we need the context to be current
              // (and don't know the handle yet)
  }

  // must be current now before doing any GL calls or render traversal through NVSG
  // (see remarks in makeCurrent() routine!)
  makeCurrent();

  if ( m_scene ) 
  {
    GLClearColor(m_scene->getBackColor()[0], m_scene->getBackColor()[1], m_scene->getBackColor()[2], 0.0f);
  }
  GLClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  // Then initiate cull/app/render traversal through base class.
  RenderArea::renderScene();

  glutSwapBuffers();
}
